/******************************************************************************
*
* Utility for loading data from .unl files into Informix database
*
* Author: Robert Sosnowski, Robert.Sosnowski@bzwbk.pl
*
* Version 1.70, 17-10-2005
*
* Copyright (c) 2001 - 2007. All Rights Reserved.
*
* Permission granted to use & distribute for free.
*
* This software can be used only at your own risk. Author is not responsible 
* for any damage caused by using this software.
* 
******************************************************************************/

#include "load2cfg.h"

#ifndef HAVE_GETOPT
extern int      optind;
extern char    *optarg;
extern int	opterr;
int getopt(int argc, char *argv[], char *optionS);
#endif

#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "load2.h"
#include <time.h>

static char *id="@(#) $Id: load.ec 2521 2007-11-21 08:55:56Z sosnor01 $";

int usage( char * );
int is_critical_dups(long int sqlcode, long int isamcode, int errtype,
                     int errcnt, int maxerrors, int *errinc);
int is_critical_std(long int sqlcode, long int isamcode, int errtype,
                    int errcnt, int maxerrors, int *errinc);
#ifdef HAVE_ZLIB
int oninitgz(const char *fname, const char *tablename, const int batchsize,
               void **handle, const void *logfile, const int fieldsnum,
               const char *insstmt2,
               onlogerror_lpt onlogerror, int verbose,
               const char *datahandleoptions);
int onreadfieldgz(const void *fd, const int sqltype, const int4 sqlxid,
                  const char *filename,
                  char **dst, const unsigned int maxdstbuf, const char delim,
                  const char escapech, int * eol, int * sqlind,
                  const unsigned long int rownum,
                  const unsigned long int loaded, const int fieldnum,
                  const char * fieldname, const int max_fld_name,
                  char *fieldname2, const void *logfile,
                  onlogerror_lpt onlogerror, const int verbose,
                  const char *datahandleoptions);
int oncrlfgz(const void *fd, const char *filename,
             const unsigned long int rownum, const unsigned long int loaded,
             const int fields_number, const int bequiet, const char escapech,
             const void *logfile,
             onlogerror_lpt onlogerror, int verbose,
             const char *datahandleoptions);
void oncleanupgz(const char *fname, void *handle, const void *logfile,
                 onlogerror_lpt onlogerror, int verbose,
                 const char *datahandleoptions);
#endif


int main( int argc, char **argv )
{

  EXEC SQL BEGIN DECLARE SECTION;
  static char dbname[TBLNAME_SIZE];
  static char server[TBLNAME_SIZE];
  static char server2[TBLNAME_SIZE+1];
  static char user[USERNAME_SIZE];
  static char password[USERNAME_SIZE];
  static char  tabname[TBLNAME_SIZE];
  EXEC SQL END DECLARE SECTION;

  static char  qstr[5000];
  static char  sqlfilename[128];
  static char  logfilename[128];
  static char  infile[128];
  short verbosity, use_ins_curs, use_commit, benchm, gzcompr;
  time_t t1, t2;
  int batchsize, maxerrors, argind;
  unsigned long int loaded_rows;
  int ch;
  FILE * fsqlfile, *logfile;
  int i;
  char *p;
  int (*is_critical)(long int sqlcode, long int isamcode, int errtype,
                     int errcnt, int maxerrors, int *errinc);

  is_critical=is_critical_std;
  verbosity=3; /* show errors and progress messages */
  use_ins_curs=0;
  logfilename[0]=0;
  sqlfilename[0]=0;
  batchsize=10000;
  server[0]=0;
  user[0]=0;
  password[0]=0;
  maxerrors=0;
  logfile=NULL;
  benchm=0;
  gzcompr=0;
 
  setvbuf(stdout,NULL,_IONBF,0);
  if ( argc == 1 ) {
    usage( argv[0] );
    exit( -1 );
  }
    

#ifdef HAVE_ZLIB
  while ((ch = getopt( argc, argv, "Vd:s:u:p:e:t:c:n:kzi:v:l:Bg" )) != EOF) {
#else
  while ((ch = getopt( argc, argv, "Vd:s:u:p:e:t:c:n:kzi:v:l:B" )) != EOF) {
#endif
    switch (ch) {
      case 'V':
        fprintf( stdout,
         "\nload ver 1.70, Copyright 2001 - 2007 Robert Sosnowski\n\n" );
        exit( 0 );
        break;
      case 'd':
        /* database name */
        strcpy( dbname, optarg );
        break;
      case 's':
        /* informix server */
        strcpy( server, optarg );
        break;
      case 'u':
        /* user */
        strcpy( user, optarg );
        break;
      case 'p':
        /* password */
        strcpy( password, optarg );
        break;
      case 'e': 
        /* SQL statement */
        strcpy( qstr, optarg );
        break;
      case 't': 
        /* table name */
        strcpy( tabname, optarg );
        break;
      case 'c': 
        /* file with SQL statement */
        strcpy( sqlfilename, optarg );
        break;
      case 'n':   
        /* batchsize */
        batchsize=(int)atoi( optarg );
        break;
      case 'k':
        /* ignore unique key violation */
        is_critical=is_critical_dups;
        break;
      case 'z':
        /* use insert cursor */
        use_ins_curs = 1;
        break;
      case 'i':
        /* ignore # bad rows */
        maxerrors = (int)atoi( optarg );
        break;
      case 'l':
        /* log file */
        strcpy( logfilename, optarg );
        break;
      case 'v':
        /* verbosity */
        verbosity=(int)atoi( optarg );
        break;
      case 'B':
        /* Benchmark mode */
	benchm=1;
        break;
      case 'g':
        /* gzipped */
        gzcompr=1;
        break;
      default:
        usage( argv[0] );
        exit(2);
    }
  }

  if (strlen( dbname ) == 0) {
    fprintf( stdout, "Must supply database (-d)!\n" );
    usage( argv[0] );
    exit(2);
  }
  FetBufSize=30000;
    
  EXEC SQL WHENEVER ERROR CONTINUE;

  if (strlen(server)==0 && (p = strstr(dbname, "@")))
  {
    p[0] = '\0';
    strcpy(server, p+1);
  }

  if (strlen(server)!=0 || strlen(user)!=0 || strlen(password)!=0) {
    if (strlen(server)==0) {
      fprintf( stderr, "Informix server should be specified with -s option\n");
      return 1;
    }
    strcpy(server2,"@");
    strcat(server2, server);
    if ( strlen(user)==0 && strlen(password)==0) {
      EXEC SQL CONNECT TO :server2 AS "conn1" WITH CONCURRENT TRANSACTION;
      if (sqlca.sqlcode<0) {
        fprintf( stderr, "Error while connecting server %s:"INT4_FORMAT"\n", server,
	         sqlca.sqlcode);
	return 1;
      }
    } else {
      if ( strlen(user)==0 || strlen(password)==0) {
    	fprintf( stderr, "Both options: -u and -p should be provided\n");
        return 1;
      }
      EXEC SQL CONNECT TO :server2 AS "conn1" USER :user USING :password
               WITH CONCURRENT TRANSACTION;
      if (sqlca.sqlcode<0) {
        fprintf( stderr, "Error while connecting server %s as user %s:"INT4_FORMAT"\n",
	         server, user, sqlca.sqlcode);
        return 1;
      }
    }
  } else {
    EXEC SQL CONNECT TO DEFAULT WITH CONCURRENT TRANSACTION;
    if (sqlca.sqlcode<0) {
      fprintf( stderr, "Error while connecting default server: "INT4_FORMAT"\n",
               sqlca.sqlcode);
      return 1;
    }
  }
  EXEC SQL DATABASE :dbname;
  if (sqlca.sqlcode < 0) {
    fprintf( stderr, "Cannot open database.  Code: "INT4_FORMAT", ISAM: "INT4_FORMAT".\n",
             sqlca.sqlcode, sqlca.sqlerrd[1] );
    return 1;
  }
  use_commit = (sqlca.sqlwarn.sqlwarn1 == 'W');

  if (strlen( qstr ) != 0 ) {
    if (strlen(tabname)!=0) {
      fprintf(stderr, "You can't specify both -e and -t options!\n");
      return 1;
    }
    if (strlen(sqlfilename)!=0) {
      fprintf(stderr, "You can't specify both -e and -c options!\n");
      return 1;
    }
  } else {
    if (strlen(tabname) != 0) {
      if (strlen(sqlfilename)!=0) {
        fprintf(stderr, "You can't specify both -t and -c options!\n");
        return 1;
      }
      sprintf(qstr, "insert into %s", tabname);
    } else {
      if (strlen(sqlfilename)!=0) {
        fsqlfile=fopen(sqlfilename, "r");
        if (fsqlfile) {
          i=fread(qstr, 1, sizeof(qstr), fsqlfile);
          qstr[i]=0;
          fclose(fsqlfile);
        } else {
          perror("Can't open sql file");
          return 1;
        }
      } else {
        fprintf(stderr, "You have to specify one of: -e -t or -c option!\n");
        return 1;
      }
    }
  }

  if (verbosity<0 || verbosity>4) {
    fprintf(stderr, "Wrong verbosity value\n");
    return 1;
  }
  if (gzcompr<0 || gzcompr>9) {
    fprintf(stderr, "Wrong gz compression value\n");
    return 1;
  }
  if (logfilename[0]) {
    logfile=fopen(logfilename,"a");
#ifdef _IONBF
    if (logfile!=NULL) {
      setvbuf(logfile, NULL, _IONBF,0);
    }
#endif
  }

  strcpy(infile,"-");
  argind=optind;
  do {
    if (argind<argc) {
      strcpy(infile, argv[argind]);
    }
    if (strcmp(infile,"-")==0) {
      infile[0]=0;
    }
    if (benchm) {
      t1=time(NULL);
      if (logfilename[0]) {
        fprintf(logfile, "Loading started at %s",ctime(&t1));
      }
      fprintf(stdout, "Loading started at %s",ctime(&t1));
    }
#ifdef HAVE_ZLIB
    if (gzcompr>0) {
      if (load(qstr, infile, '|', batchsize, use_ins_curs, use_commit, maxerrors,
               &loaded_rows, logfile, verbosity, &oninitgz, NULL, NULL, NULL, &onreadfieldgz, &oncleanupgz, NULL, &oncrlfgz,
               NULL, is_critical)!=0) {
        return 1;
      }
    } else {
#endif
      if (load(qstr, infile, '|', batchsize, use_ins_curs, use_commit, maxerrors,
               &loaded_rows, logfile, verbosity, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
               NULL, is_critical)!=0) {
        return 1;
      }
#ifdef HAVE_ZLIB
    }
#endif
    if (benchm) {
      t2=time(NULL);
      i=t2-t1;
      if (logfilename[0]) {
        fprintf(logfile, "Loading finished at %s",ctime(&t2));
        fprintf(logfile, "Elapsed seconds: %d\n",i);
      }
      fprintf(stdout, "Loading finished at %s",ctime(&t2));
      fprintf(stdout, "Elapsed seconds: %d\n",i);
    }
  } while (++argind <argc);
    
  if (logfilename[0]) {
    fclose(logfile);
  }

  return 0;
}


int usage( char *prog )
{


  fprintf( stdout, "\n");
  fprintf( stdout, "%s - load *.unl into Informix tables.\n\n", prog );

  fprintf( stdout, "Usage: \n" );
#ifdef HAVE_ZLIB
  fprintf( stdout, 
           "\t%s [-V] -d <db_name> [-s <server> -u <user> -p <password>]\n "
           "(-e <insert stmt>|-t <table>|-c <sql file>) [-n <batch size>]\n "
           "[-k] [-z] [-i <# bad rows>] [-l <logfile>] [-B] [-g] [-v <verbosity>] \n "
           "<unload files>\n\n", prog);
#else
  fprintf( stdout, 
           "\t%s [-V] -d <db_name> [-s <server> -u <user> -p <password>]\n "
           "(-e <insert stmt>|-t <table>|-c <sql file>) [-n <batch size>]\n "
           "[-k] [-z] [-i <# bad rows>] [-l <logfile>] [-B] [-v <verbosity>] \n "
           "<unload files>\n\n", prog);
#endif

  fprintf( stdout, "\t\t-V - version information\n" ); 
  fprintf( stdout, "\t\t-d - database name\n" );
  fprintf( stdout, "\t\t-s - informix server name\n" );
  fprintf( stdout, "\t\t-u - user\n" );
  fprintf( stdout, "\t\t-p - password\n" );
  fprintf( stdout, "\t\t-e - SQL insert statement in quotes\n" );
  fprintf( stdout, "\t\t-t - table name\n" );
  fprintf( stdout, "\t\t-c - file containing SQL insert statement\n" );
  fprintf( stdout, "\t\t-n - batch size\n" );
  fprintf( stdout, "\t\t-k - ignore unique key violation\n" );
  fprintf( stdout, "\t\t-z - use insert cursor\n" );
  fprintf( stdout, "\t\t-i - ignore (if posible) specified number of errors\n" );
  fprintf( stdout, "\t\t-l - log file name\n" );
  fprintf( stdout, "\t\t-B - benchmark\n" );
#ifdef HAVE_ZLIB
  fprintf( stdout, "\t\t-g - gzipped unload\n" );
#endif
  fprintf( stdout, "\t\t-v - verbosity:\n" );
  fprintf( stdout, "\t\t     0 - completely quiet\n" );
  fprintf( stdout, "\t\t     1 - show errors only\n" );
  fprintf( stdout, "\t\t     2 - summary messages\n" );
  fprintf( stdout, "\t\t     3 - detailed messages (default)\n" );
  fprintf( stdout, "\t\t     4 - debug info\n" );

  return 0;
} 

int is_critical_dups(long int sqlcode, long int isamcode, int errtype,
                     int errcnt, int maxerrors, int *errinc)
{
/* errtype:
   0 - SQL error
   1 - inserting null into non-null column
   2 - conversion from ASCII format failed
   3 - error while reading (onreadfield, oncrlf)
   4 - wrong number of columns
*/
  (*errinc)=1;
  if (errtype==0 && (sqlcode==-239 || sqlcode == -268)) {
    (*errinc)=0;
    return 0;
  }
  if (errcnt>=maxerrors) { /* test before errcnt++ */
    return 1;
  }
  return 0;
}
