/******************************************************************************
*
* Library of data loading functions for Informix database
*
* Author: Robert Sosnowski, Robert.Sosnowski@bzwbk.pl
*
* Version 1.70, 17-10-2005
*
* Copyright (c) 2001 - 2007. All rights reserved.
*
* Permission granted to use & distribute for free. Changes in source code
* require Author's acceptance.
*
* This software can be used only at your own risk. Author is not responsible 
* for any damage caused by using this software.
* 
******************************************************************************/

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif

EXEC SQL include sqltypes;
EXEC SQL include decimal;
EXEC SQL include sqlstype;
EXEC SQL include locator;

#include "load2cfg.h"
#include "load2.h"
#include "tempfile.h"
#include <stdio.h>
#ifdef HAVE_CTYPE_H
#include <ctype.h>
/* for isdigit on MSVC */
#endif
#include <stdlib.h>
#ifdef HAVE_UNISTD_H
/* unlink declaration */
#include <unistd.h>
#endif
#ifdef HAVE_IO_H
/* MSVC declaration of unlink */
#include <io.h>
#endif

#define MAX_FIELDNAME TBLNAME_SIZE+1

#define BROKEN_FREE 1


#ifndef HAVE_SNPRINTF
#ifdef HAVE__SNPRINTF
#define snprintf _snprintf
#else
#include <stdarg.h>
int snprintf(char *str, size_t count, ...)
{int i;
 va_list args;
 i= sprintf(str, args);
 va_end(args);
return i;
} 
#endif
#endif

#ifndef HAVE_SQLFREEMEM
#define SqlFreeMem sqlfreemem
#endif

#ifndef HAVE_STRICMP
#ifdef HAVE_STRCASECMP
#define stricmp strcasecmp
#else
static int stricmp(const char *s1, const char *s2)
{ int i;

  i=0;
  while(tolower(s1[i])==tolower(s2[i])) {
    if (s1[i++]=='\0') {
      return 0;
    }
  }
  return(tolower(s1[i])-tolower(s2[i]));
}
#endif
#endif


static char *id="@(#) $Id: load2.ec 2521 2007-11-21 08:55:56Z sosnor01 $";
static unsigned long int instance_id=0;

static void cleanup(loader_handle_t *loader_handle);

static int convert_ins(loader_handle_t *loader_handle);

static int convert_upd(loader_handle_t *loader_handle);

static int convert_sel(loader_handle_t *loader_handle);

static int convert_del(loader_handle_t *loader_handle);

static void extract_tablename(const char *qstr, char *tablename);

static int field_lookup(const loader_handle_t *loader_handle,
                        const char *fieldname);



int oninit_std(const char *fname, const char *tablename, const int batchsize,
               void **handle, const void *logfile, const int fieldsnum,
               const char *insstmt2,
               onlogerror_lpt onlogerror, const int verbosity,
               const char *datahandleoptions);
int onbatch_std(const char *fname, const char *tablename, const void *handle,
              const unsigned long int rownum, const unsigned long int loaded,
              const unsigned long int inserted,
              const unsigned long int updated,
              const unsigned long int deleted,
              const void *logfile,
              onlogerror_lpt onlogerror, const int verbosity);
int onfinish_std(const char *fname, const char *tablename, const void *handle,
                 const unsigned long int rownum,
                 const unsigned long int rownum2,
                 const unsigned long int inserted,
                 const unsigned long int updated,
                 const unsigned long int deleted,
                 const void *logfile,
                 onlogerror_lpt onlogerror, const int verbosity);
int onreadfield_std(const void *fd, const int sqltype, const int4 sqlxid,
                    const char *filename,
                    char **dst, const unsigned int maxdstbuf, const char delim,
                    const char escapech, int * eol, int * sqlind,
                    const unsigned long int rownum,
                    const unsigned long int loaded, const int fieldnum,
                    const char * fieldname, const int max_fld_name,
                    char *fieldname2, const void *logfile,
                    onlogerror_lpt onlogerror, const int verbosity,
                    const char *datahandleoptions);
void oncleanup_std(const char *fname, void *handle, const void *logfile,
                   onlogerror_lpt onlogerror, const int verbosity,
                   const char *datahandleoptions);
int onlogerror_std(const char * srcfile, const int pline, const char *errstr,
                   const long int sqlcode, const long int isamcode,
                   const char *ifx_str, const unsigned long int rownum,
                   const unsigned long int loaded, const int os_errno,
                   const void *logfile, const int verbosity, const int urgency);
int oncrlf_std(const void *fd, const char *filename,
               const unsigned long int rownum, const unsigned long int loaded,
               const int fields_number, const int bequiet, const char escapech,
               const void *logfile,
               onlogerror_lpt onlogerror, const int verbosity,
               const char *datahandleoptions);
int onrecordbegin_std(const void *fd, const char *filename,
                      const unsigned long int rownum,
		      const unsigned long int loaded,
                      const int fields_number, const int bequiet,
                      const char escapech, const void *logfile,
                      onlogerror_lpt onlogerror, const int verbosity,
                      const char *datahandleoptions);
int is_critical_std(const long int sqlcode, const long int isamcode,
                    const int errtype, const int errcnt, const int maxerrors,
                    int *errinc);
int onblobfree_std(loc_t *simple_lo, const void *logfile,
                   onlogerror_lpt onlogerror, int verbose);

static void blob_free(loader_handle_t *loader_handle);

static int convert_ins(loader_handle_t *loader_handle)
{
  loader_handle_t *h;
  char *p1, *p2, *p3;
  int len, param_cnt, type1, i;
  struct sqlda *tdesc; 
  EXEC SQL BEGIN DECLARE SECTION;
    char select_str [(sizeof("select * from ")+(TBLNAMEFULL_SIZE)+1)] ; 
  EXEC SQL END DECLARE SECTION;

  param_cnt=0;
  EXEC SQL WHENEVER ERROR CONTINUE;
  h=loader_handle;
/* insert statement in insstmt2 have the form: 'insert into tabname' (type 1)
 * or 'insert into tabname (columnlist)' (type 2)
 */
  if (h->insstmt2 == NULL) {
    snprintf(h->errstr, ERR_BUF_SIZE, "statement cannot be empty");
    h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode, sqlca.sqlerrd[1],
               sqlca.sqlerrm, 0, 0, 0, h->logfile, h->verbosity, 10);
    return 1;
  }
  p1=strrchr(h->insstmt2,')');
  len=strlen(h->insstmt2);
  type1=1;
  if (p1 != NULL) {
    p2=strrchr(h->insstmt2,'(');
    if (p2 != NULL && p2<p1) {
      type1=2;
      p3=strchr(p2,',');
      param_cnt=1;
      while (p3 != NULL && p3<p1) {
        param_cnt++;
        p3=strchr(p3+1,',');
      }
    } else {
      if (p2 == NULL) {
        sprintf(h->errstr, "Missing '('");
        h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                      sqlca.sqlerrd[1], sqlca.sqlerrm, 0, 0, 0, h->logfile,
                      h->verbosity, 10);
        return 1;
      } else {
        sprintf(h->errstr, "')'should be after '('");
        h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                      sqlca.sqlerrd[1], sqlca.sqlerrm, 0, 0, 0, h->logfile,
                      h->verbosity, 10);
        return 1;
      }
    }
  }
  if (type1==1) {
    p1=(char *) load2_malloc(len+1, __FILE__, __LINE__);
    if (p1 == NULL) {
      sprintf(h->errstr,"Out of memory");
      h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                    sqlca.sqlerrd[1], sqlca.sqlerrm, 0, 0, 0, h->logfile,
                    h->verbosity, 10);
      return 1;
    }
    strcpy(p1, h->insstmt2);
    p2=p1;
    while (*p2) {
      (*p2)=tolower(*p2);
      p2++;
    }
    p2=strstr(p1,"insert");
    p3=strstr(p1,"into");
    if (p2 != NULL && p3 != NULL && p2<p3) {
      p2 = (char *) load2_malloc(len - (p3+5-p1) +1, __FILE__, __LINE__);
      if (p2 == NULL) {
        sprintf(h->errstr, "Out of memory");
        h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                      sqlca.sqlerrd[1],sqlca.sqlerrm, 0, 0, 0, h->logfile,
                      h->verbosity, 10);
        load2_free(p1, __FILE__, __LINE__);
        return 1;
      }
      strcpy(p2,p3-p1+5+h->insstmt2); /* now p2 contains tablename */
      i=strlen(p2)-1;
      while (p2[i]==' ' || p2[i]=='\n' || p2[i]=='\t' || p2[i]=='\r') {
        p2[i]=0;
        i--;
      }
      strcpy(h->tablename,p2);
      snprintf(select_str, ERR_BUF_SIZE, "select * from %s", p2);
      load2_free(p2, __FILE__, __LINE__);
    } else {
      if (p2==NULL) {
        sprintf(h->errstr,"No 'insert' token");
        h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                      sqlca.sqlerrd[1], sqlca.sqlerrm, 0, 0, 0, h->logfile,
                      h->verbosity, 10);
        load2_free(p1, __FILE__, __LINE__);
        return 1;
      }
      if (p3==NULL) {
        sprintf(h->errstr,"No 'into' token");
        h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                      sqlca.sqlerrd[1], sqlca.sqlerrm, 0, 0, 0, h->logfile,
                      h->verbosity, 10);
        load2_free(p1, __FILE__, __LINE__);
        return 1;
      }
      if (p2>=p3) {
        sprintf(h->errstr,"'into' should be after 'insert'");
        h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                      sqlca.sqlerrd[1], sqlca.sqlerrm, 0, 0, 0, h->logfile,
                      h->verbosity, 10);
        load2_free(p1, __FILE__, __LINE__);
        return 1;
      }
    }
    load2_free(p1, __FILE__, __LINE__);
#ifdef DEBUG_ALLOC
printf("prepare tsqlobj\n");
#endif
    EXEC SQL prepare tsqlobj from :select_str;
    if ( sqlca.sqlcode != 0 ) {
      snprintf(h->errstr, ERR_BUF_SIZE, "Cannot prepare temporary select: "INT4_FORMAT"",
              sqlca.sqlcode);
      h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                    sqlca.sqlerrd[1], sqlca.sqlerrm, 0, 0, 0, h->logfile,
                    h->verbosity, 10);
      return 1;
    }
    EXEC SQL describe tsqlobj into tdesc;
#ifdef DEBUG_ALLOC
printf("described: %x\n", tdesc);
#endif
    if ( sqlca.sqlcode != 0 ) {
      snprintf(h->errstr, ERR_BUF_SIZE, "Cannot describe temporary select: "INT4_FORMAT"",
              sqlca.sqlcode);
      h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                    sqlca.sqlerrd[1], sqlca.sqlerrm, 0, 0, 0, h->logfile,
                    h->verbosity, 10);
#ifdef DEBUG_ALLOC
printf("Unprepare tsqlobj\n");
#endif
#ifndef BROKEN_FREE
       EXEC SQL free tsqlobj;
#endif
      return 1;
    }
    param_cnt=tdesc->sqld;
#ifdef DEBUG_ALLOC
printf("Unprepare tsqlobj\n");
#endif
#ifndef BROKEN_FREE
     EXEC SQL free tsqlobj;
#endif
#ifdef DEBUG_ALLOC
printf("After unprepare tsqlobj\n");
#endif

  } else {
    p1=(char *) load2_malloc(len+1, __FILE__, __LINE__);
    if (p1 == NULL) {
      sprintf(h->errstr,"Out of memory");
      h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                    sqlca.sqlerrd[1], sqlca.sqlerrm, 0, 0, 0, h->logfile,
                    h->verbosity, 10);
      return 1;
    }
    strcpy(p1, h->insstmt2);
    p2=p1;
    while (*p2) {
      (*p2)=tolower(*p2);
      p2++;
    }
    i=strlen(p2)-1;
    while ((p2[i]==' ' || p2[i]=='\n' || p2[i]=='\t' || p2[i]=='\r') && p2>p1 ) {
      p2[i]=0;
      i--;
    }

    p2=strstr(p1,"insert");
    p3=strstr(p1,"into");
    if (p2 != NULL && p3 != NULL && p2<p3) {
      strcpy(h->tablename,p3-p1+5+h->insstmt2);
      p3=strchr(h->tablename,'(');
      if (p3!=NULL) {
        p3[0]=0;
      }
      i=strlen(p2);
      p2=h->tablename+i;
      i--;
      while ((p2[i]==' ' || p2[i]=='\n' || p2[i]=='\t' || p2[i]=='\r') && p2>h->tablename ) {
        p2[i]=0;
        i--;
      }
    } else {
      h->tablename[0]='\0';
    }
    load2_free(p1, __FILE__, __LINE__);
  }
  p1=(char *) load2_malloc(len+1+2*param_cnt-1+2+sizeof(" values"), __FILE__, __LINE__);
  if (p1 == NULL) {
    sprintf(h->errstr,"Out of memory");
    h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode, sqlca.sqlerrd[1],
               sqlca.sqlerrm, 0, 0, 0, h->logfile, h->verbosity, 10);
    return 1;
  }
  strcpy(p1,h->insstmt2);
  strcat(p1," values(?");
  for(i=0;i<param_cnt-1;i++) {
    strcat(p1,",?");
  }
  strcat(p1,")");
  if (h->insstmt!=NULL) {
    load2_free(h->insstmt, __FILE__, __LINE__);
  }
  h->insstmt=p1;
  return 0;
}

static void cleanup(loader_handle_t *loader_handle)
{
  struct sqlvar_struct *col;
  int i;
  loader_handle_t *h;
  EXEC SQL BEGIN DECLARE SECTION;
    char stmt_id[128];
    char curs_id[128];
    char stmt_idu[128];
    char stmt_ids[128];
    char stmt_idd[128];
    char curs_ids[128];
  EXEC SQL END DECLARE SECTION;

  h=loader_handle;
  sprintf(stmt_id,"lstmt_id%ld",h->instance_id);

  if (h->iscursoropen) {
    sprintf(curs_id,"lcurs_id%ld",h->instance_id);

    sprintf(h->errstr,"Closing cursor %s...", curs_id);
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, 0, 0, 0, 0, h->logfile,
                  h->verbosity, 50);
    EXEC SQL close :curs_id;
    if ( sqlca.sqlcode != 0 ) {
      sprintf(h->errstr,"Cannot close insert cursor: "INT4_FORMAT"",sqlca.sqlcode);
      h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                    sqlca.sqlerrd[1],sqlca.sqlerrm, 0, 0, 0, h->logfile,
                    h->verbosity, 10);
      return;
    }
    h->iscursoropen=0;
    sprintf(h->errstr,"Freeing cursor %s...", curs_id);
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, 0, 0, 0, 0, h->logfile,
                  h->verbosity, 50);
#ifndef BROKEN_FREE
    EXEC SQL free :curs_id;
    if ( sqlca.sqlcode != 0 ) {
      sprintf(h->errstr,"Cannot free insert cursor: %ld",sqlca.sqlcode);
      h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                    sqlca.sqlerrd[1],sqlca.sqlerrm, 0, 0, 0, h->logfile,
                    h->verbosity, 10);
      return;
    }
#endif
  }
  sprintf(h->errstr,"Freeing statement %s...", stmt_id);
  h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, 0, 0, 0, 0, h->logfile,
                h->verbosity, 50);
#ifndef BROKEN_FREE
  EXEC SQL free :stmt_id;
  if ( sqlca.sqlcode != 0 ) {
    sprintf(h->errstr,"Cannot free insert statement: %ld",sqlca.sqlcode);
    h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                  sqlca.sqlerrd[1],sqlca.sqlerrm, 0, 0, 0, h->logfile,
                  h->verbosity, 10);
    return;
  }
#endif
  if (h->uploadmode && h->uploadmode2) {
    sprintf(curs_ids, "lcurs_ids%ld", h->instance_id);
    sprintf(stmt_ids, "lstmt_ids%ld", h->instance_id);

    sprintf(h->errstr,"Freeing cursor %s...", curs_ids);
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, 0, 0, 0, 0, h->logfile,
                  h->verbosity, 50);

#ifndef BROKEN_FREE
    EXEC SQL free :curs_ids;
    if ( sqlca.sqlcode != 0 ) {
      sprintf(h->errstr,"Cannot free select cursor: %ld",sqlca.sqlcode);
      h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                    sqlca.sqlerrd[1],sqlca.sqlerrm, 0, 0, 0, h->logfile,
                    h->verbosity, 10);
      return;
    }
#endif
    sprintf(h->errstr,"Freeing statement %s...", stmt_ids);
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, 0, 0, 0, 0, h->logfile,
                  h->verbosity, 50);
#ifndef BROKEN_FREE
    EXEC SQL free :stmt_ids;
    if ( sqlca.sqlcode != 0 ) {
      sprintf(h->errstr,"Cannot free select stmt: %ld",sqlca.sqlcode);
      h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                    sqlca.sqlerrd[1],sqlca.sqlerrm, 0, 0, 0, h->logfile,
                    h->verbosity, 10);
      return;
    }
#endif
  }
  if (h->uploadmode && h->noupdates==0) {
    sprintf(stmt_idu, "lstmt_idu%ld", h->instance_id);

    sprintf(h->errstr,"Freeing statement %s...", stmt_idu);
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, 0, 0, 0, 0, h->logfile,
                  h->verbosity, 50);
#ifndef BROKEN_FREE
    EXEC SQL free :stmt_idu;
    if ( sqlca.sqlcode != 0 ) {
      sprintf(h->errstr,"Cannot free update statement: %ld",sqlca.sqlcode);
      h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                    sqlca.sqlerrd[1],sqlca.sqlerrm, 0, 0, 0, h->logfile,
                    h->verbosity, 10);
      return;
    }
#endif
  }
  if (h->uploadmode==4) {
    sprintf(stmt_idd, "lstmt_idd%ld", h->instance_id);

    sprintf(h->errstr,"Freeing statement %s...", stmt_idd);
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, 0, 0, 0, 0, h->logfile,
                  h->verbosity, 50);
#ifndef BROKEN_FREE
    EXEC SQL free :stmt_idd;
    if ( sqlca.sqlcode != 0 ) {
      sprintf(h->errstr,"Cannot free delete stmt: %ld",sqlca.sqlcode);
      h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                    sqlca.sqlerrd[1],sqlca.sqlerrm, 0, 0, 0, h->logfile,
                    h->verbosity, 10);
      return;
    }
#endif
  }
  if (h->isinsidetrans) {

    sprintf(h->errstr,"Rollbacking transaction...");
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, 0, 0, 0, 0, h->logfile,
                  h->verbosity, 50);
    EXEC SQL rollback work;
    if ( sqlca.sqlcode != 0 ) {
      sprintf(h->errstr,"Cannot rollback work: "INT4_FORMAT"",sqlca.sqlcode);
      h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                    sqlca.sqlerrd[1],sqlca.sqlerrm, 0, 0, 0, h->logfile,
                    h->verbosity, 10);
      return;
    }
    h->isinsidetrans=0;
  }
  if (h->used_tags!=NULL) {
    sprintf(h->errstr,"Freeing used_tags...");
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, 0, 0, 0, 0, h->logfile,
                  h->verbosity, 50);
    load2_free(h->used_tags, __FILE__, __LINE__);
    h->used_tags=NULL;
  }
  if (h->oncleanup != NULL) {
    sprintf(h->errstr, "before oncleanup");
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, 0, 0, 0, 0, h->logfile,
                  h->verbosity, 50);
    h->oncleanup(h->fname, h->datahandle, h->logfile, h->onlogerror,
                 h->verbosity, h->datahandleoptions);
    sprintf(h->errstr, "after oncleanup");
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, 0, 0, 0, 0, h->logfile,
                  h->verbosity, 50);
  }
  if (h->lsdesc!=NULL) {
    for (col = h->lsdesc->sqlvar, i = 0; i < h->lsdesc->sqld; col++, i++) {
      col->sqlind = NULL;
    }
    sprintf(h->errstr,"Freeing lsdesc->sqlvar...");
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, 0, 0, 0, 0, h->logfile,
                  h->verbosity, 50);
    load2_free(h->lsdesc->sqlvar, __FILE__, __LINE__);
    h->lsdesc->sqlvar=NULL;

    sprintf(h->errstr,"Freeing lsdesc...");
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, 0, 0, 0, 0, h->logfile,
                  h->verbosity, 50);

#ifdef SQLDA_FREE
#ifdef DEBUG_ALLOC
printf("Do SqlFreeMem: %x\n", h->lsdesc);
#endif
    SqlFreeMem(h->lsdesc, SQLDA_FREE);
#else
    load2_free(h->lsdesc, __FILE__, __LINE__);
#endif
    h->lsdesc=NULL;
  }

  if (h->lsdesc2!=NULL) {
    sprintf(h->errstr,"Freeing lsdesc2...");
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, 0, 0, 0, 0, h->logfile,
                  h->verbosity, 50);
#ifdef SQLDA_FREE
#ifdef DEBUG_ALLOC
printf("Do SqlFreeMem: %x\n", h->lsdesc2);
#endif
    SqlFreeMem(h->lsdesc2, SQLDA_FREE);
#else
    load2_free(h->lsdesc2, __FILE__, __LINE__);
#endif
    h->lsdesc2=NULL;
  }

  if (h->ludesc!=NULL) {
    sprintf(h->errstr,"Freeing ludesc->sqlvar...");
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, 0, 0, 0, 0, h->logfile,
                  h->verbosity, 50);
    load2_free(h->ludesc->sqlvar, __FILE__, __LINE__);
    h->ludesc->sqlvar=NULL;
    sprintf(h->errstr,"Freeing ludesc...");
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, 0, 0, 0, 0, h->logfile,
                  h->verbosity, 50);
#ifdef SQLDA_FREE
#ifdef DEBUG_ALLOC
printf("Do SqlFreeMem: %x\n", h->ludesc);
#endif
    SqlFreeMem(h->ludesc, SQLDA_FREE);
#else
    load2_free(h->ludesc, __FILE__, __LINE__);
#endif
    h->ludesc=NULL;
  }
  if (h->lddesc!=NULL) {
    sprintf(h->errstr,"Freeing lddesc->sqlvar...");
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, 0, 0, 0, 0, h->logfile,
                  h->verbosity, 50);
    load2_free(h->lddesc->sqlvar, __FILE__, __LINE__);
    h->lddesc->sqlvar=NULL;
    sprintf(h->errstr,"Freeing lddesc...");
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, 0, 0, 0, 0, h->logfile,
                  h->verbosity, 50);
#ifdef SQLDA_FREE
#ifdef DEBUG_ALLOC
printf("Do SqlFreeMem: %x\n", h->lddesc);
#endif
    SqlFreeMem(h->lddesc, SQLDA_FREE);
#else
    load2_free(h->lddesc, __FILE__, __LINE__);
#endif
    h->lddesc=NULL;
  }

  for (col = h->ldesc->sqlvar, i = 0; i < h->ldesc->sqld; col++, i++) {
    if (col->sqltype==CFIXBINTYPE && (col->sqlxid==XID_BLOB || col->sqlxid==XID_CLOB )) {
      load2_free(col->sqldata, __FILE__, __LINE__);
    }
  }
  if (h->ldesc!=NULL) {
    sprintf(h->errstr,"Resetting sqlind...");
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, 0, 0, 0, 0, h->logfile,
                  h->verbosity, 50);
    for (col = h->ldesc->sqlvar, i = 0; i < h->ldesc->sqld; col++, i++) {
      col->sqlind = NULL;
    }
    sprintf(h->errstr,"Freeing ldesc...");
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, 0, 0, 0, 0, h->logfile,
                  h->verbosity, 50);
#ifdef SQLDA_FREE
#ifdef DEBUG_ALLOC
printf("Do SqlFreeMem: %x\n", h->ldesc);
#endif
     SqlFreeMem(h->ldesc, SQLDA_FREE);
#else
    load2_free(h->ldesc, __FILE__, __LINE__);
#endif
    h->ldesc=NULL;
  }
  if (h->recordbuffer!=NULL) {
    sprintf(h->errstr,"Freeing recordbuffer...");
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, 0, 0, 0, 0, h->logfile,
                  h->verbosity, 50);
    load2_free(h->recordbuffer, __FILE__, __LINE__);
    h->recordbuffer=NULL;
  }
  if (h->updstmt!=NULL) {
    sprintf(h->errstr,"Freeing updstmt...");
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, 0, 0, 0, 0, h->logfile,
                  h->verbosity, 50);
    load2_free (h->updstmt, __FILE__, __LINE__);
    h->updstmt=NULL;
  }
  if (h->keycols!=NULL) {
    sprintf(h->errstr,"Freeing keycols...");
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, 0, 0, 0, 0, h->logfile,
                  h->verbosity, 50);
    load2_free(h->keycols, __FILE__, __LINE__);
    h->keycols=NULL;
  }
  if (h->nullables!=NULL) {
    sprintf(h->errstr,"Freeing nullables...");
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, 0, 0, 0, 0, h->logfile,
                  h->verbosity, 50);
    load2_free(h->nullables, __FILE__, __LINE__);
    h->nullables=NULL;
  }
  if (h->indicators!=NULL) {
    sprintf(h->errstr,"Freeing indicators...");
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, 0, 0, 0, 0, h->logfile,
                  h->verbosity, 50);
    load2_free(h->indicators, __FILE__, __LINE__);
    h->indicators=NULL;
  }
  if (h->insstmt!=NULL) {
    sprintf(h->errstr,"Freeing insstmt %s...", h->insstmt);
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, 0, 0, 0, 0, h->logfile,
                  h->verbosity, 50);
    load2_free (h->insstmt, __FILE__, __LINE__);
    h->insstmt=NULL;
  }
  if (h->keycols2!=NULL) {
    sprintf(h->errstr,"Freeing keycols...");
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, 0, 0, 0, 0, h->logfile,
                  h->verbosity, 50);
    load2_free(h->keycols2, __FILE__, __LINE__);
    h->keycols2=NULL;
  }
  if (h->fname!=NULL) {
    sprintf(h->errstr,"Freeing fname...");
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, 0, 0, 0, 0, h->logfile,
                  h->verbosity, 50);
    load2_free (h->fname, __FILE__, __LINE__);
    h->fname=NULL;
  }
  if (h->insstmt2!=NULL) {
    sprintf(h->errstr,"Freeing insstmt2...");
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, 0, 0, 0, 0, h->logfile,
                  h->verbosity, 50);
    load2_free (h->insstmt2, __FILE__, __LINE__);
    h->insstmt2=NULL;
  }
  if (h->delstmt!=NULL) {
    sprintf(h->errstr,"Freeing delstmt...");
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, 0, 0, 0, 0, h->logfile,
                  h->verbosity, 50);
    load2_free (h->delstmt, __FILE__, __LINE__);
    h->delstmt=NULL;
  }
  if (h->selstmt!=NULL) {
    sprintf(h->errstr,"Freeing selstmt...");
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, 0, 0, 0, 0, h->logfile,
                  h->verbosity, 50);
    load2_free (h->selstmt, __FILE__, __LINE__);
    h->selstmt=NULL;
  }
}

static void extract_tablename(const char *qstr, char *tablename)
{ char *p1, *p2, *p3;
  int i, len;

  len=strlen(qstr);
  p1=(char *) load2_malloc(len+1, __FILE__, __LINE__);
  if (p1 == NULL) {
    tablename[0]=0;
    return;
  }
  strcpy(p1, qstr);
  p2=p1;
  while (*p2) {
    (*p2)=tolower(*p2);
    p2++;
  }
  p2=strstr(p1,"insert");
  p3=strstr(p1,"into");
  if (p2 != NULL && p3 != NULL && p2<p3) {
    strcpy(p2,p3-p1+5+qstr);
  } else {
    tablename[0]='\0';
    return;
  }
  strcpy(p1,p2);
  p2=strstr(p1,"(");
  p3=strstr(p1,"values");
  if (p2==NULL) {
    if (p3!=NULL) {
      p2=p3;
    } else {
      p2=NULL;
    }
  } else {
    if (p3!=NULL) {
      if (p2<p3) {
        /* p2=p2; */
      } else {
        p2=p3;
      }
    } else {
      /* p2=p2; */
    }
  }
  if (p2!=NULL) {
    p2[0]=0;
  }
  i=strlen(p1)-1;
  while (p1[i]==' ' || p1[i]=='\n' || p1[i]=='\t' || p1[i]=='\r') {
    p1[i]=0;
    i--;
  }
  strcpy(tablename,p1);
  load2_free(p1, __FILE__, __LINE__);
}

static int field_lookup(const loader_handle_t *loader_handle,
                        const char *fieldname)
{
  const loader_handle_t *h;
  struct sqlvar_struct *col2;
  int i;
  int fieldnum;

  h=loader_handle;
  fieldnum=-1;
  for (col2 = h->ldesc->sqlvar, i = 0; i < h->fieldsnum; col2++, i++) {
    if (stricmp(fieldname, col2->sqlname) == 0) {
      fieldnum=i;
      break;
    }
  }
  return fieldnum;
}


loader_handle_t *loader_create (void)
{
 
  loader_handle_t *loader_handle;

  loader_handle=(loader_handle_t *)load2_malloc(sizeof(loader_handle_t), __FILE__, __LINE__);
  if ( loader_handle == NULL ) {
    return NULL;
  }
  loader_handle->insstmt2=NULL;
  loader_handle->insstmt=NULL;
  loader_handle->updstmt=NULL;
  loader_handle->delstmt=NULL;
  loader_handle->selstmt=NULL;
  loader_handle->fname=NULL;
  loader_handle->delim='|';
  loader_handle->batchsize=0;
  loader_handle->usecursor=0;
  loader_handle->usecommit=0;
  loader_handle->maxerrors=0;
  loader_handle->verbosity=30;
  loader_handle->loaded_rows=0;
  loader_handle->inserted_rows=0;
  loader_handle->updated_rows=0;
  loader_handle->deleted_rows=0;
  loader_handle->inserted_rows0=0;
  loader_handle->updated_rows0=0;
  loader_handle->deleted_rows0=0;
  
  loader_handle->logfile=NULL;
  loader_handle->errcnt=0;
  loader_handle->errstr[0]=0;
  loader_handle->tablename[0]=0;
  loader_handle->fieldsnum=0;
  loader_handle->unload_type=0;
  loader_handle->extra_fields=0;
  loader_handle->escapechar='\\';
  loader_handle->datahandle=NULL;
  
  loader_handle->onlogerror=onlogerror_std;
  loader_handle->oninit=&oninit_std;
  loader_handle->onbatch=&onbatch_std;
  loader_handle->onfinish=&onfinish_std;
  loader_handle->onreadfield=&onreadfield_std;
  loader_handle->oncleanup=&oncleanup_std;
  loader_handle->oncrlf=&oncrlf_std;
  loader_handle->onrecordbegin=&onrecordbegin_std;
  loader_handle->is_critical=&is_critical_std;
  loader_handle->onblobfree=&onblobfree_std;
  loader_handle->onrow=NULL;

  loader_handle->uploadmode=0; /* no upload */
  loader_handle->uploadmode2=0; /* no select for upload */
  loader_handle->keycols2=NULL;
  loader_handle->indicators=NULL;
  loader_handle->nullables=NULL;
  loader_handle->keycols=NULL;
  loader_handle->used_tags=NULL;
  loader_handle->ldesc=NULL;
  loader_handle->ludesc=NULL;
  loader_handle->lddesc=NULL;
  loader_handle->lsdesc=NULL;
  loader_handle->lsdesc2=NULL;
  loader_handle->uploadcmd=0;
  loader_handle->recordbuffer=NULL;
  loader_handle->rownum=0;
  loader_handle->rownum2=0;
  loader_handle->rownum2_old=0;
  loader_handle->iscursoropen=0;
  loader_handle->isinsidetrans=0;
  loader_handle->instance_id=instance_id;
  loader_handle->blobs=0;
  loader_handle->noupdates=0;
  instance_id++;
   
  return loader_handle;
}

int loader_setstmt(loader_handle_t *loader_handle, const char *insstmt2)
{
  loader_handle_t *h;
  int l;

  h=loader_handle;
  if (h->insstmt!=NULL) {
    load2_free(h->insstmt, __FILE__, __LINE__);
    h->insstmt=NULL;
  }
  if (h->insstmt2!=NULL) {
    load2_free(h->insstmt2, __FILE__, __LINE__);
    h->insstmt2=NULL;
  }
  l=strlen(insstmt2)+1;
  h->insstmt=load2_malloc(l, __FILE__, __LINE__);
  if (h->insstmt==NULL) {
    snprintf(h->errstr, ERR_BUF_SIZE, "No memory for statement string %s",
            insstmt2);
    h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
        sqlca.sqlerrd[1], sqlca.sqlerrm, 0, 0, 0, h->logfile, h->verbosity, 10);
    cleanup(h);
    return 1;
  }
  h->insstmt2=load2_malloc(l, __FILE__, __LINE__);
  if (h->insstmt2==NULL) {
    load2_free(h->insstmt, __FILE__, __LINE__);
    h->insstmt=NULL;
    snprintf(h->errstr, ERR_BUF_SIZE, "No memory for statement string %s",
            insstmt2);
    h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
        sqlca.sqlerrd[1], sqlca.sqlerrm, 0, 0, 0, h->logfile, h->verbosity, 10);
    cleanup(h);
    return 1;
  }
  strcpy(h->insstmt2, insstmt2);
  strcpy(h->insstmt, insstmt2);
  return 0;
}

int loader_setdatahandlename(loader_handle_t *loader_handle, const char *fname)
{
  loader_handle_t *h;
  int l;

  h=loader_handle;
  if (h->fname!=NULL) {
    load2_free(h->fname, __FILE__, __LINE__);
    h->fname=NULL;
  }
  l=strlen(fname)+1;
  h->fname=load2_malloc(l, __FILE__, __LINE__);
  if (h->fname==NULL) {
    snprintf(h->errstr, ERR_BUF_SIZE, "No memory for filename string %s",
            fname);
    h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
        sqlca.sqlerrd[1], sqlca.sqlerrm, 0, 0, 0, h->logfile, h->verbosity, 10);
    cleanup(h);
    return 1;
  }
  strcpy(h->fname,fname);
  return 0;
}

void loader_setdelimiter(loader_handle_t *loader_handle, const char delim)
{loader_handle->delim=delim; }

void loader_setbatchsize(loader_handle_t *loader_handle, const int batchsize)
{loader_handle->batchsize=batchsize; }

void loader_setusecursor(loader_handle_t *loader_handle, const int usecursor)
{loader_handle->usecursor=usecursor; }

void loader_setusecommit(loader_handle_t *loader_handle, const int usecommit)
{loader_handle->usecommit=usecommit; }

void loader_setmaxerrors(loader_handle_t *loader_handle, const int maxerrors)
{loader_handle->maxerrors=maxerrors; }

void loader_setverbosity(loader_handle_t *loader_handle, const int verbosity)
{loader_handle->verbosity=verbosity; }

void loader_setloghandle(loader_handle_t *loader_handle, const void *logfile)
{loader_handle->logfile=logfile; }

void loader_setoninit(loader_handle_t *loader_handle,
                      int (*oninit)(const char *fname, const char *tablename,
                                    const int batchsize, void **handle,
                                    const void *logfile, const int fieldsnum,
                                    const char *insstmt2,
                                    onlogerror_lpt onlogerror,
                                    const int verbosity,
                                    const char *datahandleoptions))
{loader_handle->oninit=oninit; }

void loader_setonrow(loader_handle_t *loader_handle,
                     int (*onrow)(const char *fname, const char *tablename,
                                  const void *handle,
                                  const unsigned long int rownum,
                                  const unsigned long int loaded,
                                  const unsigned long int inserted,
                                  const unsigned long int updated,
                                  const unsigned long int deleted))
{loader_handle->onrow=onrow; }

void loader_setonbatch(loader_handle_t *loader_handle,
                       int (*onbatch)(const char *fname, const char *tablename,
                                      const void *handle,
                                      const unsigned long int rownum,
                                      const unsigned long int loaded,
                                      const unsigned long int inserted,
                                      const unsigned long int updated,
                                      const unsigned long int deleted,
                                      const void *logfile,
                                      onlogerror_lpt onlogerror,
                                      const int verbosity))
{loader_handle->onbatch=onbatch; }

void loader_setonfinish(loader_handle_t *loader_handle,
                        int (*onfinish)(const char *fname, const char *tablename,
                                        const void *handle,
                                        const unsigned long int rownum,
                                        const unsigned long int rownum2,
                                        const unsigned long int inserted,
                                        const unsigned long int updated,
                                        const unsigned long int deleted,
                                        const void *logfile,
                                        onlogerror_lpt onlogerror,
                                        const int verbosity))
                        
{loader_handle->onfinish=onfinish; }

void loader_setonreadfield(loader_handle_t *loader_handle,
                           int (*onreadfield)(const void *fd, const int sqltype,
                             const int4 sqlxid,
                             const char *filename, char **dst,
                             const unsigned int maxdstbuf, const char delim,
                             const char escapech, int * eol, int * sqlind,
                             const unsigned long int rownum,
                             const unsigned long int loaded,
                             const int fieldnum, const char *fieldname,
                             const int max_fld_name, char *fieldname2,
                             const void *logfile,
                             onlogerror_lpt onlogerror,
                             const int verbosity,
                             const char *datahandleoptions))
{loader_handle->onreadfield=onreadfield; }

void loader_setoncleanup(loader_handle_t *loader_handle,
                         void (*oncleanup)(const char *fname, void *handle,
                                           const void *logfile,
                                           onlogerror_lpt onlogerror,
                                           const int verbosity,
                                           const char *datahandleoptions))
{loader_handle->oncleanup=oncleanup; }

void loader_setonlogerror(loader_handle_t *loader_handle,
                          onlogerror_lpt onlogerror)
{loader_handle->onlogerror=onlogerror; }

void loader_setoncrlf(loader_handle_t *loader_handle,
                      int (*oncrlf)(const void *fd, const char *filename,
                                    const unsigned long int rownum,
                                    const unsigned long int loaded,
                                    const int fields_number, const int bequiet,
                                    const char escapech, const void *logfile,
                                    onlogerror_lpt onlogerror,
				    const int verbosity,
                                    const char *datahandleoptions))
{loader_handle->oncrlf=oncrlf; }

void loader_setonrecordbegin(loader_handle_t *loader_handle,
                             int (*onrecordbegin)(const void *fd,
                                                  const char *filename,
                                                  const unsigned long int rownum,
                                                  const unsigned long int loaded,
                                                  const int fields_number,
                                                  const int bequiet,
                                                  const char escapech,
                                                  const void *logfile,
                                                  onlogerror_lpt onlogerror,
                                                  const int verbosity,
                                                  const char *datahandleoptions))
{loader_handle->onrecordbegin=onrecordbegin; }

void loader_setiscritical(loader_handle_t *loader_handle,
                          int (*is_critical)(const long int sqlcode,
                                             const long int isamcode,
                                             const int errtype, const int errcnt,
                                             const int maxerors,
                                             int *errinc))
{loader_handle->is_critical=is_critical; }

void loader_setonblobfree(loader_handle_t *loader_handle,
                          int (*onblobfree)(loc_t *simple_lo,
                                            const void *logfile,
                                            onlogerror_lpt onlogerror,
                                            int verbose))
{loader_handle->onblobfree=onblobfree; }

void loader_setunload_type(loader_handle_t *loader_handle, const int unload_type)
{loader_handle->unload_type=unload_type; }

void loader_setextrafields(loader_handle_t *loader_handle, const int extra_fields)
{loader_handle->extra_fields=extra_fields; }

void loader_setescapechar(loader_handle_t *loader_handle, const char escapechar)
{loader_handle->escapechar=escapechar; }



char * loader_getstmt(loader_handle_t *loader_handle)
{return loader_handle->insstmt2; }

char * loader_getdatahandlename(loader_handle_t *loader_handle)
{return loader_handle->fname; }

char loader_getdelimiter(loader_handle_t *loader_handle)
{return loader_handle->delim; }

int loader_getbatchsize(loader_handle_t *loader_handle)
{return loader_handle->batchsize; }

int loader_getusecursor(loader_handle_t *loader_handle)
{return loader_handle->usecursor; }

int loader_getusecommit(loader_handle_t *loader_handle)
{return loader_handle->usecommit; }

int loader_getmaxerrors(loader_handle_t *loader_handle)
{return loader_handle->maxerrors; }

int loader_getverbosity(loader_handle_t *loader_handle)
{return loader_handle->verbosity; }

unsigned long int loader_getloadedrows(loader_handle_t *loader_handle)
{return loader_handle->loaded_rows; }

int loader_geterrcnt(loader_handle_t *loader_handle)
{return loader_handle->errcnt; }

char *loader_geterrstr(loader_handle_t *loader_handle)
{return loader_handle->errstr; }

char * loader_getablename(loader_handle_t *loader_handle)
{return loader_handle->tablename; }

int loader_getfieldsnum(loader_handle_t *loader_handle)
{return loader_handle->fieldsnum; }

int loader_getunload_type(loader_handle_t *loader_handle)
{return loader_handle->unload_type; }

int loader_getextrafields(loader_handle_t *loader_handle)
{return loader_handle->extra_fields; }

char loader_getescapechar(loader_handle_t *loader_handle)
{return loader_handle->escapechar; }


int loader_prepare(loader_handle_t *loader_handle)
{
  EXEC SQL BEGIN DECLARE SECTION;
    int fields_number;
    int i, j, k;
    char * insstmt;
    char * insstmt2;
    struct sqlda *ldesc;
    struct sqlda *lsdesc2;
    char stmt_id[128];
    char stmt_idu[128];
    char stmt_idd[128];
    char stmt_ids[128];
  EXEC SQL END DECLARE SECTION;
  loader_handle_t *h;
  char tablename [(TBLNAMEFULL_SIZE)+1];
#if MI_PTR_SIZE == 64
  mlong pos, cp;
#else
  mint pos, cp;
#endif
  struct sqlvar_struct *col, *col2;
  int len;
  unsigned long memsize;
  ifx_lo_t *slobp;
  char *cp2;
  
  h=loader_handle;

  h->ldesc=NULL;
  ldesc=NULL;
  h->ludesc=NULL;
  h->lddesc=NULL;
  h->indicators=NULL;
  insstmt2=h->insstmt2;
  insstmt=h->insstmt;
  h->recordbuffer=NULL;
  h->nullables=NULL;
  sprintf(stmt_id,"lstmt_id%ld",h->instance_id);
  if (h->uploadmode!=0) {
    sprintf(stmt_idu,"lstmt_idu%ld",h->instance_id);
    sprintf(stmt_idd,"lstmt_idd%ld",h->instance_id);
    sprintf(stmt_ids,"lstmt_ids%ld",h->instance_id);
  }

#ifdef DEBUG_ALLOC
printf("prepare: %s\n", stmt_id);
#endif
  EXEC SQL prepare :stmt_id from :insstmt2;
  if ( sqlca.sqlcode != 0 ) {
    snprintf(h->errstr, ERR_BUF_SIZE, "Cannot prepare: "INT4_FORMAT" from %s",
             sqlca.sqlcode, h->insstmt2);
    loader_handle->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
        sqlca.sqlerrd[1], sqlca.sqlerrm, 0, 0, 0, h->logfile, h->verbosity, 10);
    cleanup(h);
    return 1;
  }
  EXEC SQL describe :stmt_id into ldesc;
#ifdef DEBUG_ALLOC
printf("described: %x\n", ldesc);
#endif
  h->ldesc=ldesc;
  if ( sqlca.sqlcode != SQ_LDINSERT && sqlca.sqlcode != SQ_INSERT ) {
    sprintf(h->errstr,"Not an insert part of a load statement");
    h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode, sqlca.sqlerrd[1],
               sqlca.sqlerrm, 0, 0, 0, h->logfile, h->verbosity, 10);
    cleanup(h);
    return 1;
  }
  /* SQ_LDINSERT?? - read Johnattan's Leffler sources then you'll know */
  if (sqlca.sqlcode == SQ_LDINSERT) {
#ifdef SQLDA_FREE
#ifdef DEBUG_ALLOC
printf("Do SqlFreeMem: %x\n", h->ldesc);
#endif
    SqlFreeMem(ldesc, SQLDA_FREE);
#else
    load2_free(ldesc, __FILE__, __LINE__);
#endif
    ldesc=NULL;
    h->ldesc=NULL;
    if (convert_ins(h)) {
      cleanup(h);
      return 1;
    }
    insstmt=h->insstmt;
#ifdef DEBUG_ALLOC
printf("stmt free: %s\n", stmt_id);
#endif
#ifndef BROKEN_FREE
    EXEC SQL free :stmt_id;
#endif
#ifdef DEBUG_ALLOC
printf("prepare: %s\n", stmt_id);
#endif
     EXEC SQL prepare :stmt_id from :insstmt;
    if ( sqlca.sqlcode != 0 ) {
       snprintf(h->errstr, ERR_BUF_SIZE, "Cannot prepare '%s':"INT4_FORMAT"", insstmt,
                sqlca.sqlcode);
       h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                     sqlca.sqlerrd[1], sqlca.sqlerrm, 0, 0, 0, h->logfile,
                     h->verbosity, 10);
       cleanup(h);
       return 1;
    }
    EXEC SQL describe :stmt_id into ldesc;
#ifdef DEBUG_ALLOC
printf("described: %x\n", ldesc);
#endif
    h->ldesc=ldesc;
    if ( sqlca.sqlcode != SQ_INSERT) {
      sprintf(h->errstr,"Not an insert statement");
      h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                    sqlca.sqlerrd[1], sqlca.sqlerrm, 0, 0, 0, h->logfile,
                    h->verbosity, 10);
      cleanup(h);
      return 1;
    }
    fields_number=h->ldesc->sqld;
    h->fieldsnum=fields_number;
    if ( fields_number <= 0 ) {
      snprintf(h->errstr, ERR_BUF_SIZE, "There is no columns:%s",insstmt);
      h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                    sqlca.sqlerrd[1], sqlca.sqlerrm, 0, 0, 0, h->logfile,
                    h->verbosity, 10);
      cleanup(h); 
      return 1;
    }
  } else { /* SQ_INSERT */
    fields_number=h->ldesc->sqld;
    h->fieldsnum=fields_number;
    if ( fields_number <= 0 ) {
      snprintf(h->errstr, ERR_BUF_SIZE, "There is no columns:%s",insstmt2);
      h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                    sqlca.sqlerrd[1], sqlca.sqlerrm, 0, 0, 0, h->logfile,
                    h->verbosity, 10);
      cleanup(h);
      return 1;
    }
    extract_tablename(insstmt2, tablename);
    strcpy(h->tablename,tablename);
  }

  EXEC SQL allocate descriptor 'desc1' WITH MAX :fields_number;
  if ( sqlca.sqlcode != 0 ) {
    snprintf(h->errstr,  ERR_BUF_SIZE, "Cannot allocate descriptor "INT4_FORMAT"",
            sqlca.sqlcode);
    h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode, sqlca.sqlerrd[1],
               sqlca.sqlerrm, 0, 0, 0, h->logfile, h->verbosity, 10);
    cleanup(h);
    return 1;
  }
                                    
  EXEC SQL describe :stmt_id using sql descriptor 'desc1';

  h->indicators=(int *)load2_malloc((h->ldesc->sqld)*sizeof(int), __FILE__, __LINE__);
  if ( h->indicators == NULL ) {
    sprintf(h->errstr,"Out of memory");
    h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode, sqlca.sqlerrd[1],
               sqlca.sqlerrm, 0, 0, 0, h->logfile, h->verbosity, 10);
    cleanup(h);
    EXEC SQL deallocate descriptor 'desc1';
    return 1;
  }

  for (i=0;i<h->ldesc->sqld;i++) {
    h->indicators[i]=0;
  }

  h->nullables=(short *)load2_malloc((h->ldesc->sqld) * sizeof(short), __FILE__, __LINE__);
  if ( h->nullables == NULL ) {
    sprintf(h->errstr,"Out of memory");
    h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode, sqlca.sqlerrd[1],
               sqlca.sqlerrm, 0, 0, 0, h->logfile, h->verbosity, 10);
    cleanup(h);
    EXEC SQL deallocate descriptor 'desc1';
    return 1;
  }

  if (h->uploadmode!=0) {
    h->keycols=(short *)load2_malloc((h->ldesc->sqld)* sizeof(short), __FILE__, __LINE__);
    if ( h->keycols == NULL ) {
      sprintf(h->errstr,"Out of memory");
      h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode, sqlca.sqlerrd[1],
                 sqlca.sqlerrm, 0, 0, 0, h->logfile, h->verbosity, 10);
      cleanup(h);
      EXEC SQL deallocate descriptor 'desc1';
      return 1;
    }
    if (convert_upd(h)) {
      EXEC SQL deallocate descriptor 'desc1';
      return 1;
    }
    if (h->uploadmode==5) {
      h->noinserts=1;
      h->uploadmode=2;
    }
    if (h->noupdates==0) {
      insstmt2=h->updstmt;
#ifdef DEBUG_ALLOC
printf("prepare: %s\n", stmt_idu);
#endif
      EXEC SQL prepare :stmt_idu from :insstmt2;
      insstmt2=h->insstmt2;
      if ( sqlca.sqlcode != 0 ) {
        snprintf(h->errstr, ERR_BUF_SIZE, "Cannot prepare: "INT4_FORMAT" from %s",sqlca.sqlcode,
                 h->updstmt);
        loader_handle->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                                  sqlca.sqlerrd[1], sqlca.sqlerrm, 0, 0, 0,
                                  h->logfile, h->verbosity, 10);
        cleanup(h);
        EXEC SQL deallocate descriptor 'desc1';
        return 1;
      }
    }
    if (h->uploadmode && h->uploadmode2) {
      if (convert_sel(h)) {
        EXEC SQL deallocate descriptor 'desc1';
        return 1;
      }
      insstmt2=h->selstmt;
#ifdef DEBUG_ALLOC
printf("prepare: %s\n", stmt_ids);
#endif
       EXEC SQL prepare :stmt_ids from :insstmt2;
      insstmt2=h->insstmt2;
      if ( sqlca.sqlcode != 0 ) {
        snprintf(h->errstr, ERR_BUF_SIZE, "Cannot prepare: "INT4_FORMAT" from %s",
                 sqlca.sqlcode, h->selstmt);
        loader_handle->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                                  sqlca.sqlerrd[1], sqlca.sqlerrm, 0, 0, 0,
                                  h->logfile, h->verbosity, 10);
        cleanup(h);
        EXEC SQL deallocate descriptor 'desc1';
        return 1;
      }
    }
    if (h->uploadmode==4) {
      if (convert_del(h)) {
        EXEC SQL deallocate descriptor 'desc1';
        return 1;
      }
      insstmt2=h->delstmt;
#ifdef DEBUG_ALLOC
printf("prepare: %s\n", stmt_idd);
#endif
      EXEC SQL prepare :stmt_idd from :insstmt2;
      insstmt2=h->insstmt2;
      if ( sqlca.sqlcode != 0 ) {
        snprintf(h->errstr, ERR_BUF_SIZE, "Cannot prepare: "INT4_FORMAT" from %s",sqlca.sqlcode,
                 h->delstmt);
        loader_handle->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                                  sqlca.sqlerrd[1], sqlca.sqlerrm, 0, 0, 0,
                                  h->logfile, h->verbosity, 10);
        cleanup(h);
        EXEC SQL deallocate descriptor 'desc1';
        return 1;
      }
    }
  }

/* Step 1: analyze ldesc to determine type and memory requirements
 *         of each item in the select list.  rtypalign() returns a
 *         pointer to the next appropriate boundary (starting at
 *         pos) for the indicated data type.
 * Unhandled types: SQLSET, SQLMULTISET, SQLLIST,
 * SQLROW, SQLCOLLECTION, SQLROWREF, SQLUDTVAR, SQLUDTFIXED, SQLREFSER8
 */
  pos = 0;
  h->blobs=0;
  for (col = h->ldesc->sqlvar, i = 0; i < h->ldesc->sqld; col++, i++) {
    len=stleng(col->sqlname);
    cp2=col->sqlname+len-1;
    while(len>1 && *cp2==' ') {
      *cp2='\0';
      len--, cp2--;
    }
    j=i+1;
    EXEC SQL get descriptor 'desc1' VALUE :j :k = NULLABLE;
    if ( sqlca.sqlcode != 0 ) {
      snprintf(h->errstr, ERR_BUF_SIZE, "Cannot get descriptor "INT4_FORMAT"",
              sqlca.sqlcode);
      h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                    sqlca.sqlerrd[1], sqlca.sqlerrm, 0, 0, 0, h->logfile,
                    h->verbosity, 10);
      cleanup(h);
      EXEC SQL deallocate descriptor 'desc1';
      return 1;
    }
    h->nullables[i]=k; /* 1 the column allow null values, 0 - does not */

    switch(col->sqltype) {
      case SQLSMFLOAT:
        col->sqltype = CFLOATTYPE;
        break;
      case SQLFLOAT:
        col->sqltype = CDOUBLETYPE;
        break;
      case SQLMONEY:
      case SQLDECIMAL:
        col->sqltype = CDECIMALTYPE;
        break;
      case SQLCHAR:
      case SQLNCHAR:
        col->sqltype = CCHARTYPE;
/*        col->sqltype=CFIXCHARTYPE; */
        break;
      case SQLLVARCHAR: 
        col->sqltype = CLVCHARTYPE;
        break;
      case SQLINT:
      case SQLSERIAL:
        col->sqltype = CINTTYPE;
        break;
      case SQLSMINT:
        col->sqltype = CSHORTTYPE;
        break;
      case SQLINT8:
      case SQLSERIAL8:
        col->sqltype = CINT8TYPE;
        break;
      case SQLVCHAR:
      case SQLNVCHAR:
        col->sqltype = CVCHARTYPE;
        break;
      case SQLDATE:
        col->sqltype = CDATETYPE;
        break;
      case SQLDTIME:
        col->sqltype = CDTIMETYPE;
        break;
      case SQLINTERVAL:
        col->sqltype = CINVTYPE;
        break;
      case SQLBOOL:
        col->sqltype = CBOOLTYPE;
        break;
      case SQLTEXT:
      case SQLBYTES:
/*        col->sqltype = CLOCATORTYPE; */
        (h->blobs)++;
	break;
#ifdef SQLUDTFIXED
      case SQLUDTFIXED:
        switch (col->sqlxid) {
          case XID_BLOB:
            break;
          case XID_CLOB:
            break;
          default:
         /*  Usupported data type.  */
                snprintf(h->errstr,ERR_BUF_SIZE, "unknown extended type:%s(idtype="INT4_FORMAT") for column %s\n", col->sqltypename,
                        col->sqlxid, col->sqlname);
                h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, NULL, 0, 0, 0,
                           h->logfile, h->verbosity, 10);
                cleanup(h);
                EXEC SQL deallocate descriptor 'desc1';
                return 1;
        }
        col->sqltype=CFIXBINTYPE;
        break;
#endif
      default:
        /*  Usupported data type.  */
        snprintf(h->errstr, ERR_BUF_SIZE, "unsupported type:%s:%s:%d",
                col->sqlname, rtypname(col->sqltype), col->sqltype);
        h->onlogerror(__FILE__, __LINE__, h->errstr, 0,
                      0, NULL, 0, 0, 0, h->logfile,
                      h->verbosity, 10);
        cleanup(h);
        EXEC SQL deallocate descriptor 'desc1';
        return 1;
    }
    memsize = rtypmsize(col->sqltype,col->sqllen);
    if (col->sqltype!=CFIXBINTYPE) {
      pos = rtypalign( pos, col->sqltype) + memsize;
    }
    col->sqlind = (short *)(h->indicators + i);
  }
  EXEC SQL deallocate descriptor 'desc1';


/* Step 2: Allocate memory to receive a row of the table returned
 *         by the SELECT statement.  The variable pos has an integer
 *         value equal to the number of bytes occupied by the
 *         row.
 */
  if (pos!=0) { /* this can happen when loading only smart BLOBs */
    h->recordbuffer = (char *)load2_malloc((size_t)pos, __FILE__, __LINE__);
    if ( h->recordbuffer == NULL ) {
      sprintf(h->errstr,"Out of memory");
      h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode, sqlca.sqlerrd[1],
                 sqlca.sqlerrm, 0, 0, 0, h->logfile, h->verbosity, 10);
      cleanup(h);
      return 1;
    }
  }

/* Step 3: Set pointers in the allocated memory to receive each
 * item in the select list.
 */
#if MI_PTR_SIZE == 64
  cp = (mlong)h->recordbuffer;
#else
  cp = (mint)h->recordbuffer;
#endif

  for (col = h->ldesc->sqlvar, i = 0; i < h->ldesc->sqld; col++, i++) {
    memsize = rtypmsize(col->sqltype, col->sqllen);
    if (col->sqltype==CFIXBINTYPE) {
      if (col->sqlxid==XID_BLOB || col->sqlxid==XID_CLOB ) {
        col->sqldata=load2_malloc(memsize, __FILE__, __LINE__);
        slobp=(ifx_lo_t *) col->sqldata;
        if ( col->sqldata == NULL ) {
          sprintf(h->errstr,"Out of memory\n");
          h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode, sqlca.sqlerrd[1],
                     sqlca.sqlerrm, h->rownum, 0, 0, h->logfile, h->verbosity, 10);
          cleanup(h);
          return 1;
        }
      }
    } else {
      cp = rtypalign( cp, col->sqltype);
      col->sqldata = (char *)cp;

      cp += memsize;
      if (col->sqltype==CCHARTYPE || col->sqltype==CVCHARTYPE
      || col->sqltype==CLVCHARTYPE /* || col->sqltype==CFIXCHARTYPE */) {
        col->sqllen=memsize;
      }
      if (col->sqltype==CDTIMETYPE) {
        ((dtime_t *)col->sqldata)->dt_qual=(int2)col->sqllen;
      }
      if (col->sqltype==CINVTYPE) {
         ((intrvl_t *)col->sqldata)->in_qual=(int2)col->sqllen;
      }
    }
  }
  if (h->uploadmode!=0) {
    h->ludesc=(struct sqlda *) load2_malloc(sizeof(struct sqlda), __FILE__, __LINE__);
    if (h->ludesc == NULL) {
      sprintf(h->errstr,"Out of memory");
      h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                    sqlca.sqlerrd[1], sqlca.sqlerrm, 0, 0, 0, h->logfile,
                    h->verbosity, 10);
      cleanup(h);
      return 1;
    }
    h->ludesc->sqld = h->ldesc->sqld;
    h->ludesc->sqlvar=(struct sqlvar_struct *) load2_malloc((h->ludesc->sqld)*
                                sizeof(struct sqlvar_struct), __FILE__, __LINE__);
    if (h->ludesc == NULL) {
      sprintf(h->errstr,"Out of memory");
      h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                    sqlca.sqlerrd[1], sqlca.sqlerrm, 0, 0, 0, h->logfile,
                    h->verbosity, 10);
      cleanup(h);
      return 1;
    }
    col2=h->ludesc->sqlvar;
    for (col = h->ldesc->sqlvar, i = 0; i < h->ldesc->sqld; col++, i++) {
      if (!h->keycols[i]) {
        col2[0]=col[0];
        col2++;
      }
    }
    for (col = h->ldesc->sqlvar, i = 0; i < h->ldesc->sqld; col++, i++) {
      if (h->keycols[i]) {
        col2[0]=col[0];
        col2++;
      }
    }
    if (h->uploadmode==4) {
      h->lddesc=(struct sqlda *) load2_malloc(sizeof(struct sqlda), __FILE__, __LINE__);
      if (h->lddesc == NULL) {
        sprintf(h->errstr,"Out of memory");
        h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                      sqlca.sqlerrd[1], sqlca.sqlerrm, 0, 0, 0, h->logfile,
                      h->verbosity, 10);
        cleanup(h);
        return 1;
      }
      j=0;
      for (i = 0; i < h->ldesc->sqld; i++) {
        if (h->keycols[i]) {
          j++;
        }
      }
      h->lddesc->sqld = j;
      h->lddesc->sqlvar=(struct sqlvar_struct *) load2_malloc((h->lddesc->sqld)*
                                  sizeof(struct sqlvar_struct), __FILE__, __LINE__);
      if (h->lddesc == NULL) {
        sprintf(h->errstr,"Out of memory");
        h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                      sqlca.sqlerrd[1], sqlca.sqlerrm, 0, 0, 0, h->logfile,
                      h->verbosity, 10);
        cleanup(h);
        return 1;
      }
      col2=h->lddesc->sqlvar;
      for (col = h->ldesc->sqlvar, i = 0; i < h->ldesc->sqld; col++, i++) {
        if (h->keycols[i]) {
          col2[0]=col[0];
          col2++;
        }
      }
    }
  }
  if (h->uploadmode && h->uploadmode2) {
      h->lsdesc=(struct sqlda *) load2_malloc(sizeof(struct sqlda), __FILE__, __LINE__);
    if (h->lsdesc == NULL) {
      sprintf(h->errstr,"Out of memory");
      h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                    sqlca.sqlerrd[1], sqlca.sqlerrm, 0, 0, 0, h->logfile,
                    h->verbosity, 10);
      cleanup(h);
      return 1;
    }
    j=0;
    for (i = 0; i < h->ldesc->sqld; i++) {
      if (h->keycols[i]) {
        j++;
      }
    }
    h->lsdesc->sqld = j;
    h->lsdesc->sqlvar=(struct sqlvar_struct *)
                         load2_malloc(j * sizeof(struct sqlvar_struct), __FILE__, __LINE__);
    if (h->lsdesc == NULL) {
      sprintf(h->errstr,"Out of memory");
      h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                    sqlca.sqlerrd[1], sqlca.sqlerrm, 0, 0, 0, h->logfile,
                    h->verbosity, 10);
      cleanup(h);
      return 1;
    }
    col2=h->lsdesc->sqlvar;
    for (col = h->ldesc->sqlvar, i = 0; i < h->ldesc->sqld; col++, i++) {
      if (h->keycols[i]) {
        col2[0]=col[0];
        col2++;
      }
    }
    EXEC SQL describe :stmt_ids into lsdesc2;
#ifdef DEBUG_ALLOC
printf("described: %x\n", lsdesc2);
#endif
    h->lsdesc2=lsdesc2;
    if ( sqlca.sqlcode != 0) {
      snprintf(h->errstr, ERR_BUF_SIZE, "Not a select statement %s", h->selstmt);
      h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                    sqlca.sqlerrd[1], sqlca.sqlerrm, 0, 0, 0, h->logfile,
                    h->verbosity, 10);
      cleanup(h);
      return 1;
    }
  }
  return 0;
}

int loader_open(loader_handle_t *loader_handle)
{
  loader_handle_t *h;
  void *h1;
  EXEC SQL BEGIN DECLARE SECTION;
    char stmt_id[128];
    char curs_id[128];
    char stmt_ids[128];
  EXEC SQL END DECLARE SECTION;
  int i;

  h=loader_handle;


  sprintf(stmt_id,"lstmt_id%ld",h->instance_id);
  if (h->usecursor) {
    sprintf(curs_id,"lcurs_id%ld",h->instance_id);
#ifdef DEBUG_ALLOC
printf("declare cursor: %s\n", curs_id);
#endif
     if (h->usecommit) {
      EXEC SQL declare :curs_id cursor with hold for :stmt_id;
    } else {
      EXEC SQL declare :curs_id cursor for :stmt_id;
    }
    if ( sqlca.sqlcode != 0 ) {
      snprintf(h->errstr, ERR_BUF_SIZE, "Cannot declare cursor: "INT4_FORMAT"",
              sqlca.sqlcode);
      h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                    sqlca.sqlerrd[1], sqlca.sqlerrm, 0, 0, 0, h->logfile,
                    h->verbosity, 10);
      cleanup(h);
      return 1;
    }
 
    EXEC SQL open :curs_id;
    if ( sqlca.sqlcode != 0 ) {
      snprintf(h->errstr, ERR_BUF_SIZE, "Cannot open cursor: "INT4_FORMAT":"INT4_FORMAT":%s",
              sqlca.sqlcode,
              sqlca.sqlerrd[1], sqlca.sqlerrm);
      h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                    sqlca.sqlerrd[1], sqlca.sqlerrm, 0, 0, 0, h->logfile,
                    h->verbosity, 10);
      cleanup(h);
      return 1;
    }
    h->iscursoropen=1;
  }
  if (h->uploadmode && h->uploadmode2) {
    sprintf(curs_id, "lcurs_ids%ld", h->instance_id);
    sprintf(stmt_ids, "lstmt_ids%ld", h->instance_id);

#ifdef DEBUG_ALLOC
printf("declare cursor : %s\n", curs_id);
#endif
     EXEC SQL declare :curs_id cursor for :stmt_ids;
    if ( sqlca.sqlcode != 0 ) {
      snprintf(h->errstr, ERR_BUF_SIZE, "Cannot declare select cursor: "INT4_FORMAT"",
              sqlca.sqlcode);
      h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                    sqlca.sqlerrd[1], sqlca.sqlerrm, 0, 0, 0, h->logfile,
                    h->verbosity, 10);
      cleanup(h);
      return 1;
    }
  }
  h1=h->datahandle;
  if (h->oninit(h->fname,h->tablename, h->batchsize, &h1, h->logfile,
    h->fieldsnum, h->insstmt, h->onlogerror, h->verbosity,
    h->datahandleoptions)!=0) {
    h->datahandle=h1;
    cleanup(h);
    return 1;
  }
  h->datahandle=h1;
  if (h->unload_type==3 || h->unload_type==4) {
    h->used_tags=(short *)load2_malloc((h->ldesc->sqld)* sizeof(short), __FILE__, __LINE__);
    if ( h->used_tags == NULL ) {
      sprintf(h->errstr,"Out of memory");
      h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                    sqlca.sqlerrd[1], sqlca.sqlerrm, 0, 0, 0, h->logfile,
                    h->verbosity, 10);
      cleanup(h);
      return 1;
    }
    for (i=0;i<h->fieldsnum;i++) {
      h->used_tags[i]=0;
    }
  }
  if (h->usecommit) {
    if (h->usecommit==2) {
      EXEC SQL begin work without replication;
    } else {
      EXEC SQL begin work;
    }
    if ( sqlca.sqlcode != 0 ) {
      snprintf(h->errstr, ERR_BUF_SIZE, "Cannot begin transaction: "INT4_FORMAT":"INT4_FORMAT":%s",
              sqlca.sqlcode, sqlca.sqlerrd[1], sqlca.sqlerrm);
      h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                    sqlca.sqlerrd[1], sqlca.sqlerrm, 0, 0, 0, h->logfile,
                    h->verbosity, 10);
      cleanup(h);
      return 1;
    }
    loader_handle->isinsidetrans=1;
  }
 
  h->rownum=1;
  h->rownum2=0;
  h->rownum2_old=0;
  h->inserted_rows=0;
  h->updated_rows=0;
  h->deleted_rows=0;
  h->inserted_rows0=0;
  h->updated_rows0=0;
  h->deleted_rows0=0;
  return 0;
}

int loader_readfield(loader_handle_t *loader_handle, int *fieldnum,
                     void *buf, unsigned long int buf_size,
                     unsigned long int *bytesread, int *all_records,
                     char *fieldname, const int maxfieldname, int *emptyread,
                     int *all_fields, int *badrecord, int *sqlind)
{
  loader_handle_t *h;
  char *charp;
  int len;
  struct sqlvar_struct *col;
  int i;
  int errinc;
  int eol;
  int extracolumn;
  unsigned long int cvstat;
  int j;
  int do_delete;

  *emptyread=0;
  *all_fields=0;
  *badrecord=0;
  *bytesread=0;
  *all_records=0;
  fieldname[0]=0;
  h=loader_handle;
  if ((*fieldnum)==-1 && h->uploadmode==4 && h->unload_type!=3
       && h->unload_type!=4) {
    extracolumn=1;
  } else {
    extracolumn=0;
  }
  while (1) {
    charp=buf;
    if (extracolumn) {
      len=h->onreadfield(h->datahandle, CSHORTTYPE, 0, h->fname,
                         &charp, buf_size, h->delim, h->escapechar, &eol, sqlind,
                         h->rownum, h->rownum2_old, 0,
                         (h->ldesc->sqlvar)->sqlname, maxfieldname,
                         fieldname, h->logfile, h->onlogerror, h->verbosity,
                         h->datahandleoptions);
    } else {
      len=h->onreadfield(h->datahandle, (h->ldesc->sqlvar+(*fieldnum))->sqltype,
                         (h->ldesc->sqlvar+(*fieldnum))->sqlxid, h->fname,
                         &charp, buf_size, h->delim, h->escapechar, &eol, sqlind,
                         h->rownum, h->rownum2_old, *fieldnum,
                         (h->ldesc->sqlvar+(*fieldnum))->sqlname, maxfieldname,
                         fieldname, h->logfile, h->onlogerror, h->verbosity,
                         h->datahandleoptions);
    }
    if (len==-1) { /* error while read */
      cleanup(h);
      return 1;
    }
    if (h->unload_type==3 || h->unload_type==4) { /* tag format */
      if ((fieldname && fieldname[0]) || len==-2 || eol) {
        /* len==-2 when EOF, fieldname - field name read from file */
        break;
      }
    } else {
      break;
    }
  }
  *bytesread=len;
  if (len<0) {
    *bytesread=0; /* onreadfield handler have to set flag len==-2
                     only if no field data were read */
  }
  if (len==-2) {
    *all_records=1;
    *emptyread=1;
    return 0;
  }
  if ((h->unload_type==3 || h->unload_type==4) && !eol) {
    i=field_lookup(h, fieldname);
    if (i<0) { /* when field name not found */
      if (stricmp(fieldname,"uploadcmd")==0) {
        extracolumn=1;
      } else {
        snprintf(h->errstr, ERR_BUF_SIZE,
                "Field %s does not exist in insert stmt at line %ld",
                fieldname, h->rownum);
        if (h->extra_fields) {
          *emptyread=1;
          *bytesread=0;
  /*      *eol=*eol; */
          return 0;
        } else {
          if (!h->is_critical(0, 0, 4, h->errcnt, h->maxerrors, &errinc)) {
            h->errcnt+=errinc;
            *emptyread=1;
            *bytesread=0;
  /*        *eol=*eol; */
            h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, NULL, 0, 0, 0,
                          h->logfile, h->verbosity, 20);
            return 0;
          } else {
            h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, NULL, 0, 0, 0,
                          h->logfile, h->verbosity, 10);
            h->loaded_rows=h->rownum2_old;
            cleanup(h);
            return 1;
          }
        }
      }
    } else {
      *fieldnum=i;
      h->used_tags[i]=1;
    }
  } else {
    i=*fieldnum;
  }
  if (extracolumn) {
    cvstat=rstoi(charp,(mint *)&j);
    if (cvstat!=0) {
      snprintf(h->errstr, ERR_BUF_SIZE,
               "Line %ld: not a short integer format for upload cmd column (%ld): %s",
               h->rownum, cvstat, charp);
      h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                 sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum, h->rownum2_old,
                 0, h->logfile, h->verbosity, 10);
      cleanup(h);
      return 1;
    }
    if (h->uploadcmd<0 || h->uploadcmd>2) {
      snprintf(h->errstr, ERR_BUF_SIZE,
               "Line %ld: incorrect value in upload cmd column (%ld): %s",
               h->rownum, cvstat, charp);
      h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                 sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum, h->rownum2_old,
                 0, h->logfile, h->verbosity, 10);
      cleanup(h);
      return 1;
    }
    h->uploadcmd=j;
    *emptyread=1;
    (*fieldnum)++;
    return 0;
  }
  col=((h->ldesc)->sqlvar)+i;

/* 0 - ending delimiter not required, 1 - reqired, 
   2 - ending delimiter not allowed,
   3 - tag format, non-existing field is null
   4 - tag format, non-existing field is an error */
                      
  *all_fields=0;
  switch (h->unload_type) {
    case 0:
    case 1:
    case 2: 
      if (i>=h->ldesc->sqld-1) {
        *all_fields=1;
      }
      break;
    case 3:
    case 4:
      if (eol) {
        *all_fields=1;
      }
      break;
    default:
      snprintf(h->errstr, ERR_BUF_SIZE, "Unknown unload type %d",
              h->unload_type);
      h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, NULL, 0, 0, 0,
                    h->logfile, h->verbosity, 10);
      h->loaded_rows=h->rownum2_old;
      cleanup(h);
      return 1;
  }
  if (eol) {
    if ( ((i<h->fieldsnum-1) && (h->unload_type==0))
    || ((i<h->fieldsnum) && (h->unload_type==1))
    || ((i<h->fieldsnum-1) && (h->unload_type==2))) {
      snprintf(h->errstr, ERR_BUF_SIZE,
              "At line %ld: number of columns %d<expected %d",
              h->rownum,i,h->fieldsnum);
      if (!h->is_critical(0, 0, 4, h->errcnt, h->maxerrors, &errinc)) {
        h->rownum++;
        h->errcnt+=errinc;
        *emptyread=1;
        *bytesread=0;
/*      *eol=*eol; */
        h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                      sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum,
                      h->rownum2_old, 0, h->logfile, h->verbosity, 20);
        return 0;
      } else {
        h->loaded_rows=h->rownum2_old;
        h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                      sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum,
                      h->rownum2_old, 0, h->logfile, h->verbosity, 10);
        cleanup(h);
        return 1;
      }
    }
  }
  do_delete=0;
  if (h->uploadmode==4 && h->uploadcmd==2 && h->keycols[i]==0) {
    do_delete=1;
  }
  if (h->nullables[i]==0 && *sqlind==-1 && !do_delete) {
    snprintf(h->errstr, ERR_BUF_SIZE, "Column %d:%s cannot be null at line %ld",
            i, col -> sqlname, h->rownum);

    if (!h->is_critical(0, 0, 1, h->errcnt, h->maxerrors, &errinc)) {
      h->errcnt+=errinc;
      if(h->oncrlf(h->datahandle, h->fname, h->rownum, h->fieldsnum, h->rownum,
         h->rownum2_old, h->escapechar, h->logfile, h->onlogerror, h->verbosity,
         h->datahandleoptions)==1) {
        h->loaded_rows=h->rownum2_old;
        cleanup(h);
        return 1;
      }
      h->rownum++;
      h->errcnt+=errinc;
      *emptyread=1;
      *bytesread=0;
/*    *eol=*eol; */
      *badrecord=1;
      h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                    sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum, h->rownum2_old,
                    0, h->logfile, h->verbosity, 20);
      return 0;
    } else {
      h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                    sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum, h->rownum2_old,
                    0, h->logfile, h->verbosity, 10);
      cleanup(h);
      return 1;
    }
  }
  if (*all_fields) {
/* 0 - ending delimiter not required,
   1 - reqired, 
   2 - ending delimiter not allowed,
   3 - tag format, non-existing field is null
   4 - tag format, non-existing field is an error */


    if (eol && h->unload_type==1) {
      snprintf(h->errstr, ERR_BUF_SIZE,
              "At line %ld: ending delimiter is required", h->rownum);
      if (!h->is_critical(0, 0, 4, h->errcnt, h->maxerrors, &errinc)) {
        h->rownum++;
        h->errcnt+=errinc;
        *emptyread=1;
        *bytesread=0;
/*      *eol=*eol; */
        *badrecord=1;
        h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                      sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum,
                      h->rownum2_old, 0, h->logfile, h->verbosity, 20);
        return 0;
      } else {
        h->loaded_rows=h->rownum2_old;
        h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                      sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum,
                      h->rownum2_old, 0, h->logfile, h->verbosity, 10);
        cleanup(h);
        return 1;
      }
    }
    if (!eol && h->unload_type==1) {
      if ((!h->extra_fields && h->oncrlf(h->datahandle, h->fname, h->rownum,
           h->rownum2_old, h->fieldsnum, 0, h->escapechar, h->logfile,
           h->onlogerror, h->verbosity, h->datahandleoptions)!=0)
        || (h->extra_fields && h->oncrlf(h->datahandle, h->fname, h->rownum,
            h->rownum2_old, h->fieldsnum, 1, h->escapechar, h->logfile,
            h->onlogerror, h->verbosity, h->datahandleoptions)!=0)) {
        if (!h->is_critical(0, 0, 4, h->errcnt, h->maxerrors, &errinc)) {
          h->rownum++;
          h->errcnt+=errinc;
          *emptyread=1;
          *bytesread=0;
/*        *eol=*eol; */
          *badrecord=1;
          return 0;
        } else {
          h->loaded_rows=h->rownum2_old;
          cleanup(h);
          return 1;
        }
      }
    }
    if (!eol && h->unload_type==2) {
      if (!h->extra_fields) {
        snprintf(h->errstr, ERR_BUF_SIZE,
                "At line %ld: number of columns in file >expected %d",
                h->rownum, h->fieldsnum);
        if (!h->is_critical(0, 0, 4, h->errcnt,h->maxerrors, &errinc)) {
          if (h->oncrlf(h->datahandle, h->fname, h->rownum, h->fieldsnum,
              h->rownum, h->rownum2_old, h->escapechar, h->logfile,
              h->onlogerror, h->verbosity, h->datahandleoptions)==1) {
            h->loaded_rows=h->rownum2_old;
            cleanup(h);
            return 1; 
          }
          h->rownum++;
          h->errcnt+=errinc;
          *emptyread=1;
          *bytesread=0;
/*        *eol=*eol; */
          *badrecord=1;
          h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                        sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum,
                        h->rownum2_old, 0, h->logfile, h->verbosity, 20);
          return 0;
        } else {
          h->loaded_rows=h->rownum2_old;
          h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                        sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum,
                        h->rownum2_old, 0, h->logfile, h->verbosity, 10);
          cleanup(h);
          return 1;
        }

      } else {
        if (h->oncrlf(h->datahandle, h->fname, h->rownum, h->rownum2_old,
            h->fieldsnum, 1, h->escapechar, h->logfile, h->onlogerror,
            h->verbosity, h->datahandleoptions)!=0) {
          h->loaded_rows=h->rownum2_old;
          cleanup(h);
          return 1;
        }
      }
    }
    if (!eol && h->unload_type==0) {
      if ((!h->extra_fields && h->oncrlf(h->datahandle, h->fname, h->rownum,
           h->rownum2_old, h->fieldsnum, 0, h->escapechar, h->logfile,
           h->onlogerror, h->verbosity, h->datahandleoptions)!=0)
        || (h->extra_fields && h->oncrlf(h->datahandle, h->fname, h->rownum,
            h->rownum2_old, h->fieldsnum, 1, h->escapechar, h->logfile,
            h->onlogerror, h->verbosity, h->datahandleoptions)!=0)) {
        if (!h->is_critical(0, 0, 4, h->errcnt, h->maxerrors, &errinc)) {
          h->rownum++;
          h->errcnt+=errinc;
          *emptyread=1;
          *bytesread=0;
/*        *eol=*eol; */
          *badrecord=1;
          return 0;
        } else {
          h->loaded_rows=h->rownum2_old;
          cleanup(h);
          return 1;
        }
      }
    }
  }
  *emptyread=0;
/*  *bytesread=0; */
/* *eol=*eol; */
  *badrecord=0;
  return 0;
}

int loader_beforerow(loader_handle_t *loader_handle)
{
  loader_handle_t *h;

  h=loader_handle;
  if (h->onrecordbegin(h->datahandle, h->fname, h->rownum, h->rownum2_old,
      h->fieldsnum, 0, h->escapechar, h->logfile, h->onlogerror,
      h->verbosity, h->datahandleoptions)!=0) {
    h->loaded_rows=h->rownum2_old;
    cleanup(h);
    return 1;
  }
  return 0;
}

int loader_getfieldnum(loader_handle_t *loader_handle, const char *fieldname)
{ loader_handle_t *h;
  int errinc;
  int i;
  int fieldnum2;
 
/* return value: 0 .. n-1 field number, -1 extra column containing command 
   for loadinc, -2 hard error (cleanup), -3 soft error (continue)
*/
  h=loader_handle;
  i=field_lookup(h, fieldname);
  if (i<0) { /* when field name not found */
    if (stricmp(fieldname,"uploadcmd")==0) {
      return -1;
    } else {
      snprintf(h->errstr, ERR_BUF_SIZE,
               "Field %s does not exist in insert statement at line %ld",
               fieldname, h->rownum);
      if (h->extra_fields ||
          !h->is_critical(0, 0, 4, h->errcnt, h->maxerrors, &errinc)) {
        h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                      sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum,
                      h->rownum2_old, 0, h->logfile, h->verbosity, 20);
        h->errcnt+=errinc;
        return -3;
      } else {
        h->loaded_rows=h->rownum2_old;
        h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                      sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum,
                      h->rownum2_old, 0, h->logfile, h->verbosity, 10);
        cleanup(h);
        return -2;
      }
    }
  }
  fieldnum2=i;
  return fieldnum2;
}

int loader_getfieldtype(loader_handle_t *loader_handle, const int fieldnum)
{ struct sqlvar_struct *col;

  if (fieldnum==-1) {
    return SQLINT;
  }
  if (fieldnum<0 || fieldnum>=loader_handle->ldesc->sqld) {
    return -1;
  }
  col=((loader_handle->ldesc)->sqlvar)+fieldnum;
  return col->sqltype;
}

int loader_getfieldtypeext(loader_handle_t *loader_handle, const int fieldnum)
{ struct sqlvar_struct *col;

  if (fieldnum==-1) {
    return -1;
  }
  if (fieldnum<0 || fieldnum>=loader_handle->ldesc->sqld) {
    return -1;
  }
  col=((loader_handle->ldesc)->sqlvar)+fieldnum;
  if (col->sqltype!=CFIXBINTYPE) {
    return -1;
  }
  return col->sqlxid;
}


int loader_putfield(loader_handle_t *loader_handle, const int fieldnum,
                    const void *buffer, const int buflen, const int sqlind,
                    int *badrecord)
{ loader_handle_t *h;
  struct sqlvar_struct *col;
  char *charp;
  int errinc;
  int cvstat;
  double doubleval;
  float floatval;
  char c;
  int i;
  int do_delete;
  loc_t simple_lo;
  mint bug1a;
  int2 bug1b;
  mlong bug2a;
  int4 bug2b;
  
  cvstat=0;
  h=loader_handle;
  charp=(char *)buffer;
  if (fieldnum==-1) {
    cvstat=rstoi(charp,(mint *)&i);
    if (cvstat!=0) {
      snprintf(h->errstr, ERR_BUF_SIZE,
               "Line %ld: not a short integer format for upload cmd column (%d): %s",
               h->rownum, cvstat, charp);
      h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                 sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum, h->rownum2_old,
                 0, h->logfile, h->verbosity, 10);
      cleanup(h);
      return 1;
    }
    if (i<0 || i>2) {
      snprintf(h->errstr, ERR_BUF_SIZE,
               "Line %ld: incorrect value in upload cmd column (%d): %s",
               h->rownum, cvstat, charp);
      h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                 sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum, h->rownum2_old,
                 0, h->logfile, h->verbosity, 10);
      cleanup(h);
      return 1;
    }
    h->uploadcmd=i;
    return 0;
  }
  if (h->unload_type==3 || h->unload_type==4) {
       h->used_tags[fieldnum]=1;
  }
  col=((h->ldesc)->sqlvar)+fieldnum;
  *badrecord=0;

  do_delete=0;
  if (h->uploadmode==4 && h->uploadcmd==2 && h->keycols[fieldnum]==0) {
    do_delete=1;
  }
  if (h->nullables[fieldnum]==0 && sqlind==-1 && !do_delete) {
    snprintf(h->errstr, ERR_BUF_SIZE,
            "Column %d:%s cannot be null at line %ld", fieldnum,
            col -> sqlname, h->rownum);
    if (!h->is_critical(0, 0, 1, h->errcnt,h->maxerrors, &errinc)) {
      h->errcnt+=errinc;
      if(h->oncrlf(h->datahandle, h->fname, h->rownum, h->fieldsnum, h->rownum,
                   h->rownum2_old, h->escapechar, h->logfile, h->onlogerror,
                   h->verbosity, h->datahandleoptions)==1) {
        h->loaded_rows=h->rownum2_old;
        cleanup(h);
        return 1;
      }
      h->rownum++;
      *badrecord=1;
      h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                    sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum, h->rownum2_old,
                    0, h->logfile, h->verbosity, 20);
      return 0; 
     } else {
      h->loaded_rows=h->rownum2_old;
      h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                    sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum, h->rownum2_old,
                    0, h->logfile, h->verbosity, 10);
      cleanup(h);
      return 1;
    }
  }
  h->indicators[fieldnum]=sqlind;
/*  sqlind: -1 if null, 0 if not null & no truncation,
           <full size> if data truncated */
  if (sqlind == -1) {
    switch (col->sqltype) {
      case CLOCATORTYPE:
      case SQLTEXT:
      case SQLBYTES: /* for these types rsetnull is not enough */
                     simple_lo.loc_indicator= -1; /* Null */
                     simple_lo.loc_bufsize=0;
                     simple_lo.loc_buffer=NULL;
                     simple_lo.loc_type=col->sqltype;
                     simple_lo.loc_loctype=LOCMEMORY;
                     simple_lo.loc_size=0;
                     bycopy((void *)&simple_lo, col->sqldata, sizeof(loc_t));
                     break;
    }
    rsetnull(col->sqltype,col->sqldata);
    return 0;
  }
  switch (col->sqltype) {
/* Unhandled types:
 * SQLSET, SQLMULTISET, SQLLIST, SQLROW, SQLCOLLECTION, SQLROWREF, SQLUDTVAR,
 * SQLUDTFIXED, SQLREFSER8
 */ 
    case CFLOATTYPE:
      cvstat=rstod(charp,&doubleval);
      floatval=(float)doubleval; /* how to detect overflow in portable way?
                             now problem is ignored */
      bycopy((char *)&floatval,col->sqldata,sizeof(floatval));
      if (cvstat!=0) {
        snprintf(h->errstr, ERR_BUF_SIZE, "Line %ld: not a float format: %s",
                 h->rownum,charp);
        h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                   sqlca.sqlerrd[1],sqlca.sqlerrm, h->rownum, h->rownum2_old,
                   0, h->logfile, h->verbosity, 10);
      }
      break;

    case CDOUBLETYPE:
      cvstat=rstod(charp, (double *)col->sqldata);
      if (cvstat!=0) {
        snprintf(h->errstr, ERR_BUF_SIZE, "Line %ld: not a double format: %s",
                 h->rownum,charp);
        h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                   sqlca.sqlerrd[1],sqlca.sqlerrm, h->rownum, h->rownum2_old,
                   0, h->logfile, h->verbosity, 10);
      }
      break;

    case CDECIMALTYPE:
      cvstat=deccvasc(charp,buflen,(struct decimal *)col->sqldata);
      if (cvstat!=0) {
        snprintf(h->errstr, ERR_BUF_SIZE,
                 "Line %ld: not a decimal format (%d): %s", h->rownum,
                 cvstat, charp);
        h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                   sqlca.sqlerrd[1],sqlca.sqlerrm, h->rownum, h->rownum2_old,
                   0, h->logfile, h->verbosity, 10);
      }
      break;

    case CINTTYPE:
/*      cvstat=rstol(charp,(mlong *)col->sqldata); */
      cvstat=rstol(charp,(mlong *)&bug2a);
     
      if (cvstat!=0) {
        snprintf(h->errstr, ERR_BUF_SIZE,
                 "Line %ld: not an integer format (%d): %s", h->rownum,
                 cvstat, charp);
        h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                   sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum, h->rownum2_old,
                   0, h->logfile, h->verbosity, 10);
      }
      bug2b=bug2a;
      bycopy((void *)&bug2b, col->sqldata, sizeof(bug2b));
      break;

    case CSHORTTYPE:
/*      cvstat=rstoi(charp,(mint *)col->sqldata); 
      above overrides 4 bytes in memory. it should be only 2 */
      cvstat=rstoi(charp,(mint *)&bug1a);
      if (cvstat!=0) {
        snprintf(h->errstr, ERR_BUF_SIZE,
                 "Line %ld: not a short integer format (%d): %s",
                 h->rownum, cvstat, charp);
        h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                   sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum, h->rownum2_old,
                   0, h->logfile, h->verbosity, 10);
      }
      bug1b=bug1a;
      bycopy((void *)&bug1b, col->sqldata, sizeof(bug1b));
      break;

    case CINT8TYPE:
      cvstat=ifx_int8cvasc(charp, buflen, (ifx_int8_t *)col->sqldata);
      if (cvstat!=0) {
        snprintf(h->errstr, ERR_BUF_SIZE,
                 "Line %ld: not a long integer format (%d): %s",
                 h->rownum,cvstat,charp);
        h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                   sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum, h->rownum2_old,
                   0, h->logfile, h->verbosity, 10);
      }
      break;

    case CBOOLTYPE:
      c=charp[0];
      if ((c!='t' && c!='f' && c!='T' && c!='F') || (buflen !=1)) {
        snprintf(h->errstr, ERR_BUF_SIZE,
                 "Line %ld: not a boolean format (%d): %s", h->rownum,
                 cvstat, charp);
        h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                   sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum, h->rownum2_old,
                   0, h->logfile, h->verbosity, 10);
      }
      bycopy(&c,col->sqldata,sizeof(c));
      break;

    case CDATETYPE:
      cvstat=rstrdate(charp,(int4 *)col->sqldata);
      if (cvstat!=0) {
        snprintf(h->errstr, ERR_BUF_SIZE,
                 "Line %ld: not a date format (%d): %s", h->rownum,
                 cvstat, charp);
        h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                   sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum, h->rownum2_old,
                   0, h->logfile, h->verbosity, 10);
      }
      break;

    case CDTIMETYPE:
      cvstat=dtcvasc(charp,(dtime_t *)col->sqldata);
      if (cvstat!=0) {
        snprintf(h->errstr, ERR_BUF_SIZE,
                 "Line %ld: not a datetime format (%d): %s", h->rownum,
                 cvstat, charp);
        h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                   sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum, h->rownum2_old,
                   0, h->logfile, h->verbosity, 10);
      }
      break;

    case CINVTYPE:
      cvstat=incvasc(charp,(intrvl_t *)col->sqldata);
      if (cvstat!=0) {
        snprintf(h->errstr, ERR_BUF_SIZE,
                 "Line %ld: not an interval format (%d): %s", h->rownum,
                 cvstat, charp);
        h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                   sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum, h->rownum2_old,
                   0, h->logfile, h->verbosity, 10);
      }
      break;

    case CCHARTYPE:
      stcopy(charp, (char *)col->sqldata);
      break;

    case CFIXCHARTYPE:
      byfill(col->sqldata, col->sqllen, ' ');
      bycopy(charp, col->sqldata, col->sqllen);
      break;

    case CVCHARTYPE:
      stcopy(charp,(char *)col->sqldata);
      break;

    case CLVCHARTYPE:
      stcopy(charp,(char *)col->sqldata);
      break;

    case CLOCATORTYPE:
    case SQLTEXT:
    case SQLBYTES:
      bycopy(charp, col->sqldata, sizeof(loc_t));
      break;
      
#ifdef SQLUDTFIXED
    case CFIXBINTYPE:
      bycopy(charp, col->sqldata, sizeof(ifx_lo_t));
      break;
#endif

    default:
      /*  Usupported data type.  */
      snprintf(h->errstr, ERR_BUF_SIZE, "unsupported type:%s:%s:%d",
              col->sqlname, rtypname(col->sqltype), col->sqltype);
      h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                 sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum, h->rownum2_old, 0,
                 h->logfile, h->verbosity, 10);
      h->loaded_rows=h->rownum2_old;
      cleanup(h);
      return 1;
  }
  return 0;
}

int loader_putrecord(loader_handle_t *loader_handle)
{
  EXEC SQL BEGIN DECLARE SECTION;
    struct sqlda *ldesc;
    struct sqlda *ludesc;
    struct sqlda *lddesc;
    struct sqlda *lsdesc;
    struct sqlda *lsdesc2;
    char stmt_id[128];
    char stmt_idu[128];
    char stmt_idd[128];
    char stmt_ids[128];
    char curs_id[128];
    char curs_ids[128];
  EXEC SQL END DECLARE SECTION;
  loader_handle_t *h;
  int errinc;
  struct sqlvar_struct *col;
  int i;
  unsigned long int buf2;
  int uploadmode;
  int do_delete;
  int j;

  h=loader_handle;

  ldesc=h->ldesc;
  ludesc=h->ludesc;
  lddesc=h->lddesc;
  lsdesc=h->lsdesc;
  lsdesc2=h->lsdesc2;
  sprintf(stmt_id, "lstmt_id%ld", h->instance_id);
  if (h->uploadmode!=0) {
    sprintf(stmt_idu, "lstmt_idu%ld", h->instance_id);
    sprintf(stmt_idd, "lstmt_idd%ld", h->instance_id);
    sprintf(stmt_ids, "lstmt_ids%ld", h->instance_id);
  }
  if (h->usecursor) {
    sprintf(curs_id,"lcurs_id%ld",h->instance_id);
  }
  if (h->uploadmode && h->uploadmode2) {
    sprintf(curs_ids, "lcurs_ids%ld",h->instance_id);
  }
  if (h->unload_type==3 || h->unload_type==4) {
    for (col = h->ldesc->sqlvar, i = 0; i < h->ldesc->sqld; col++, i++) {
      if (h->used_tags[i]==0) {
        h->indicators[i]=-1;
        rsetnull(col->sqltype,col->sqldata);
        if (h->unload_type==4) {
          snprintf(h->errstr, ERR_BUF_SIZE, "Record:  %ld: missing field %s", h->rownum,
                  col->sqlname);
          if (!h->is_critical(0, 0, 4, h->errcnt, h->maxerrors, &errinc)) {
            h->rownum++;
            h->errcnt+=errinc;
            h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                          sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum,
                          h->rownum2_old, 0, h->logfile, h->verbosity, 20);
            return 0;
          } else {
            h->loaded_rows=h->rownum2_old;
            h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                          sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum,
                          h->rownum2_old, 0, h->logfile, h->verbosity, 10);
            cleanup(h);
            return 1;
          } 
        }
      }
      h->used_tags[i]=0;
      do_delete=0;
      if (h->uploadmode==4 && h->uploadcmd==2 && h->keycols[i]==0) {
        do_delete=1;
      }
      if (h->nullables[i]==0 && h->indicators[i]==-1 && !do_delete) {
        snprintf(h->errstr, ERR_BUF_SIZE,
                "Column %d:%s cannot be null at line %ld",
                i, col -> sqlname, h->rownum);

        if (!h->is_critical(0, 0, 1, h->errcnt, h->maxerrors, &errinc)) {
          h->errcnt+=errinc;
          h->rownum++;
          h->errcnt+=errinc;
          h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                        sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum,
                        h->rownum2_old, 0, h->logfile, h->verbosity, 20);
          return 0;
        } else {
          h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                        sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum,
                        h->rownum2_old, 0, h->logfile, h->verbosity, 10);
          cleanup(h);
          return 1;
        }
      }
    }
  }
  if (h->usecursor && !h->uploadmode) {
    EXEC SQL PUT :curs_id using descriptor ldesc;
    if ( sqlca.sqlcode != 0) {
      snprintf(h->errstr, ERR_BUF_SIZE,
              "Row: %ld : cannot insert record: "INT4_FORMAT":"INT4_FORMAT"%s",
              h->rownum, sqlca.sqlcode, sqlca.sqlerrd[1], sqlca.sqlerrm);
      if (h->is_critical(sqlca.sqlcode, sqlca.sqlerrd[1], 0, h->errcnt,
                          h->maxerrors, &errinc)) {
        h->loaded_rows=h->rownum2_old;
        h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                      sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum, h->rownum2_old,
                      0, h->logfile, h->verbosity, 10);
        cleanup(h);
        return 1;
      } else {
        h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                      sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum, h->rownum2_old,
                      0, h->logfile, h->verbosity, 20);
        h->errcnt+=errinc;
      }
    }
    h->rownum2+=sqlca.sqlerrd[2];
    h->inserted_rows0+=sqlca.sqlerrd[2];
    h->ldesc=ldesc;
  } else {
    uploadmode=h->uploadmode;
    if (uploadmode<0 || uploadmode>4) {
      h->uploadmode=0;
      uploadmode=0;
    }
/* uploadcmd: 0 - update, 1 - insert, 2 - delete */
/* uploadcmd2: <>0 - use select to determine insert or update */
/* uploadmode:
     0 - inserts only,
     1 - try insert first; if failed try update
     2 - try update first; if failed try insert
     3 - insert or update according to statistics
     4 - update/insert/delete according to uploadcmd value
*/
    if (h->uploadmode==4) {
      if (h->uploadcmd==0) { /* do update */
        uploadmode=2;
      } else {
        if (h->uploadcmd==2) { /* do delete */
          uploadmode=4;
        } else {
          if (h->uploadcmd==1) { /* do insert */
            uploadmode=1;
          } else {
            uploadmode=0; /* incorrect upload command; shouldn't happen */
          }
        }
      }
    }
    /* now uploadmode 4 means something different: delete record */
    if (uploadmode==3) {
      if (h->inserted_rows0 > h->updated_rows0) {
        uploadmode=1;
      } else {
        uploadmode=2;
      }
    }
    /* now uploadmode value can be: 0,1,2,4 */
    if (uploadmode && h->uploadmode2 && uploadmode!=4) {
      /* do select and depending on return value set uploadmode to 1 or 2 */
      j=0;
      for (col = h->ldesc->sqlvar, i = 0; i < h->ldesc->sqld; col++, i++) {
        if ( h->indicators[i]==-1 && h->keycols[i]) {
          snprintf(h->errstr, ERR_BUF_SIZE,
                  "Row: %ld : key field %d (%s) cannot be null",
                  h->rownum, i, col->sqlname );
          if (h->is_critical(sqlca.sqlcode, sqlca.sqlerrd[1], 0, h->errcnt,
                             h->maxerrors, &errinc)) {
            h->loaded_rows=h->rownum2_old;
            h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                          sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum,
                          h->rownum2_old, 0, h->logfile, h->verbosity, 10);
            cleanup(h);
            return 1;
          } else {
            h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                          sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum,
                          h->rownum2_old, 0, h->logfile, h->verbosity, 20);
            h->errcnt+=errinc;
            j++;
            break;
          }
        }
      }
      if (j==0) {
          /* do select here */
        EXEC SQL open :curs_ids using descriptor lsdesc;
        if (sqlca.sqlcode!=0) {
          snprintf(h->errstr, ERR_BUF_SIZE,
                  "Row: %ld : cannot open select for record: "INT4_FORMAT","INT4_FORMAT" %s",
                  h->rownum, sqlca.sqlcode, sqlca.sqlerrd[1], sqlca.sqlerrm);
          if (h->is_critical(sqlca.sqlcode, sqlca.sqlerrd[1], 0, h->errcnt,
                             h->maxerrors, &errinc)) {
            h->loaded_rows=h->rownum2_old;
            h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                          sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum,
                          h->rownum2_old, 0, h->logfile, h->verbosity, 10);
            cleanup(h);
            return 1;
          } else {
            h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                          sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum,
                          h->rownum2_old, 0, h->logfile, h->verbosity, 20);
            h->errcnt+=errinc;
          }
        }
        lsdesc2->sqlvar->sqldata=(void *)&buf2;
        if (sqlca.sqlcode==0) {
          EXEC SQL fetch :curs_ids using descriptor lsdesc2;
          i=sqlca.sqlcode;
          lsdesc2->sqlvar->sqldata=NULL;
          if (i!=0 && i!=100) {
            snprintf(h->errstr, ERR_BUF_SIZE,
                    "Row: %ld : cannot fetch record: "INT4_FORMAT","INT4_FORMAT" %s",
                    h->rownum, sqlca.sqlcode, sqlca.sqlerrd[1], sqlca.sqlerrm);
            EXEC SQL close :curs_ids;
            if (h->is_critical(sqlca.sqlcode, sqlca.sqlerrd[1], 0, h->errcnt,
                               h->maxerrors, &errinc)) {
              h->loaded_rows=h->rownum2_old;
              h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                            sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum,
                            h->rownum2_old, 0, h->logfile, h->verbosity, 10);
              cleanup(h);
              return 1;
            } else {
              h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                            sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum,
                            h->rownum2_old, 0, h->logfile, h->verbosity, 20);
              h->errcnt+=errinc;
            }
          }
          if (i==0 || i==100) {
            EXEC SQL close :curs_ids;
            if (sqlca.sqlcode!=0) {
              snprintf(h->errstr, ERR_BUF_SIZE,
                      "Row: %ld : cannot close select cursor: "INT4_FORMAT","INT4_FORMAT" %s",
                      h->rownum, sqlca.sqlcode, sqlca.sqlerrd[1], sqlca.sqlerrm);
              if (h->is_critical(sqlca.sqlcode, sqlca.sqlerrd[1], 0, h->errcnt,
                                 h->maxerrors, &errinc)) {
                h->loaded_rows=h->rownum2_old;
                h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                              sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum,
                              h->rownum2_old, 0, h->logfile, h->verbosity, 10);
                cleanup(h);
                return 1;
              } else {
                h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                              sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum,
                              h->rownum2_old, 0, h->logfile, h->verbosity, 20);
                h->errcnt+=errinc;
              }
            }
            if (sqlca.sqlcode==0) {
              if (i==100) {
                uploadmode=1;
              } else {
                uploadmode=2;
              }
            }
          }
        }
      }
    }
    if (uploadmode==2 && h->noupdates==0) {
      j=0;
      for (col = h->ldesc->sqlvar, i = 0; i < h->ldesc->sqld; col++, i++) {
        if ( h->indicators[i]==-1 && h->keycols[i]) {
          snprintf(h->errstr, ERR_BUF_SIZE,
                  "Row: %ld : key field %d (%s) cannot be null",
                  h->rownum, i, col->sqlname );
          if (h->is_critical(sqlca.sqlcode, sqlca.sqlerrd[1], 0, h->errcnt,
                             h->maxerrors, &errinc)) {
            h->loaded_rows=h->rownum2_old;
            h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                          sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum,
                          h->rownum2_old, 0, h->logfile, h->verbosity, 10);
            cleanup(h);
            return 1;
          } else {
            h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                          sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum,
                          h->rownum2_old, 0, h->logfile, h->verbosity, 20);
            h->errcnt+=errinc;
            j++;
            break;
          }
        }
      }
      if (j==0) {
        EXEC SQL execute :stmt_idu using descriptor ludesc;
        h->ludesc=ludesc;
        if (sqlca.sqlcode!=0 && sqlca.sqlcode!=-239 && sqlca.sqlcode!=-268
          && sqlca.sqlerrd[1]!=-100) {
          snprintf(h->errstr, ERR_BUF_SIZE,
                  "Row: %ld : cannot update record: "INT4_FORMAT","INT4_FORMAT" %s",
                  h->rownum, sqlca.sqlcode, sqlca.sqlerrd[1], sqlca.sqlerrm);
          if (h->is_critical(sqlca.sqlcode, sqlca.sqlerrd[1], 0, h->errcnt,
                             h->maxerrors, &errinc)) {
            h->loaded_rows=h->rownum2_old;
            h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                          sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum, h->rownum2_old,
                          0, h->logfile, h->verbosity, 10);
            cleanup(h);
            return 1;
          } else {
            h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                          sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum, h->rownum2_old,
                          0, h->logfile, h->verbosity, 20);
            h->errcnt+=errinc;
          }
        }
        if (sqlca.sqlcode==0) {
          h->rownum2+=sqlca.sqlerrd[2];
          h->updated_rows0+=sqlca.sqlerrd[2];
        }
      }
    }
    if (uploadmode<2 ||
      (uploadmode==2 && sqlca.sqlerrd[2]==0 && !h->noinserts &&
         (sqlca.sqlcode==0 || sqlca.sqlcode==-239 || sqlca.sqlcode!=-268 ||
          sqlca.sqlerrd[1]==-100)) ||
      (uploadmode==2 && h->noupdates && !h->noinserts) ) {
      EXEC SQL execute :stmt_id using descriptor ldesc;
      h->ldesc=ldesc;
      if ( sqlca.sqlcode != 0 && (uploadmode==0 || (uploadmode==1
          && sqlca.sqlcode!=-239 && sqlca.sqlcode!=-268
          && sqlca.sqlerrd[1]!=-100))) {
        snprintf(h->errstr, ERR_BUF_SIZE,
                "Row: %ld : cannot insert record: "INT4_FORMAT","INT4_FORMAT" %s",
                h->rownum, sqlca.sqlcode, sqlca.sqlerrd[1], sqlca.sqlerrm);
        if (h->is_critical(sqlca.sqlcode, sqlca.sqlerrd[1], 0, h->errcnt,
                            h->maxerrors, &errinc)) {
          h->loaded_rows=h->rownum2_old;
          h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                        sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum, h->rownum2_old,
                        0, h->logfile, h->verbosity, 10);
          cleanup(h);
          return 1;
        } else {
          h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                        sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum, h->rownum2_old,
                        0, h->logfile, h->verbosity, 20);
          h->errcnt+=errinc;
        }
      }
      if (sqlca.sqlcode==0) {
        h->rownum2+=sqlca.sqlerrd[2];
        h->inserted_rows0+=sqlca.sqlerrd[2];
      }
    }
    if (uploadmode==1 && h->noupdates==0 && (sqlca.sqlcode==-239  || sqlca.sqlcode==-268
         || sqlca.sqlerrd[1]==-100)) {
      j=0;
      for (col = h->ldesc->sqlvar, i = 0; i < h->ldesc->sqld; col++, i++) {
        if ( h->indicators[i]==-1 && h->keycols[i]) {
          snprintf(h->errstr, ERR_BUF_SIZE,
                  "Row: %ld : key field %d (%s) cannot be null",
                  h->rownum, i, col->sqlname );
          if (h->is_critical(sqlca.sqlcode, sqlca.sqlerrd[1], 0, h->errcnt,
                             h->maxerrors, &errinc)) {
            h->loaded_rows=h->rownum2_old;
            h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                          sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum,
                          h->rownum2_old, 0, h->logfile, h->verbosity, 10);
            cleanup(h);
            return 1;
          } else {
            h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                          sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum,
                          h->rownum2_old, 0, h->logfile, h->verbosity, 20);
            h->errcnt+=errinc;
            j++;
            break;
          }
        }
      }
      if (j==0) {
        EXEC SQL execute :stmt_idu using descriptor ludesc;
        if ( sqlca.sqlcode != 0) {
          snprintf(h->errstr, ERR_BUF_SIZE,
                  "Row: %ld : cannot update record: "INT4_FORMAT","INT4_FORMAT" %s",
                  h->rownum, sqlca.sqlcode, sqlca.sqlerrd[1], sqlca.sqlerrm);
          if (h->is_critical(sqlca.sqlcode, sqlca.sqlerrd[1], 0, h->errcnt,
                             h->maxerrors, &errinc)) {
            h->loaded_rows=h->rownum2_old;
            h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                          sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum, h->rownum2_old,
                          0, h->logfile, h->verbosity, 10);
            cleanup(h);
            return 1;
          } else {
            h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                          sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum, h->rownum2_old,
                          0, h->logfile, h->verbosity, 20);
            h->errcnt+=errinc;
          }
        }
        if (sqlca.sqlcode==0) {
          h->rownum2+=sqlca.sqlerrd[2];
          h->updated_rows0+=sqlca.sqlerrd[2];
        }
        h->ludesc=ludesc;
      }
    }
    if (uploadmode==4) {
      j=0;
      for (col = h->ldesc->sqlvar, i = 0; i < h->ldesc->sqld; col++, i++) {
        if ( h->indicators[i]==-1 && h->keycols[i]) {
          snprintf(h->errstr, ERR_BUF_SIZE,
                  "Row: %ld : key field %d (%s) cannot be null",
                  h->rownum, i, col->sqlname );
          if (h->is_critical(sqlca.sqlcode, sqlca.sqlerrd[1], 0, h->errcnt,
                             h->maxerrors, &errinc)) {
            h->loaded_rows=h->rownum2_old;
            h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                          sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum,
                          h->rownum2_old, 0, h->logfile, h->verbosity, 10);
            cleanup(h);
            return 1;
          } else {
            h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                          sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum,
                          h->rownum2_old, 0, h->logfile, h->verbosity, 20);
            h->errcnt+=errinc;
            j++;
            break;
          }
        }
      }
      if (j==0) {
        EXEC SQL execute :stmt_idd using descriptor lddesc;
        if ( sqlca.sqlcode != 0) {
          snprintf(h->errstr, ERR_BUF_SIZE,
                  "Row: %ld : cannot delete record: "INT4_FORMAT","INT4_FORMAT "%s",
                  h->rownum, sqlca.sqlcode, sqlca.sqlerrd[1], sqlca.sqlerrm);
          if (h->is_critical(sqlca.sqlcode, sqlca.sqlerrd[1], 0, h->errcnt,
                             h->maxerrors, &errinc)) {
            h->loaded_rows=h->rownum2_old;
            h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                          sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum, h->rownum2_old,
                          0, h->logfile, h->verbosity, 10);
            cleanup(h);
            return 1;
          } else {
            h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                          sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum, h->rownum2_old,
                          0, h->logfile, h->verbosity, 20);
            h->errcnt+=errinc;
          }
        }
        if (sqlca.sqlcode==0) {
          h->rownum2+=sqlca.sqlerrd[2];
          h->deleted_rows0+=sqlca.sqlerrd[2];
        }
        h->lddesc=lddesc;
      }
    }
  }
  if (h->blobs) {
    blob_free(h);
  }
  if (h->unload_type==3 || h->unload_type==4) {
    for (col = h->ldesc->sqlvar, i = 0; i < h->ldesc->sqld; col++, i++) {
      h->indicators[i]=0;
    }
  }

  if (h->onrow) {
    if ( h->onrow(h->fname, h->tablename, h->datahandle, h->rownum, h->rownum2,
         h->inserted_rows0, h->updated_rows0, h->deleted_rows0)
      !=0 ) {
      h->loaded_rows=h->rownum2_old;
      cleanup(h);
      return 1;
    }
  }

/*         if (batchsize>0 && (((rownum) % batchsize) == 0) && sqlca.sqlcode==0)  */
  if (h->batchsize>0 && ((h->rownum2 / h->batchsize)!= (h->rownum2_old /
      h->batchsize)) ) {
    if (h->usecursor) {
      EXEC SQL FLUSH :curs_id;
      h->rownum2+=sqlca.sqlerrd[2];
      h->inserted_rows0+=sqlca.sqlerrd[2];
    }
    if ( sqlca.sqlcode != 0) {
      snprintf(h->errstr, ERR_BUF_SIZE,
              "Row: %ld : cannot flush record: "INT4_FORMAT","INT4_FORMAT" %s",
              h->rownum, sqlca.sqlcode, sqlca.sqlerrd[1],sqlca.sqlerrm);
      if (h->is_critical(sqlca.sqlcode, sqlca.sqlerrd[1], 0, h->errcnt,
          h->maxerrors, &errinc)) {
        h->loaded_rows=h->rownum2_old;
        h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                      sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum,
                      h->rownum2_old, 0, h->logfile, h->verbosity, 10);
        cleanup(h);
        return 1;
      } else {
        h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                      sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum,
                      h->rownum2_old, 0, h->logfile, h->verbosity, 20);
        h->errcnt+=errinc;
      }
    }
 
    if (h->usecommit) {
      EXEC SQL commit work;
      if ( sqlca.sqlcode != 0 ) {
        snprintf(h->errstr, ERR_BUF_SIZE,
                "Cannot commit transaction: "INT4_FORMAT":"INT4_FORMAT":%s",
                sqlca.sqlcode, sqlca.sqlerrd[1], sqlca.sqlerrm);
        h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                   sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum, h->rownum2_old, 0,
                   h->logfile, h->verbosity, 10);
        h->loaded_rows=h->rownum2_old;
        cleanup(h);
        return 1;
      }
      h->isinsidetrans=0;
    }
    h->rownum2_old=h->rownum2;
    h->inserted_rows=h->inserted_rows0;
    h->updated_rows=h->updated_rows0;
    h->deleted_rows=h->deleted_rows0;
    if ( h->onbatch(h->fname, h->tablename, h->datahandle, h->rownum, h->rownum2,
         h->inserted_rows0, h->updated_rows0, h->deleted_rows0, h->logfile,
         h->onlogerror, h->verbosity) !=0 ) {
      h->loaded_rows=h->rownum2_old;
      cleanup(h);
      return 1;
    }
    if (h->usecommit) {
      if (h->usecommit==2) {
        EXEC SQL begin work without replication;
      } else {
        EXEC SQL begin work;
      }
      if ( sqlca.sqlcode != 0 ) {
        snprintf(h->errstr, ERR_BUF_SIZE,
                "Cannot begin transaction: "INT4_FORMAT":"INT4_FORMAT":%s",
                sqlca.sqlcode, sqlca.sqlerrd[1], sqlca.sqlerrm);
        h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                   sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum, h->rownum2_old, 0,
                   h->logfile, h->verbosity, 10);
        h->loaded_rows=h->rownum2_old;
        cleanup(h);
        return 1;
      }
      h->isinsidetrans=1;
     }
  } else {
    if (!h->usecommit) {
      h->rownum2_old=h->rownum2;
      h->inserted_rows=h->inserted_rows0;
      h->updated_rows=h->updated_rows0;
      h->deleted_rows=h->deleted_rows0;
    }
  }
  h->rownum++;
  h->uploadcmd=0; 
  return 0;
}

int loader_close(loader_handle_t *loader_handle)
{
  loader_handle_t *h;
  int errinc;
  EXEC SQL BEGIN DECLARE SECTION;
    char curs_id[128];
  EXEC SQL END DECLARE SECTION;
 
  h=loader_handle;
  
  if (h->usecursor) {
    sprintf(curs_id,"lcurs_id%ld",h->instance_id);
    EXEC SQL FLUSH :curs_id;
    h->rownum2+=sqlca.sqlerrd[2];
    h->inserted_rows0+=sqlca.sqlerrd[2];
    if ( sqlca.sqlcode != 0) {
      snprintf(h->errstr, ERR_BUF_SIZE,
              "Row: %ld : cannot flush record: "INT4_FORMAT","INT4_FORMAT" %s",
              h->rownum, sqlca.sqlcode, sqlca.sqlerrd[1], sqlca.sqlerrm);
      if (h->is_critical(sqlca.sqlcode, sqlca.sqlerrd[1], 0, h->errcnt,
          h->maxerrors, &errinc)){
        h->loaded_rows=h->rownum2_old;
        h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                      sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum,
                      h->rownum2_old, 0, h->logfile, h->verbosity, 10);
        cleanup(h);
        return 1;
      } else {
        h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                      sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum,
                      h->rownum2_old, 0, h->logfile, h->verbosity, 20);
        h->errcnt+=errinc;
      }
    }
  }
  if (h->usecommit) {
    EXEC SQL commit work;
    if ( sqlca.sqlcode != 0 ) {
      snprintf(h->errstr, ERR_BUF_SIZE,
              "Cannot commit transaction: "INT4_FORMAT":"INT4_FORMAT":%s",sqlca.sqlcode,
              sqlca.sqlerrd[1], sqlca.sqlerrm);
      h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                    sqlca.sqlerrd[1], sqlca.sqlerrm, h->rownum, h->rownum2_old,
                    0, h->logfile, h->verbosity, 10);
      h->loaded_rows=h->rownum2_old;
      cleanup(h);
      return 1;
    }
    h->isinsidetrans=0;
    h->rownum2_old=h->rownum2;
    h->inserted_rows=h->inserted_rows0;
    h->updated_rows=h->updated_rows0;
    h->deleted_rows=h->deleted_rows0;
  }
  if (h->onfinish(h->fname, h->tablename, h->datahandle, h->rownum-1,
      h->rownum2, h->inserted_rows, h->updated_rows, h->deleted_rows,
      h->logfile, h->onlogerror, h->verbosity) !=0) {
    h->loaded_rows=h->rownum2_old;
    cleanup(h);
    return 1;
  }
  h->loaded_rows=h->rownum2_old;
  return 0;
}

int loader_free(loader_handle_t *loader_handle)
{
  loader_handle_t *h;

  h=loader_handle;
  h->fieldsnum=0;
  h->tablename[0]=0;
  cleanup(h);
  load2_free(h,__FILE__,__LINE__); /* = loader_destroy */
  return 0;
}

int load (const char * insstmt2, const char *fname, const char delim,
          const int batchsize, const int usecursor, const int usecommit,
          const int maxerrors,  unsigned long int *loaded_rows,
          const void *logfile, int verbosity,
          int (*oninit)(const char *fname, const char *tablename,
                        const int batchsize, void **handle,
                        const void *logfile, const int fieldsnum,
                        const char *insstmt2,
                        onlogerror_lpt onlogerror, const int verbosity,
                        const char *datahandleoptions),
          int (*onrow)(const char *fname, const char *tablename,
                       const void *handle, const unsigned long int rownum,
                       const unsigned long int loaded,
                       const unsigned long int inserted,
                       const unsigned long int updated,
                       const unsigned long int deleted),
          int (*onbatch)(const char *fname, const char *tablename,
                         const void *handle, const unsigned long int rownum,
                         const unsigned long int loaded,
                         const unsigned long int inserted,
                         const unsigned long int updated,
                         const unsigned long int deleted,
                         const void *logfile,
                         onlogerror_lpt onlogerror, const int verbosity),
          int (*onfinish)(const char *fname, const char *tablename,
                          const void *handle,
                          const unsigned long int rownum,
                          const unsigned long int rownum2,
                          const unsigned long int inserted,
                          const unsigned long int updated,
                          const unsigned long int deleted,
                          const void *logfile,
                          onlogerror_lpt onlogerror, const int verbosity),
          int (*onreadfield)(const void *fd, const int sqltype,
                             const int4 sqlxid,
                             const char *filename, char **dst,
                             const unsigned int maxdstbuf, const char delim,
                             const char escapech, int * eol,
                             int * sqlind, const unsigned long int rownum,
                             const unsigned long int loaded, const int fieldnum,
                             const char * fieldname, const int max_fld_name,
                             char *fieldname2, const void *logfile,
                             onlogerror_lpt onlogerror, const int verbosity,
                             const char *datahandleoptions),
          void (*oncleanup)(const char *fname, void *handle,
                            const void *logfile,
                            onlogerror_lpt onlogerror, const int verbosity,
                            const char *datahandleoptions),
          onlogerror_lpt onlogerror,
          int (*oncrlf)(const void *fd, const char *filename,
                        const unsigned long int rownum,
                        const unsigned long int loaded, const int fields_number,
                        const int bequiet, const char escapech,
                        const void *logfile,
                        onlogerror_lpt onlogerror, const int verbosity,
                        const char *datahandleoptions),
          int (*onrecordbegin)(const void *fd, const char *filename,
                               const unsigned long int rownum,
                               const unsigned long int loaded,
                               const int fields_number, const int bequiet,
                               const char escapech, const void *logfile,
                               onlogerror_lpt onlogerror, const int verbosity,
                               const char *datahandleoptions),
          int (*is_critical)(const long int sqlcode, const long int isamcode,
                             const int errtype, const int errcnt,
                             const int maxerors, int *errinc)
)

/* Parameters:
 *  - insstmt2 - insert statement,
 *  - fname - file name to be loaded,
 *  - delim - field delimiter,
 *  - batchsize - number of inserted records after which the following occurs:
 *    - cursor is flushed (if cursor used),
 *    - commit ends old transaction and begin starts a new one (if transactions
 *      are used),
 *    - onbatch event handler is called
 *    batchsize: 0 - all records in 1 batch, n - n transactions per batch, n >0,
 *  - usecursor
 *    - 1 use insert cursor,
 *    - 0 use plain insert
 *  - usecommit
 *    - 1 issue begin work/commit work/rollback work statements
 *    - 2 same as 1 except "begin work without replication"
 *    - 0 do not use begin work/commit work/rollback work statements
 *  - maxerrors - maximm number of errors before give up (0 - return after first
 *                error),
 *  - loaded_rows - returns number of rows actually loaded,
 *  - oninit - procedure is called before reading of a file starts (typically 
 *             it opens a file for read),
 *  - onrow - procedure called after each row to show progress, it can be used
 *            to allow user cancel loading,
 *  - onbatch - procedure called after each but last batch of transactions
 *            (typically to show progress message, it can be used to allow
 *             user cancel loading)
 *  - onfinish - procedure is called after succesfull end of reading a file
 *               (typically it closes a file)
 *  - onreadfield - procedure for reading a value for a single field from file
 *  - oncleanup - procedure for freeing file read resources (typically frees
 *                buffers and closes a file)
 *  - onlogerror - procedure responsible for writing error message information
 *                 to a file/console,
 *  - oncrlf - procedure responsible for reading/skipping data until EOL
 *  - onrecordbegin - procedure responsible for reading/skipping record header
 *  - is_critical - procedure which determines if an error should be ignored
 *                  by a procedure
 * Procedure returns 0 on success, 1 otherwise.
 * Limitations:
 * - not handled IDS2000 new datatypes.
 */
 
{
  loader_handle_t *loader_handle;
  int sqlind;
  int i, k;
  char buffer[32767+2];
  int all_fields;
  char sqlname2[MAX_FIELDNAME]; /* fieldname read from file */
  unsigned long int bytesread;
  int emptyread;
  int badrecord;
  int all_records;
  int batchsize2;
    
  buffer[0]=0;

  loader_handle=loader_create ();
  if (loader_handle==NULL) {
/*    sprintf(errstr,"Could not initialize loader - no memory");
    onlogerror(__FILE__, __LINE__, errstr, sqlca.sqlcode, sqlca.sqlerrd[1],
               sqlca.sqlerrm, 0, 0, 0, logfile);
    cleanup(fname, datahandle, 0, 0, udesc, indicators, nullables, used_tags,
            buffer, logfile, NULL, onlogerror); */
    return 1;
  }

  k=loader_setstmt(loader_handle, insstmt2);
  if (k!=0) {
    return 1;
  }

  k=loader_setdatahandlename(loader_handle, fname);
  if (k!=0) {
    return 1;
  }

  loader_setdelimiter(loader_handle, delim);


  if (batchsize<0) {
    batchsize2=0;
  } else {
    batchsize2=batchsize;
  }
  loader_setbatchsize(loader_handle, batchsize2);

  loader_setusecursor(loader_handle, usecursor);

  loader_setusecommit(loader_handle, usecommit);

  loader_setmaxerrors(loader_handle, maxerrors);

  if (logfile==NULL) {
    loader_setloghandle(loader_handle, NULL);
  } else {
    loader_setloghandle(loader_handle, logfile);
  }
  switch (verbosity) { 
    case 0: 
            loader_setverbosity(loader_handle, 0);
            break;
    case 1: 
            loader_setverbosity(loader_handle, 20);
            break;
    case 2: 
            loader_setverbosity(loader_handle, 30);
            break;
    case 3: 
            loader_setverbosity(loader_handle, 40);
            break;
    case 4: 
            loader_setverbosity(loader_handle, 50);
            break;
    default: 
            loader_setverbosity(loader_handle, 30);
            break;
  }



  if (oninit!=NULL) {
    loader_setoninit(loader_handle, oninit);
  }

  loader_setonrow(loader_handle, onrow);

  if (onbatch!=NULL) {
    loader_setonbatch(loader_handle, onbatch);
  }
  
  if (onfinish!=NULL) {
     loader_setonfinish(loader_handle, onfinish);
  }

  if (onreadfield!=NULL) {
     loader_setonreadfield(loader_handle, onreadfield);
  }

  if (oncleanup!=NULL) {
     loader_setoncleanup(loader_handle, oncleanup);
  }

  if (onlogerror!=NULL) {
     loader_setonlogerror(loader_handle, onlogerror);
  }

  if (oncrlf!=NULL) {
     loader_setoncrlf(loader_handle, oncrlf);
  }

  if (onrecordbegin!=NULL) {
     loader_setonrecordbegin(loader_handle, onrecordbegin);
  }

  if (is_critical!=NULL) {
     loader_setiscritical(loader_handle, is_critical);
  }

  *loaded_rows=0;
  k=loader_prepare(loader_handle);
  if (k!=0) {
    return 1;
  }
  k=loader_open(loader_handle);
  if (k!=0) {
    return 1;
  }
  i=0; /* current field number */
  if (loader_handle->uploadmode==4) {
    i=-1;
  }
  while (1) {
    if (i==0) {
      if (loader_beforerow(loader_handle)) {
        return 1;
      }
    }
    while (1) {
      k=loader_readfield(loader_handle, &i,
                       buffer, sizeof(buffer)-1,
                       &bytesread, &all_records,
                       sqlname2, MAX_FIELDNAME, &emptyread,
                       &all_fields, &badrecord, &sqlind);
      if (k!=0) {
        return 1;
      }
      if (all_records && i==0) {
        break;
      }
      if (emptyread && !all_fields && !all_records) {
        continue;
      }
      break;
    }
    /* assertion here: !emptyread || all_fields || all_records */

    if (badrecord) {
      i=0;
      continue;
    }
    if (!emptyread) {
      k=loader_putfield(loader_handle, i, buffer, bytesread, sqlind,
                        &badrecord);
      if (k!=0) {
        return 1;
      }
      i++;
    }
    if (all_fields) {
      k=loader_putrecord(loader_handle);
      if (k!=0) {
        return 1;
      }
      i=0;
      if (loader_handle->uploadmode==4) {
        i=-1;
      }
    }
    if (badrecord) {
      i=0;
      continue;
    }
    if (all_records) {
      break;
    }
  }
  if (loader_close(loader_handle)) {
    return 1;
  }
  *loaded_rows=loader_getloadedrows(loader_handle);
  if (loader_free(loader_handle)) { /* = loader_unprepare */
    return 1;
  }
  loader_handle=NULL;
  return 0;
}

void loader_setuploadmode(loader_handle_t *loader_handle, const short uploadmode)
{
  loader_handle->uploadmode=uploadmode;
}

void loader_setuploadmode2(loader_handle_t *loader_handle, const short uploadmode2)
{
  loader_handle->uploadmode2=uploadmode2;
}


int loader_setkeycols(loader_handle_t *loader_handle, const char *collist)
{ char *p1;
  int len;
  loader_handle_t *h;

  h=loader_handle;
  len=strlen(collist);
  p1=(char *) load2_malloc(len+1, __FILE__, __LINE__);
  if (p1==NULL) {
    sprintf(h->errstr,"Out of memory");
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0,
                  sqlca.sqlerrd[1], sqlca.sqlerrm, 0, 0, 0, h->logfile,
                  h->verbosity, 10);
    cleanup(h);
    return 1;
  }
  strcpy(p1,collist);
  h->keycols2=p1;

  return 0;
}

int convert_keycols(loader_handle_t *loader_handle)
{
  loader_handle_t *h;
  int i;
  char *p1, *p2, *p3;
  int len;
  int fieldnum;
  char ch;
  int isnum;
  char *collist;
  int endloop;


  /* setting keycols array */

  h=loader_handle;
  for (i=0; i<h->ldesc->sqld; i++) {
    h->keycols[i]=0;
  }

  collist=h->keycols2;
  p1=collist;
  endloop=0;
  ch=' ';
  while (1) {
    p2=strchr(p1, ',');
    if (p2==NULL) {
      if (!endloop) {
        len=strlen(p1);
        p2=p1+len;
        endloop=1;
      } else {
        break;
      }
    }
    while (p1<p2-1 && p1) {
      if ((*p1)==' ') {
        p1++;
      } else {
        break;
      }
    }
    p3=p2;
    while (p1<p2-1 ) {
      if (*(p2-1)==' ') {
        p2--;
      } else {
        break;
      }
    }
    len=p2-p1;
    if (len==0) {
      snprintf(h->errstr, ERR_BUF_SIZE,
              "Key field name must not be empty %s", p1);
      *p2=ch;
      h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, NULL, 0, 0,
                    0, h->logfile, h->verbosity, 10);
      cleanup(h);
      return 1;
    }
    ch=*p2;
    *p2=0;
    i=0;
    isnum=1;
    while (i<len) {
      if (!isdigit(p1[i])) {
        isnum=0;
        break;
      }
      i++;
    }
    if (isnum) {
      fieldnum=atoi(p1)-1;
    } else {
      fieldnum=field_lookup(h, p1);
    }
    if (fieldnum<0 || fieldnum>=h->fieldsnum) {
      snprintf(h->errstr, ERR_BUF_SIZE, "Specified key field %s does not exist",
              p1);
      *p2=ch;
      h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, NULL, 0, 0,
                    0, h->logfile, h->verbosity, 10);
      cleanup(h);
      return 1;
    }
    *p2=ch;
    p2=p3;
    h->keycols[fieldnum]=1;
    if (p2[0]==0) {
      break;
    }
    p2++;
    p1=p2;
  }
  return 0;
}

static int convert_upd(loader_handle_t *loader_handle)
{
  loader_handle_t *h;
  int i;
  char *p1;
 
  struct sqlvar_struct *col;
  char wherestr[50];
  char setstr[50];
  int l1, l2, l;
  int j;


  /* setting keycols array */

  h=loader_handle;

  i=convert_keycols(h);
  if (i!=0) {
    return 1;
  }

  /* actual statement conversion (insert stmt -> update stmt */

  l1=0;
  l2=0;
  j=0;
  l=0;
  for (col = h->ldesc->sqlvar, i = 0; i < h->ldesc->sqld; col++, i++) {
    wherestr[0]=0;
    setstr[0]=0;
    if (h->keycols[i]) {
      if (l!=0) {
        strcat(wherestr, " AND ");
      }
      strcat(wherestr, col->sqlname);
      strcat(wherestr, "=?");
      l++;
    } else {
      if (j!=0) {
        strcat(setstr, ", ");
      }
      strcat(setstr, col->sqlname);
      strcat(setstr, "=?");
      j++;
    }
    l1+=strlen(wherestr);
    l2+=strlen(setstr);
  }
  if (j==0) {
    /* This means that all fields in unload are in the key.
       In such a case update should not occur at all!
       It is easy when Update is tried after unsuccesfull insert.
       What to do if Update is first and Insert is made after?
       For this reason update statement have to be prepared anyway. */
    h->noupdates=1;
    return 0;
  }
  l=strlen("update ")+strlen(h->tablename)+strlen(" set ")+l2+
    strlen(" where ")+l1+1;
  p1=(char *)load2_malloc(l, __FILE__, __LINE__);
  if ( p1 == NULL ) {
    sprintf(h->errstr,"Out of memory");
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0,
                    sqlca.sqlerrd[1], sqlca.sqlerrm, 0, 0, 0, h->logfile,
                    h->verbosity, 10);
    cleanup(h);
    return 1;
  }
  strcpy(p1,"update ");
  strcat(p1,h->tablename);
  strcat(p1," set ");
  j=0;
  for (col = h->ldesc->sqlvar, i = 0; i < h->ldesc->sqld; col++, i++) {
    if (!h->keycols[i]) {
      if (j!=0) {
        strcat(p1, ", ");
      }
      strcat(p1, col->sqlname);
      strcat(p1, "=?");
      j++;
    }
  }
  strcat(p1, " where ");
  j=0;
  for (col = h->ldesc->sqlvar, i = 0; i < h->ldesc->sqld; col++, i++) {
    if (h->keycols[i]) {
      if (j!=0) {
        strcat(p1, " AND ");
      }
      strcat(p1, col->sqlname);
      strcat(p1, "=?");
      j++;
    }
  }
  h->updstmt=p1;
  return 0;
}

static int convert_sel(loader_handle_t *loader_handle)
{
  loader_handle_t *h;
  int i;
  char *p1;
  struct sqlvar_struct *col;
  char wherestr[50];
  int l1, l;
  int j;


  l1=0;
  l=0;
  h=loader_handle;
  for (col = h->ldesc->sqlvar, i = 0; i < h->ldesc->sqld; col++, i++) {
    wherestr[0]=0;
    if (h->keycols[i]) {
      if (l!=0) {
        strcat(wherestr, " AND ");
      }
      strcat(wherestr, col->sqlname);
      strcat(wherestr, "=?");
      l++;
    }
    l1+=strlen(wherestr);
  }
  l=strlen("select 1 from ")+strlen(h->tablename)+ strlen(" where ")+l1+1;
  p1=(char *)load2_malloc(l, __FILE__, __LINE__);
  if ( p1 == NULL ) {
    sprintf(h->errstr,"Out of memory");
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0,
                    sqlca.sqlerrd[1], sqlca.sqlerrm, 0, 0, 0, h->logfile,
                    h->verbosity, 10);
    cleanup(h);
    return 1;
  }
  strcpy(p1,"select 1 from ");
  strcat(p1,h->tablename);
  strcat(p1," where ");
  j=0;
  for (col = h->ldesc->sqlvar, i = 0; i < h->ldesc->sqld; col++, i++) {
    if (h->keycols[i]) {
      if (j!=0) {
        strcat(p1, " AND ");
      }
      strcat(p1, col->sqlname);
      strcat(p1, "=?");
      j++;
    }
  }
  h->selstmt=p1;
  return 0;  
}

static int convert_del(loader_handle_t *loader_handle)
{
  loader_handle_t *h;
  int i;
  char *p1;
  struct sqlvar_struct *col;
  char wherestr[50];
  int l1, l;
  int j;


  l1=0;
  l=0;
  h=loader_handle;
  for (col = h->ldesc->sqlvar, i = 0; i < h->ldesc->sqld; col++, i++) {
    wherestr[0]=0;
    if (h->keycols[i]) {
      if (l!=0) {
        strcat(wherestr, " AND ");
      }
      strcat(wherestr, col->sqlname);
      strcat(wherestr, "=?");
      l++;
    }
    l1+=strlen(wherestr);
  }
  l=strlen("delete from ")+strlen(h->tablename)+ strlen(" where ")+l1+1;
  p1=(char *)load2_malloc(l, __FILE__, __LINE__);
  if ( p1 == NULL ) {
    sprintf(h->errstr,"Out of memory");
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0,
                    sqlca.sqlerrd[1], sqlca.sqlerrm, 0, 0, 0, h->logfile,
                    h->verbosity, 10);
    cleanup(h);
    return 1;
  }
  strcpy(p1,"delete from ");
  strcat(p1,h->tablename);
  strcat(p1," where ");
  j=0;
  for (col = h->ldesc->sqlvar, i = 0; i < h->ldesc->sqld; col++, i++) {
    if (h->keycols[i]) {
      if (j!=0) {
        strcat(p1, " AND ");
      }
      strcat(p1, col->sqlname);
      strcat(p1, "=?");
      j++;
    }
  }
  h->delstmt=p1;
  return 0;  
}

static void blob_free(loader_handle_t *loader_handle)
{
  loader_handle_t *h;
  int i, j;
  struct sqlvar_struct *col;
  loc_t simple_lo;

  h=loader_handle;
  for (col = h->ldesc->sqlvar, i = 0; i < h->ldesc->sqld; col++, i++) {
    if (col->sqltype==CLOCATORTYPE
     || col->sqltype==SQLBYTES
     || col->sqltype==SQLTEXT) {

      bycopy(col->sqldata, (void *) &simple_lo, sizeof(loc_t));
      if (simple_lo.loc_indicator!=-1 && h->indicators[i]!=-1) {
        j=h->onblobfree(&simple_lo, h->logfile, h->onlogerror, h->verbosity);
        if (j!=0) {
          return;
        }
      }
    }
  } 
}

int simple_load (const char * insstmt2, const char *fname)
/* Parameters:
 *  - insstmt2 - insert statement,
 *  - fname - file name to be loaded,
 *                  by a procedure
 * Procedure returns 0 on success, 1 otherwise.
 * Limitations:
 * - not handled IDS2000 new datatypes.
 */
 
{
  loader_handle_t *loader_handle;
  int sqlind;
  int i, k;
  char buffer[32767+2];
  int all_fields;
  char sqlname2[MAX_FIELDNAME]; /* fieldname read from file */
  unsigned long int bytesread;
  int emptyread;
  int badrecord;
  int all_records;
  char delim;
  int batchsize;
  int usecursor;
  int usecommit;
  int maxerrors;

  delim='|';
  batchsize=10000;
  buffer[0]=0;
  usecursor=0;
  usecommit=0;
  maxerrors=0;

  loader_handle=loader_create ();
  if (loader_handle==NULL) {
/*    sprintf(errstr,"Could not initialize loader - no memory");
    onlogerror(__FILE__, __LINE__, errstr, sqlca.sqlcode, sqlca.sqlerrd[1],
               sqlca.sqlerrm, 0, 0, 0, logfile);
    cleanup(fname, datahandle, 0, 0, udesc, indicators, nullables, used_tags,
            buffer, logfile, NULL, onlogerror); */
    return 1;
  }

  k=loader_setstmt(loader_handle, insstmt2);
  if (k!=0) {
    return 1;
  }

  k=loader_setdatahandlename(loader_handle, fname);
  if (k!=0) {
    return 1;
  }

  loader_setdelimiter(loader_handle, delim);

  loader_setbatchsize(loader_handle, batchsize);

  loader_setusecursor(loader_handle, usecursor);

  loader_setusecommit(loader_handle, usecommit);

  loader_setmaxerrors(loader_handle, maxerrors);

  loader_setloghandle(loader_handle, NULL);

  loader_setverbosity(loader_handle, 0);

  k=loader_prepare(loader_handle);
  if (k!=0) {
    return 1;
  }
  k=loader_open(loader_handle);
  if (k!=0) {
    return 1;
  }
  i=0; /* current field number */
  while (1) {
    if (i==0) {
      if (loader_beforerow(loader_handle)) {
        return 1;
      }
    }
    while (1) {
      k=loader_readfield(loader_handle, &i,
                       buffer, sizeof(buffer)-1,
                       &bytesread, &all_records,
                       sqlname2, MAX_FIELDNAME, &emptyread,
                       &all_fields, &badrecord, &sqlind);
      if (k!=0) {
        return 1;
      }
      if (all_records && i==0) {
        break;
      }
      if (emptyread && !all_fields && !all_records) {
        continue;
      }
      break;
    }
    /* assertion here: !emptyread || all_fields || all_records */

    if (badrecord) {
      i=0;
      continue;
    }
    if (!emptyread) {
      k=loader_putfield(loader_handle, i, buffer, bytesread, sqlind,
                        &badrecord);
      if (k!=0) {
        return 1;
      }
      i++;
    }
    if (all_fields) {
      k=loader_putrecord(loader_handle);
      if (k!=0) {
        return 1;
      }
      i=0;
    }
    if (badrecord) {
      i=0;
      continue;
    }
    if (all_records) {
      break;
    }
  }
  if (loader_close(loader_handle)) {
    return 1;
  }
  if (loader_free(loader_handle)) { /* = loader_unprepare */
    return 1;
  }
  loader_handle=NULL;
  return 0;
}

