/******************************************************************************
*
* Library of data unloading functions for Informix database
*
* Author: Robert Sosnowski, Robert.Sosnowski@bzwbk.pl
*
* Version 1.70, 17-10-2005
*
* Copyright (c) 2001 - 2007. All rights reserved.
*
* Permission granted to use & distribute for free. Changes in source code
* require Author's acceptance.
*
* This software can be used only at your own risk. Author is not responsible 
* for any damage caused by using this software.
* 
******************************************************************************/

#include "load2cfg.h"

#include <stdio.h>

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif

#include <stdlib.h>
#include "unload2.h"
#include "load2cfg.h"
#include "tempfile.h"

#ifndef HAVE_SNPRINTF
#ifdef HAVE__SNPRINTF
#define snprintf _snprintf
#else
#include <stdarg.h>
int snprintf(char *str, size_t count, ...)
{int i;
 va_list args;
 i= sprintf(str, args);
 va_end(args);
return i;
} 
#endif
#endif 

#ifndef HAVE_SQLFREEMEM
#define SqlFreeMem sqlfreemem
#endif

/*EXEC SQL include sqlca;
EXEC SQL include sqlda; */
EXEC SQL include sqltypes;
EXEC SQL include decimal;
EXEC SQL include sqlstype;


static char *id="@(#) $Id: unload2.ec 2521 2007-11-21 08:55:56Z sosnor01 $";
static unsigned long int instance_id=0;

#define MAX_NBLOBS 2000
#define BLOB_BUF_SIZE 5000
#define BROKEN_FREE 1


static void cleanup(unl_handle_t *unl_handle);
int onlogerror_stdu(const char * srcfile, const int pline,
                    const char *errstr, const long int sqlcode,
                    const long int isamcode, const char *ifx_str,
                    const unsigned long int rownum,
                    const unsigned long int processed,
                    const int os_errno, const void *logfile,
                    const int verbosity, const int urgency);
int oninit_stdu(char *fname, const char *tablename, int batchsize,
                void **handle, char *escapech, const void *logfile, int fieldsnum,
                char *slctstmt2, onlogerror_upt onlogerror, int verbose,
                const char *datahandleoptions);
int onfinish_stdu(char *fname, char *tablename, void *handle,
                  unsigned long int rownum, const void *logfile,
                  onlogerror_upt onlogerror, int verbose);
int onwritefield_stdu(void *fd, int sqltype, int4 sqlxid, char *filename,
                      const char *src, const unsigned int buflen,
                      char delim, char escapech, short int sqlind,
                      unsigned long int rownum,  int fieldnum,
                      char *fieldname, const void *logfile,
                      onlogerror_upt onlogerror, int verbose,
                      const char *datahandleoptions,
                      void *fieldstate);
int onbatch_stdu(char *fname, const char *tablename, void *handle,
                 unsigned long int rownum,
                 const void *logfile,
                 onlogerror_upt onlogerror, int verbose);
void oncleanup_stdu(char *fname, void *handle, const void *logfile,
                    onlogerror_upt onlogerror, int verbose,
                    const char *datahandleoptions);
int oncrlf_stdu(void *fd, char *filename, unsigned long int rownum,
                const void *logfile,
                onlogerror_upt onlogerror, int verbose,
                const char *datahandleoptions);
int onrecordbegin_stdu(void *fd, char *filename, unsigned long int rownum,
                       const void *logfile,
                       onlogerror_upt onlogerror, int verbose,
                       const char *datahandleoptions);
int onblobfree_stdu(loc_t *simple_lo, const void *logfile,
                    onlogerror_upt onlogerror, int verbose);

mint openit(loc_t *loc, mint flag, mint bsize);
mint closeit(loc_t *loc);
mint writeit(loc_t *loc, char *buffer, mint bufflen);

int unl_setdatahandlename(unl_handle_t *unl_handle, const char *fname)
{
  unl_handle_t *h;
  int l;

  h=unl_handle;
  if (h->fname!=NULL) {
    load2_free(h->fname, __FILE__, __LINE__);
    h->fname=NULL;
  }
  l=strlen(fname)+1;
  h->fname=(char *)load2_malloc(l, __FILE__, __LINE__);
  if (h->fname==NULL) {
    snprintf(h->errstr, ERR_BUF_SIZE, "No memory for filename string %s",
            fname);
    h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
        sqlca.sqlerrd[1], sqlca.sqlerrm, 0, 0, 0, h->logfile, h->verbosity, 10);
    cleanup(h);
    return 1;
  }
  strcpy(h->fname,fname);
  return 0;
}

int unl_setdatahandleoptions(unl_handle_t *unl_handle, const char *datahandleoptions)
{
  unl_handle_t *h;
  int l;

  h=unl_handle;
  if (h->datahandleoptions!=NULL) {
    load2_free(h->datahandleoptions, __FILE__, __LINE__);
    h->datahandleoptions=NULL;
  }
  l=strlen(datahandleoptions)+1;
  h->datahandleoptions=load2_malloc(l, __FILE__, __LINE__);
  if (h->datahandleoptions==NULL) {
    snprintf(h->errstr, ERR_BUF_SIZE, "No memory for file options string %s",
            datahandleoptions);
    h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
        sqlca.sqlerrd[1], sqlca.sqlerrm, 0, 0, 0, h->logfile, h->verbosity, 10);
    cleanup(h);
    return 1;
  }
  strcpy(h->datahandleoptions,datahandleoptions);
  return 0;
}

void unl_setdelimiter(unl_handle_t *unl_handle, const char delim)
{
  unl_handle->delim=delim;
}

void unl_setbatchsize(unl_handle_t *unl_handle, const int batchsize)
{
  unl_handle->batchsize=batchsize;
}

void unl_setverbosity(unl_handle_t *unl_handle, const int verbosity)
{
  unl_handle->verbosity=verbosity;
}

void unl_setloghandle(unl_handle_t *unl_handle, const void *logfile)
{
  unl_handle->logfile=logfile;
}

void unl_setfetcharray(unl_handle_t *unl_handle, const short fetcharray)
{
  unl_handle->fetcharray=fetcharray;
}

void unl_setoninit(unl_handle_t *unl_handle,
                   int (*oninit)(char *fname, const char *tablename,
                                 int batchsize, void **handle, char *escapech,
                                 const void *logfile, int fieldsnum, char *slctstmt2,
                                 onlogerror_upt onlogerror, int verbose,
                                 const char *datahandleoptions))
{
  unl_handle->oninit=oninit;
}

void unl_setonrow(unl_handle_t *unl_handle,
                  int (*onrow)(char *fname, const char *tablename,
                               void *handle, unsigned long int rownum))
{
  unl_handle->onrow=onrow;
}

void unl_setonbatch(unl_handle_t *unl_handle,
                    int (*onbatch)(char *fname, const char *tablename,
                                   void *handle, unsigned long int rownum,
                                   const void *logfile, onlogerror_upt onlogerror,
                                   int verbose))
{
  unl_handle->onbatch=onbatch;
}


void unl_setonfinish(unl_handle_t *unl_handle,
                     int (*onfinish)(char *fname, char *tablename, void *handle,
                     unsigned long int rownum, const void *logfile,
                     onlogerror_upt onlogerror, int verbose))
{
  unl_handle->onfinish=onfinish;
}

void unl_setonwritefield(unl_handle_t *unl_handle,
                         int (*onwritefield)(void *fd, int sqltype,
                         int4 sqlxid, char *filename, const char *src,
                         const unsigned int buflen, char delim,
                         char escapech, short int sqlind,
                         unsigned long int rownum, int fieldnum,
                         char *fieldname, const void *logfile,
                         onlogerror_upt onlogerror, int verbose,
                         const char *datahandleoptions,
                         void *fieldstate))
{
  unl_handle->onwritefield=onwritefield;
}

void unl_setoncleanup(unl_handle_t *unl_handle,
                      void (*oncleanup)(char *fname, void *handle,
                      const void *logfile, onlogerror_upt onlogerror, int verbose,
                      const char *datahandleoptions))
{
  unl_handle->oncleanup=oncleanup;
}

void unl_setonlogerror(unl_handle_t *unl_handle,
                          onlogerror_upt onlogerror)
{
  unl_handle->onlogerror=onlogerror;
}

void unl_setoncrlf(unl_handle_t *unl_handle,
                   int (*oncrlf)(void *fd, char *filename,
                                 unsigned long int rownum,
                                 const void *logfile,
                                 onlogerror_upt onlogerror, int verbose,
                                 const char *datahandleoptions))
{
  unl_handle->oncrlf=oncrlf;
}

void unl_setonrecordbegin(unl_handle_t *unl_handle,
                          int (*onrecordbegin)(void *fd, char *filename,
                          unsigned long int rownum, const void *logfile,
                          onlogerror_upt onlogerror, int verbose,
                          const char *datahandleoptions))
{
  unl_handle->onrecordbegin=onrecordbegin;
}

void unl_setblobfree(unl_handle_t *unl_handle,
                     int (*onblobfree)(loc_t *simple_lo, const void *logfile,
                                       onlogerror_upt onlogerror,
                                       int verbose))
{
  unl_handle->onblobfree=onblobfree;
}

void unl_setescapechar(unl_handle_t *unl_handle, const char escapechar)
{
  unl_handle->escapechar=escapechar;
}

char * unl_getstmt(unl_handle_t *unl_handle)
{
  return unl_handle->slctstmt2;
}

char * unl_getdatahandlename(unl_handle_t *unl_handle)
{
  return unl_handle->fname;
}

char * unl_getdatahandleoptions(unl_handle_t *unl_handle)
{
  return unl_handle->datahandleoptions;
}

char unl_getdelimiter(unl_handle_t *unl_handle)
{
  return unl_handle->delim;
}

int unl_getbatchsize(unl_handle_t *unl_handle)
{
  return unl_handle->batchsize;
}

int unl_getverbosity(unl_handle_t *unl_handle)
{
  return unl_handle->verbosity;
}

unsigned long int unl_getunloadedrows(unl_handle_t *unl_handle)
{
  return unl_handle->unloaded_rows;
}

char *unl_geterrstr(unl_handle_t *unl_handle)
{
  return unl_handle->errstr;
}

char *unl_getablename(unl_handle_t *unl_handle)
{
  return unl_handle->tablename;
}

int unl_getfieldsnum(unl_handle_t *unl_handle)
{
  return unl_handle->fieldsnum;
}

char unl_getescapechar(unl_handle_t *unl_handle)
{
  return unl_handle->escapechar;
}

short unl_getfetcharray(unl_handle_t *unl_handle)
{
  return unl_handle->fetcharray;
}


unl_handle_t *unl_create (void)
{
  unl_handle_t *unl_handle;

  unl_handle=(unl_handle_t *)load2_malloc(sizeof(unl_handle_t), __FILE__, __LINE__);
  if ( unl_handle == NULL ) {
    return NULL;
  }

  unl_handle->slctstmt2=NULL;
  unl_handle->fname=NULL;
  unl_handle->delim='|';
  unl_handle->batchsize=0;
  unl_handle->verbosity=30;
  unl_handle->unloaded_rows=0;

  unl_handle->logfile=NULL;
  unl_handle->errstr[0]=0;
  unl_handle->tablename[0]=0;
  unl_handle->fieldsnum=0;
  unl_handle->escapechar='\\';
  unl_handle->datahandle=NULL;
  unl_handle->datahandleoptions=NULL;
 
  unl_handle->onlogerror=&onlogerror_stdu;
  unl_handle->oninit=&oninit_stdu;
  unl_handle->onrow=NULL;
  unl_handle->onbatch=&onbatch_stdu;
  unl_handle->onfinish=&onfinish_stdu;
  unl_handle->onwritefield=&onwritefield_stdu;
  unl_handle->oncleanup=&oncleanup_stdu;
  unl_handle->oncrlf=&oncrlf_stdu;
  unl_handle->onrecordbegin=&onrecordbegin_stdu;
  unl_handle->onblobfree=&onblobfree_stdu;
  unl_handle->openblob=&openit;
  unl_handle->closeblob=&closeit;
  unl_handle->writeblob=&writeit;

  unl_handle->indicators=NULL;
  unl_handle->udesc=NULL;
  unl_handle->recordbuffer=NULL;
  unl_handle->rownum=0;
  unl_handle->iscursoropen=0;
  unl_handle->instance_id=instance_id;
  unl_handle->blobs=0;
  unl_handle->fetarrsize=0;
  unl_handle->fetarridx=0;
  unl_handle->fetcharray=0;
  instance_id++;

  return unl_handle;
}

void unl_settablename(unl_handle_t *unl_handle, const char *tablename)
{
  unl_handle_t *h;

  h=unl_handle;
  strcpy(h->tablename, tablename);
}

int unl_setstmt(unl_handle_t *unl_handle, const char *selstmt)
{
  unl_handle_t *h;
  int l;

  h=unl_handle;
  if (h->slctstmt2!=NULL) {
    load2_free(h->slctstmt2, __FILE__, __LINE__);
    h->slctstmt2=NULL;
  }
  l=strlen(selstmt)+1;
  h->slctstmt2=(char *)load2_malloc(l, __FILE__, __LINE__);
  if (h->slctstmt2==NULL) {
    h->slctstmt2=NULL;
    snprintf(h->errstr, ERR_BUF_SIZE, "No memory for statement string %s",
             selstmt);
    h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
        sqlca.sqlerrd[1], sqlca.sqlerrm, 0, 0, 0, h->logfile, h->verbosity, 10);
    cleanup(h);
    return 1;
  }
  strcpy(h->slctstmt2, selstmt);
  return 0;
}


int unl_prepare(unl_handle_t *unl_handle)
{
  unl_handle_t *h;
#if MI_PTR_SIZE == 64
  mlong pos, cp, pos2;
#else
  mint pos, cp, pos2;
#endif
  char *cp2;
  int len;
  int i;
  struct sqlvar_struct *col;
EXEC SQL BEGIN DECLARE SECTION;
  int fields_number;
  char *slctstmt;
  char stmt_id[128];
  struct sqlda *udesc2;
EXEC SQL END DECLARE SECTION;
  long memsize;
  loc_t *blobp;
  ifx_lo_t *slobp;
  int j;

  h=unl_handle;
  sprintf(stmt_id,"ustmt_id%ld",h->instance_id);



  h->rownum=0;
  h->recordbuffer=NULL;
  slctstmt=h->slctstmt2;
  h->udesc=NULL;
  h->indicators=NULL;
  h->datahandle=NULL;
  h->unloaded_rows=0;
  if (h->batchsize<0) {
    h->batchsize=0;
  }
  EXEC SQL whenever error continue;

#ifdef DEBUG_ALLOC
printf("prepare stmt_id\n");
#endif
  EXEC SQL prepare :stmt_id from :slctstmt;
  if ( sqlca.sqlcode != 0 ) {
    sprintf(h->errstr,"Cannot prepare: "INT4_FORMAT":%s\n",sqlca.sqlcode, sqlca.sqlerrm);
    h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode, sqlca.sqlerrd[1],
               sqlca.sqlerrm, h->rownum, 0, 0, h->logfile, h->verbosity, 10);
    cleanup(h);
    return 1;
  }
  udesc2=h->udesc;
  EXEC SQL describe :stmt_id into udesc2;
#ifdef DEBUG_ALLOC
printf("described: %x\n", udesc2);
#endif
  h->udesc=udesc2;
  fields_number=h->udesc->sqld;
  h->fieldsnum=fields_number;
  if ( sqlca.sqlcode != 0 /* SELECT */ && sqlca.sqlcode != SQ_EXECPROC) {
    sprintf(h->errstr,"Not a select statement\n");
    h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode, sqlca.sqlerrd[1],
               sqlca.sqlerrm, h->rownum, 0, 0, h->logfile, h->verbosity, 10);
    cleanup(h);
    return 1;
  }
  if ( fields_number<=0 ) {
    sprintf(h->errstr, "There is no output columns\n");
    h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode, sqlca.sqlerrd[1],
               sqlca.sqlerrm, h->rownum, 0, 0, h->logfile, h->verbosity, 10);
    cleanup(h);
    return 1;
  }

  h->typmsizes=load2_malloc(sizeof(int)*(h->udesc->sqld), __FILE__, __LINE__);
  if ( h->typmsizes == NULL ) {
    sprintf(h->errstr,"Out of memory\n");
    h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode, sqlca.sqlerrd[1],
               sqlca.sqlerrm, h->rownum, 0, 0, h->logfile, h->verbosity, 10);
    cleanup(h);
    return 1;
  }
  memset(h->typmsizes,0,h->udesc->sqld);
  /* Step 1: analyze udesc to determine type and memory requirements
   *         of each item in the select list.  rtypalign() returns a
   *         pointer to the next appropriate boundary (starting at
   *         pos) for the indicated data type.
   * Unhandled types: SQLSET, SQLMULTISET, SQLLIST,
   * SQLROW, SQLCOLLECTION, SQLROWREF, SQLUDTVAR, SQLUDTFIXED, SQLREFSER8
   */
  pos = 0;
  pos2 = 0;
  h->blobs=0;
  for (col = h->udesc->sqlvar, i = 0; i < h->udesc->sqld; col++, i++) {
    len=stleng(col->sqlname);
    cp2=col->sqlname+len-1;
    while(len>1 && *cp2==' ') {
      *cp2='\0';
      len--, cp2--;
    }

    switch(col->sqltype) {
      case SQLSMFLOAT:
        col->sqltype = CFLOATTYPE;
        break;
      case SQLFLOAT:
        col->sqltype = CDOUBLETYPE;
        break;
      case SQLMONEY:
      case SQLDECIMAL:
        col->sqltype = CDECIMALTYPE;
        break;
      case SQLCHAR:
      case SQLNCHAR:
        col->sqltype = CCHARTYPE;
        break;
      case SQLLVARCHAR: 
        col->sqltype = CLVCHARTYPE;
        break;
      case SQLINT:
      case SQLSERIAL:
        col->sqltype = CINTTYPE;
        break;
      case SQLSMINT:
        col->sqltype = CSHORTTYPE;
        break;
      case SQLINT8:
      case SQLSERIAL8:
        col->sqltype = CINT8TYPE;
        break;
      case SQLVCHAR:
      case SQLNVCHAR:
        col->sqltype = CVCHARTYPE;
        break;
      case SQLDATE:
        col->sqltype = CDATETYPE;
        break;
      case SQLDTIME:
        col->sqltype = CDTIMETYPE;
        break;
      case SQLINTERVAL:
        col->sqltype = CINVTYPE;
        break;
      case SQLBOOL:
        col->sqltype = CBOOLTYPE;
        break;
      case SQLTEXT:
      case SQLBYTES:
        (h->blobs)++;
        break;

#ifdef SQLUDTFIXED
      case SQLUDTFIXED:
        switch (col->sqlxid) {
          case XID_BLOB:
            break;
          case XID_CLOB:
            break;
          default:
         /*  Usupported data type.  */
                sprintf(h->errstr,"unknown extended type:%s(idtype="INT4_FORMAT") for column %s\n", col->sqltypename,
                        col->sqlxid, col->sqlname);
                h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, NULL, 0, 0, 0,
                           h->logfile, h->verbosity, 10);
                cleanup(h);
                return 1;
        }
        col->sqltype=CFIXBINTYPE;
        break;
#endif

      default:
         /*  Usupported data type.  */
         sprintf(h->errstr,"unknown type:%s:%s:%d\n", col->sqlname,
                 rtypname(col->sqltype), col->sqltype);
         h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, NULL, 0, 0, 0,
                    h->logfile, h->verbosity, 10);
         cleanup(h);
         return 1;
    }

    memsize = rtypmsize(col->sqltype,col->sqllen);
    h->typmsizes[i]=memsize;
    if (col->sqltype!=CFIXBINTYPE) {
      pos = rtypalign( pos, col->sqltype) + memsize;
    }
      pos2 = rtypalign( pos2, col->sqltype) + memsize;  
  }
  if (h->fetcharray) {
#if MI_PTR_SIZE == 64
    if ( pos2 > (mlong) FetBufSize ) {
#else
    if ( pos2 > (mint) FetBufSize ) {
#endif
      /* Record size too large to use fetch array */ 
      FetArrSize = 1;
    } else {
      FetArrSize = (FetBufSize / pos2); /* global ESQL/C variable:
                                                        size of fetch array */
    }
    if (FetArrSize<1) {
      FetArrSize=0;
    }
  } else {
    FetArrSize=0;
  }

  if (FetArrSize) {
    h->indicators=(int *)load2_malloc(sizeof(int)*(h->udesc->sqld)*FetArrSize, __FILE__, __LINE__);
  } else {
    h->indicators=(int *)load2_malloc(sizeof(int)*(h->udesc->sqld), __FILE__, __LINE__);
  }
  if ( h->indicators == NULL ) {
    sprintf(h->errstr,"Out of memory\n");
    h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode, sqlca.sqlerrd[1],
               sqlca.sqlerrm, h->rownum, 0, 0, h->logfile, h->verbosity, 10);
    cleanup(h);
    return 1;
  }

  if (FetArrSize>1) {
  /* recalculate everything from the beginning: exact size would have changed */
    pos = 0;
    for (col = h->udesc->sqlvar, i = 0; i < h->udesc->sqld; col++, i++) {
      if (col->sqltype!=CFIXBINTYPE) {
        memsize = rtypmsize(col->sqltype,col->sqllen)*FetArrSize;
        pos = rtypalign( pos, col->sqltype) + memsize;
      }
    }
  }
  /* Step 2: Allocate memory to receive a row of the table returned
   *         by the SELECT statement.  The variable pos has an integer
   *         value equal to the number of bytes occupied by the
   *         row.
   */
  if (pos!=0) {
    h->recordbuffer = load2_malloc(pos, __FILE__, __LINE__);
    if ( h->recordbuffer == NULL ) {
      sprintf(h->errstr,"Out of memory\n");
      h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode, sqlca.sqlerrd[1],
                 sqlca.sqlerrm, h->rownum, 0, 0, h->logfile, h->verbosity, 10);
      cleanup(h);
      return 1;
    }
  } else {
    h->recordbuffer=NULL; /* this can happen when unloading only smart blob */
  }

  memset(h->recordbuffer,0,pos);

/* Step 3: Set pointers in the allocated memory to receive each
 * item in the select list.
 */
#if MI_PTR_SIZE == 64
  cp = (mlong)h->recordbuffer;
#else
  cp = (mint)h->recordbuffer;
#endif
  for (col = h->udesc->sqlvar, i = 0; i < h->udesc->sqld; col++, i++) {
    memsize = rtypmsize(col->sqltype, col->sqllen);
    if (col->sqltype==CFIXBINTYPE) {
      if (col->sqlxid==XID_BLOB || col->sqlxid==XID_CLOB ) {
        col->sqldata=load2_malloc(memsize, __FILE__, __LINE__);
        slobp=(ifx_lo_t *) col->sqldata;
        if ( col->sqldata == NULL ) {
          sprintf(h->errstr,"Out of memory\n");
          h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode, sqlca.sqlerrd[1],
                     sqlca.sqlerrm, h->rownum, 0, 0, h->logfile, h->verbosity, 10);
          cleanup(h);
          return 1;
        }
      }
    } else {
      cp = rtypalign( cp, col->sqltype);
      col->sqldata = (char *)cp;
      memsize = rtypmsize(col->sqltype, col->sqllen);
      cp += memsize;
      if (col->sqltype==SQLBYTES || col->sqltype==SQLTEXT ||
          col->sqltype==CLOCATORTYPE ) {
        col->sqllen=memsize; /* Bug in ESQLC ? without this assignment
                                fetch statement hangs
                                when using fetch array and LOCUSER;
                                memsize=sizeof(loc_t)=64, col->sqllen=56
                                8 byte difference because of 4 pointers
                                to functions: open, read, write, close.
                                Apparently pointers are overwritten when 56.
                                Redhat, CSDK 2.81UC2 */
        blobp=(loc_t *)col->sqldata;
        blobp->loc_type=col->sqltype;
        blobp->loc_loctype=LOCUSER;
        blobp->loc_bufsize=0;
        blobp->loc_buffer=NULL;
        blobp->loc_user_env=(char *)h;
        blobp->loc_open=h->openblob;
        blobp->loc_write=h->writeblob;
        blobp->loc_close=h->closeblob;
      }
      if (col->sqltype==CDTIMETYPE) {
        ((dtime_t *)col->sqldata)->dt_qual=(int2)col->sqllen;
      }
      if (col->sqltype==CINVTYPE) {
         ((intrvl_t *)col->sqldata)->in_qual=(int2)col->sqllen;
      }
      if (col->sqltype==CCHARTYPE || col->sqltype==CVCHARTYPE
        || col->sqltype==CLVCHARTYPE){
        col->sqllen=memsize;
      }
    }
    if (FetArrSize) {
      col->sqlind = (short *)h->indicators+i*FetArrSize;
    } else {
      col->sqlind = (short *)h->indicators+i;
    }

    for (j=1; j<FetArrSize; j++) {
      if (col->sqltype==CFIXBINTYPE) {
        if (col->sqlxid==XID_BLOB || col->sqlxid==XID_CLOB ) {
          slobp=(ifx_lo_t *) col->sqldata;
          col->sqldata=load2_malloc(memsize, __FILE__, __LINE__);
          if ( col->sqldata == NULL ) {
            sprintf(h->errstr,"Out of memory\n");
            h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode, sqlca.sqlerrd[1],
                       sqlca.sqlerrm, h->rownum, 0, 0, h->logfile, h->verbosity, 10);
            cleanup(h);
            return 1;
          }
            switch (col->sqlxid) {
              case XID_BLOB:
                break;
              case XID_CLOB:
               break;
              default: break;
            }
          }
      } else {
        cp += memsize;
        if (col->sqltype==SQLBYTES || col->sqltype==SQLTEXT ||
            col->sqltype==CLOCATORTYPE ) {
          col->sqllen=memsize;
          blobp=(loc_t *)(col->sqldata+j*memsize);
          blobp->loc_type=col->sqltype;
          blobp->loc_loctype=LOCUSER;
          blobp->loc_bufsize=0;
          blobp->loc_buffer=NULL;
          blobp->loc_user_env=(char *)h;
          blobp->loc_open=h->openblob;
          blobp->loc_write=h->writeblob;
          blobp->loc_close=h->closeblob;
        }
        if (col->sqltype==CDTIMETYPE) {
          ((dtime_t *)(col->sqldata+j*memsize))->dt_qual=(int2)col->sqllen;
        }
        if (col->sqltype==CINVTYPE) {
          ((intrvl_t *)(col->sqldata+j*memsize))->in_qual=(int2)col->sqllen;
        } 
      }
    }
  }

  return 0; 
}

int unl_open(unl_handle_t *unl_handle)
{
  unl_handle_t *h;
  char escapech;
  void *h1;
  EXEC SQL BEGIN DECLARE SECTION;
    char stmt_id[128];
    char curs_id[128];
  EXEC SQL END DECLARE SECTION;
  struct sqlvar_struct *col;
  int i;

  h=unl_handle;
  sprintf(stmt_id,"ustmt_id%ld",h->instance_id);
  sprintf(curs_id,"ucurs_id%ld",h->instance_id);

#ifdef DEBUG_ALLOC
printf("Declare cursor: %s\n", curs_id);
#endif
  EXEC SQL declare :curs_id cursor with hold for :stmt_id;
  if ( sqlca.sqlcode != 0 ) {
    sprintf(h->errstr,"Cannot declare cursor: "INT4_FORMAT":"INT4_FORMAT":%s\n",sqlca.sqlcode,
            sqlca.sqlerrd[1], sqlca.sqlerrm);
    h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode, sqlca.sqlerrd[1],
               sqlca.sqlerrm, h->rownum, 0, 0, h->logfile, h->verbosity, 10);
    cleanup(h);
    return 1;
  }
 
  EXEC SQL open :curs_id;
  if ( sqlca.sqlcode != 0 ) {
    sprintf(h->errstr,"Cannot open cursor: "INT4_FORMAT":"INT4_FORMAT":%s\n",sqlca.sqlcode,
            sqlca.sqlerrd[1], sqlca.sqlerrm);
    h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode, sqlca.sqlerrd[1],
               sqlca.sqlerrm, h->rownum, 0, 0, h->logfile, h->verbosity, 10);
    cleanup(h);
    return 1;
  }
  h->iscursoropen=1;
  h->fetarridx=h->fetarrsize;

  h1=h->datahandle;
  h->fieldstates=load2_malloc(h->fieldsnum*sizeof(void *), __FILE__, __LINE__);
  if (h->fieldstates==NULL) {
    snprintf(h->errstr, ERR_BUF_SIZE, "No memory for field states table");
    h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
        sqlca.sqlerrd[1], sqlca.sqlerrm, 0, 0, 0, h->logfile, h->verbosity, 10);
    cleanup(h);
    return 1;
  }
  byfill((char *)h->fieldstates,h->fieldsnum*sizeof(void *),0);
  for (col = h->udesc->sqlvar, i = 0; i < h->fieldsnum; col++, i++) {
    if (col->sqltype==CFIXBINTYPE &&
        (col->sqlxid==XID_BLOB || col->sqlxid==XID_CLOB )) {
      h->fieldstates[i]=load2_malloc(sizeof(slob_handle_t), __FILE__, __LINE__);
      if (h->fieldstates[i]==NULL) {
        snprintf(h->errstr, ERR_BUF_SIZE, "No memory for slob handle");
        h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
            sqlca.sqlerrd[1], sqlca.sqlerrm, 0, 0, 0, h->logfile, h->verbosity, 10);
        cleanup(h);
        return 1;
      }
      byfill(h->fieldstates[i],sizeof(slob_handle_t),0);
    }
  }
  if (h->oninit &&
      (h->oninit(h->fname, h->tablename, h->batchsize, &h1, &escapech,
                h->logfile, h->fieldsnum, h->slctstmt2, h->onlogerror,
                h->verbosity, h->datahandleoptions )!=0)) {
    cleanup(h);
    return 1;
  }
  h->escapechar=escapech;
  h->datahandle=h1;
  h->rownum=1;
  h->unloaded_rows=0;
 
  return 0;
}

int unl_getfield(unl_handle_t *unl_handle, const int fieldnum,
                 void *buffer, const unsigned int buflen,
                 unsigned long int *bytesread, int *sqlind)
{
  const int max_numeric_string=40;
  struct sqlvar_struct *col;
  unl_handle_t *h;
  int len;
  char *fieldp, *cp;
  char *sqldata;
  short *sqlind2;
  loc_t simple_lo;

  h=unl_handle;
  fieldp=buffer;
  col=h->udesc->sqlvar+fieldnum;

  if (h->fetarridx) {
    sqldata=col->sqldata+h->fetarridx*h->typmsizes[fieldnum];
  } else {
    sqldata=col->sqldata;
  }

  if (FetArrSize<=1) {
    sqlind2 = col->sqlind;
  } else {
    sqlind2 = col->sqlind+h->fetarridx;
  }

  *sqlind=*sqlind2;
/*  byfill(fieldp,buflen,0); */
/*  byfill(fieldp,max_numeric_string+1,0);*/ /* setting whole buffer is too slow */
  fieldp[max_numeric_string]=0;

/* Issue: strncpy is slow because often it initializes whole buffer to zeros
   Buffer is big, strncpy is in innermost loop.
   Result: from 10 to 20 times slower unload. Tested on SCO, Win, Linux */

 #define SLOW_STRNCPY 1
  switch (col->sqltype) {
    case CFLOATTYPE:
      if ( *sqlind2 == -1 ) {
        fieldp[0]='\0';
      } else {
#ifndef SLOW_STRNCPY
        snprintf(fieldp,buflen-1,"%f", (double) *((float *) sqldata));
#else
        sprintf(fieldp,"%f", (double) *((float *) sqldata));
#endif
      }
      break;
    case CDOUBLETYPE:
      if ( *sqlind2 == -1 ) {
             fieldp[0]='\0';
         } else {
#ifndef SLOW_STRNCPY
             snprintf(fieldp,buflen-1,"%f",*((double *) sqldata));
#else
             sprintf(fieldp,"%f",*((double *) sqldata));
#endif
         }
         break;
    case CDECIMALTYPE:
      if ( *sqlind2 == -1 ) {
        fieldp[0]='\0';
      } else {
        dectoasc( (dec_t *) sqldata, fieldp, max_numeric_string,-1);
        fieldp[max_numeric_string]=0;
      }
      break;
    case CINTTYPE:
      if ( *sqlind2 == -1 ) {
        fieldp[0]='\0';
      } else {
#ifndef SLOW_STRNCPY
        snprintf(fieldp,buflen-1,INT4_FORMAT,*((int4 *) sqldata));
#else
        sprintf(fieldp,INT4_FORMAT,*((int4 *) sqldata));
#endif
/*        rfmtlong(*((int4 *) sqldata),"-<<<<<<<<", fieldp); */
      }
      break;
    case CSHORTTYPE:
      if ( *sqlind2 == -1 ) {
        fieldp[0]='\0';
      } else {
#ifndef SLOW_STRNCPY
        snprintf(fieldp,buflen-1, "%d",*((int2 *) sqldata));
#else
        sprintf(fieldp, "%d",*((int2 *) sqldata));
#endif
      }
      break;
    case CINT8TYPE:
      if ( *sqlind2 == -1 ) {
        fieldp[0]='\0';
      } else {
        ifx_int8toasc((struct ifx_int8 *) sqldata,fieldp,max_numeric_string); /* 27 is enough for maximum value */
        fieldp[max_numeric_string]=0;
      }
      break;
    case CBOOLTYPE:
      if ( *sqlind2 == -1 ) {
        fieldp[0]='\0';
      } else {
        if ((*sqldata)=='\0'){
          fieldp[0]='F';
        } else {
          fieldp[0]='T';
        }
        fieldp[1]='\0';
      }
      break;
    case CDATETYPE:
      if ( *sqlind2 == -1 ) {
        fieldp[0]='\0';
      } else {
        rdatestr( *((int4 *) sqldata),fieldp);
      }
      break;
    case CDTIMETYPE:
      if ( *sqlind2 == -1 ) {
        fieldp[0]='\0';
      } else {
        dttoasc(((dtime_t *) sqldata), fieldp);
      }
      break;
    case CINVTYPE:
      if ( *sqlind2 == -1 ) {
        fieldp[0]='\0';
      } else {
        intoasc(((intrvl_t *) sqldata), fieldp);
      }
      break;
    case CCHARTYPE:
      if ( *sqlind2 == -1 ) {
        fieldp[0]='\0';
      } else {
#ifndef SLOW_STRNCPY
        strncpy(fieldp, sqldata, buflen-1);
#else
        strcpy(fieldp,sqldata);
#endif
/*        stcopy(sqldata,fieldp); */
      }
      break;
    case CVCHARTYPE:
      if ( *sqlind2 == -1 ) {
        fieldp[0]='\0';
      } else {
#ifndef SLOW_STRNCPY
        strncpy(fieldp, sqldata, buflen-1);
#else
        strcpy(fieldp, sqldata);
#endif
/*        stcopy(sqldata, fieldp); */
      }
      break; 
    case CLVCHARTYPE:
      if ( *sqlind2 == -1 ) {
        fieldp[0]='\0';
      } else {
#ifndef SLOW_STRNCPY
        strncpy(fieldp, sqldata, buflen-1);
#else
        strcpy(fieldp, sqldata);
#endif
      }
      break;

    case CLOCATORTYPE:
    case SQLTEXT:
    case SQLBYTES:
      if ( *sqlind2 == -1 ) {
        simple_lo.loc_indicator= -1; /* Null */
        simple_lo.loc_bufsize=0;
        simple_lo.loc_buffer=NULL;
        simple_lo.loc_type=col->sqltype;
        simple_lo.loc_loctype=LOCMEMORY;
        simple_lo.loc_size=0;
        bycopy((void *)&simple_lo, fieldp, sizeof(loc_t));
      } else {
        bycopy(sqldata, fieldp, sizeof(loc_t));
      }
      break;

#ifdef SQLUDTFIXED
    case CFIXBINTYPE:
      if ( *sqlind2 == -1 ) {
        fieldp[0]='\0';
      } else {
        bycopy(sqldata, fieldp, sizeof(ifx_lo_t));
      }
      break;
#endif

    default:
      /*  Usupported data type.  */
      snprintf(h->errstr, ERR_BUF_SIZE, "unknown type:%s:%s:%d\n", col->sqlname,
              rtypname(col->sqltype), col->sqltype);
      h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, NULL, h->rownum, 0, 0,
                   h->logfile, h->verbosity, 10);
      cleanup(h);
      return 1;
  }
  len = stleng(fieldp);
  if (col->sqltype == CCHARTYPE || col->sqltype == CDECIMALTYPE
      || col->sqltype == CINT8TYPE) {
    len = stleng(fieldp);
    cp = fieldp + len - 1;
    while ( len > 1 && *cp == ' ') {
      *cp = '\0';
      len--, cp--;
    }
    len = cp - fieldp + 1;
 
  }
  *bytesread=len;
  return 0;
}

int unl_beforerow(unl_handle_t *unl_handle)
{
  unl_handle_t *h;

  h=unl_handle;
  if (h->onrecordbegin(h->datahandle, h->fname, h->rownum, h->logfile,
      h->onlogerror, h->verbosity, h->datahandleoptions)!=0) {
    cleanup(h);
    return 1;
  }
  return 0;
}

int unl_afterrow(unl_handle_t *unl_handle)
{
  unl_handle_t *h;

  h=unl_handle;
  if (h->oncrlf(h->datahandle, h->fname, h->rownum, h->logfile, h->onlogerror,
      h->verbosity, h->datahandleoptions)!=0) {
    cleanup(h);
    return 1;
  }
  h->rownum++;
  return 0;
}

int unl_writefield(unl_handle_t *unl_handle, const int fieldnum,
                 const void *buffer, const int buflen, const int sqlind,
                 void *fieldstate)
{
  struct sqlvar_struct *col;
  unl_handle_t *h;

  h=unl_handle;
  col=h->udesc->sqlvar+fieldnum;
  if (h->onwritefield(h->datahandle, col->sqltype, col->sqlxid, h->fname, buffer, buflen,
      h->delim, h->escapechar, (short int) sqlind, h->rownum,  fieldnum, col->sqlname,
      h->logfile, h->onlogerror, h->verbosity, h->datahandleoptions,
      fieldstate)!=0) {
    cleanup(h);
    return 1;
  }
  return 0;
}

int unl_getrecord(unl_handle_t *unl_handle, int *all_records)
{
  unl_handle_t *h;
  int i;
  EXEC SQL BEGIN DECLARE SECTION;
  struct sqlda *udesc2;
  char curs_id[128];

  EXEC SQL END DECLARE SECTION;

  h=unl_handle;
  *all_records=0;

  if (FetArrSize && h->fetarridx<h->fetarrsize-1) {
    h->fetarridx++;
  } else {
    sprintf(curs_id,"ucurs_id%ld",h->instance_id);
    udesc2=h->udesc;

    EXEC SQL fetch :curs_id using descriptor udesc2;
    i=sqlca.sqlcode;
    h->fetarrsize=sqlca.sqlerrd[2]; /* number of fetched records */
    h->fetarridx=0;
    h->udesc=udesc2;

    if (i==100) {
      *all_records=1;
      return 0;
    }
    if (i!=0) {
      snprintf(h->errstr, ERR_BUF_SIZE, "Cannot fetch data: "INT4_FORMAT","INT4_FORMAT",%s",
               sqlca.sqlcode, sqlca.sqlerrd[1], sqlca.sqlerrm);
      h->onlogerror(__FILE__, __LINE__, h->errstr, sqlca.sqlcode,
                    sqlca.sqlerrd[1], sqlca.sqlerrm, 0, 0, 0, h->logfile,
                    h->verbosity, 10);
      cleanup(h);
      return 1;
    }
  }
  if (h->onrow) {
   if ( h->onrow(h->fname, h->tablename, h->datahandle, h->rownum) !=0 ) {
      cleanup(h);
      return 1;
    }
  }
  if (h->batchsize>0 && (((h->rownum % h->batchsize)==0)) ) {
    if ( h->onbatch(h->fname, h->tablename, h->datahandle, h->rownum,
         h->logfile, h->onlogerror, h->verbosity) !=0 ) {
      cleanup(h);
      return 1;
    }
  }

  return 0;
}

int unl_close(unl_handle_t *unl_handle)
{
  unl_handle_t *h;

  h=unl_handle;
  if (h->onfinish &&
      (h->onfinish(h->fname, h->tablename, h->datahandle, h->rownum-1,
      h->logfile, h->onlogerror, h->verbosity) !=0)) {
    cleanup(h);
    return 1;
  }
  return 0;
}

int unl_free(unl_handle_t *unl_handle)
{
  unl_handle_t *h;

  h=unl_handle;
  h->fieldsnum=0;
  h->tablename[0]=0;
  cleanup(h);
  load2_free(h,__FILE__,__LINE__); /* = loader_destroy */
  return 0;
}

/* return value: 0 .. n-1 field number, -1 extra column containing command 
   for loadinc, -2 hard error (cleanup), -3 soft error (continue)
*/
int unl_getfieldtype(unl_handle_t *unl_handle, const int fieldnum)
{
  struct sqlvar_struct *col;

  col=unl_handle->udesc->sqlvar+fieldnum;
  return col->sqltype;
}

int unload (const char * slctstmt2, const char *fname, const char delim,
            const int batchsize, unsigned long int *unloaded_rows,
            const void *logfile, const int verbosity,
            const char *tablename,
            int (*oninit)(char *fname, const char *tablename, int batchsize,
                          void **handle, char *escapech, const void *logfile,
                          int fieldsnum, char *slctstmt2,
                          onlogerror_upt onlogerror, int verbose,
                          const char *datahandleoptions),
            int (*onrow)(char *fname, const char *tablename, void *handle,
                         unsigned long int rownum),
            int (*onbatch)(char *fname, const char *tablename, void *handle,
                           unsigned long int rownum,
                           const void *logfile,
                           onlogerror_upt onlogerror, int verbose),
            int (*onfinish)(char *fname, char *tablename, void *handle,
                            unsigned long int rownum,
                            const void *logfile,
                            onlogerror_upt onlogerror, int verbose),
            int (*onwritefield)(void *fd, int sqltype, int4 sqlxid,
                                char *filename, const char *src,
                                const unsigned int buflen,
                                char delim, char escapech,
                                short int sqlind, unsigned long int rownum,
                                int fieldnum, char *fieldname, const void *logfile,
                                onlogerror_upt onlogerror, int verbose,
                                const char *datahandleoptions,
                                void *fieldstate),
            void (*oncleanup)(char *fname, void *handle, const void *logfile,
                              onlogerror_upt onlogerror, int verbose,
                              const char *datahandleoptions),
            onlogerror_upt onlogerror,
            int (*oncrlf)(void *fd, char *filename, unsigned long int rownum,
                          const void *logfile,
                          onlogerror_upt onlogerror, int verbose,
                          const char *datahandleoptions),
            int (*onrecordbegin)(void *fd, char *filename,
                                 unsigned long int rownum, const void *logfile,
                                 onlogerror_upt onlogerror, int verbose,
                                 const char *datahandleoptions)
)
{
  unl_handle_t *unl_handle;
  int i, k;
  char buffer[32767+1];
  int batchsize2;
  int fieldsnum;
  int all_records;
  unsigned long int bytesread;
  int sqlind;


  unl_handle=unl_create();
  if (unl_handle==NULL) {
    return 1;
  }
  k=unl_setstmt(unl_handle, slctstmt2);
  if (k!=0) {
    return 1;
  }

  k=unl_setdatahandlename(unl_handle, fname);
  if (k!=0) {
    return 1;
  }

  unl_setdelimiter(unl_handle, delim);

  if (batchsize<0) {
    batchsize2=0;
  } else {
    batchsize2=batchsize;
  }
  unl_setbatchsize(unl_handle, batchsize2);

  if (logfile==NULL) {
    unl_setloghandle(unl_handle, NULL);
  } else {
    unl_setloghandle(unl_handle, logfile);
  }
  switch (verbosity) { 
    case 0: 
            unl_setverbosity(unl_handle, 0);
            break;
    case 1: 
            unl_setverbosity(unl_handle, 20);
            break;
    case 2: 
            unl_setverbosity(unl_handle, 30);
            break;
    case 3: 
            unl_setverbosity(unl_handle, 40);
            break;
    case 4: 
            unl_setverbosity(unl_handle, 50);
            break;
    default: 
            unl_setverbosity(unl_handle, 30);
            break;
  }
  unl_settablename(unl_handle, tablename);

  if (oninit!=NULL) {
    unl_setoninit(unl_handle, oninit);
  }
  unl_setonrow(unl_handle, onrow);

  if (onbatch!=NULL) {
    unl_setonbatch(unl_handle, onbatch);
  }
  
  if (onfinish!=NULL) {
     unl_setonfinish(unl_handle, onfinish);
  }

  if (onwritefield!=NULL) {
     unl_setonwritefield(unl_handle, onwritefield);
  }

  if (oncleanup!=NULL) {
     unl_setoncleanup(unl_handle, oncleanup);
  }

  if (onlogerror!=NULL) {
     unl_setonlogerror(unl_handle, onlogerror);
  }

  if (oncrlf!=NULL) {
     unl_setoncrlf(unl_handle, oncrlf);
  }

  if (onrecordbegin!=NULL) {
     unl_setonrecordbegin(unl_handle, onrecordbegin);
  }
  *unloaded_rows=0;


  k=unl_prepare(unl_handle);
   if (k!=0) {
    return 1;
  }
  fieldsnum=unl_getfieldsnum(unl_handle);

  k=unl_open(unl_handle);
  if (k!=0) {
    return 1;
  }
  if (onwritefield!=NULL) {
    while (1) {
      k=unl_getrecord(unl_handle, &all_records);
      if (k!=0) {
        return 1;
      }
      if (all_records) {
        break;
      }
      if (unl_beforerow(unl_handle)) {
        return 1;
      }
      i=0; /* current field number */

      while (i<fieldsnum) {
        k=unl_getfield(unl_handle, i, buffer, sizeof(buffer), &bytesread, &sqlind);
        if (k!=0) {
          return 1;
        }
        k=unl_writefield(unl_handle, i, buffer, bytesread, sqlind, unl_handle->fieldstates[i]);
        if (k!=0) {
          return 1;
        }
        i++;
      }

      if (i>=fieldsnum) {
        if (unl_afterrow(unl_handle)) {
          return 1;
        }
      }
    }
  }
  if (unl_close(unl_handle)) {
    return 1;
  }
  *unloaded_rows=unl_getunloadedrows(unl_handle);
  if (unl_free(unl_handle)) { /* = loader_unprepare */
    return 1;
  }
  unl_handle=NULL;
  return 0; 
}

static void cleanup(unl_handle_t *unl_handle)
{
  unl_handle_t *h;
  char errstr[ERR_BUF_SIZE];
  struct sqlvar_struct *col;
  int i;
  EXEC SQL BEGIN DECLARE SECTION;
  char stmt_id[128];
  char curs_id[128];
  EXEC SQL END DECLARE SECTION;

  h=unl_handle;
  sprintf(stmt_id,"ustmt_id%ld",h->instance_id);
  sprintf(curs_id,"ucurs_id%ld",h->instance_id);
  if (h->oncleanup != NULL) {
    sprintf(h->errstr, "before oncleanup");
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, NULL, h->rownum, 0, 0,
                    h->logfile, h->verbosity, 50);
    h->oncleanup(h->fname, h->datahandle, h->logfile, h->onlogerror,10,
                 h->datahandleoptions);
    sprintf(h->errstr, "after oncleanup");
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, NULL, h->rownum, 0, 0,
                    h->logfile, h->verbosity, 50);
  }
  if (h->iscursoropen) {
    sprintf(h->errstr,"Closing cursor %s...", curs_id);
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, NULL, h->rownum, 0, 0,
                    h->logfile, h->verbosity, 50);
    EXEC SQL close :curs_id;
    h->iscursoropen=0;
    if ( sqlca.sqlcode != 0 ) {
      sprintf(errstr,"Cannot close cursor: "INT4_FORMAT":"INT4_FORMAT":%s\n",sqlca.sqlcode,
              sqlca.sqlerrd[1], sqlca.sqlerrm);
      h->onlogerror(__FILE__, __LINE__, errstr, sqlca.sqlcode, sqlca.sqlerrd[1],
                    sqlca.sqlerrm, h->rownum, 0, 0,
                    h->logfile, h->verbosity, 10);
      return;
    }
    sprintf(h->errstr,"Freeing cursor %s...", curs_id);
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, NULL, h->rownum, 0, 0,
                    h->logfile, h->verbosity, 50);
#ifdef DEBUG_ALLOC
printf("Unprepare curs_id\n");
#endif
#ifndef BROKEN_FREE
    EXEC SQL free :curs_id;
#endif
#ifdef DEBUG_ALLOC
printf("After unprepare curs_id\n");
#endif
    sprintf(h->errstr,"Freeing statement %s...", curs_id);
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, NULL, h->rownum, 0, 0,
                    h->logfile, h->verbosity, 50);
#ifdef DEBUG_ALLOC
printf("Unprepare stmt_id\n");
#endif
#ifndef BROKEN_FREE
    EXEC SQL free :stmt_id;
#endif
#ifdef DEBUG_ALLOC
printf("After unprepare stmt_id\n");
#endif
  }
  for (col = h->udesc->sqlvar, i = 0; i < h->udesc->sqld; col++, i++) {
    sprintf(h->errstr,"Freeing fieldstates[%d]...", i);
    if (h->fieldstates!=NULL && (h->fieldstates)[i]!=NULL) {
      h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, NULL, h->rownum, 0, 0,
                      h->logfile, h->verbosity, 50);
      load2_free((h->fieldstates)[i], __FILE__, __LINE__);
    }
  }
  if (h->fieldstates!=NULL) {
    sprintf(h->errstr,"Freeing fieldstates...");
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, NULL, h->rownum, 0, 0,
                    h->logfile, h->verbosity, 50);
    load2_free(h->fieldstates, __FILE__, __LINE__);
  }
  for (col = h->udesc->sqlvar, i = 0; i < h->udesc->sqld; col++, i++) {
    sprintf(h->errstr,"Freeing slobs...");
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, NULL, h->rownum, 0, 0,
                    h->logfile, h->verbosity, 50);
    if (col->sqltype==CFIXBINTYPE) {
      if (col->sqlxid==XID_BLOB || col->sqlxid==XID_CLOB ) {
        load2_free(col->sqldata, __FILE__, __LINE__);
      }
    }
  }
  if (h->recordbuffer!=NULL) {
    sprintf(h->errstr,"Freeing recordbuffer...");
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, NULL, h->rownum, 0, 0,
                    h->logfile, h->verbosity, 50);
    load2_free(h->recordbuffer, __FILE__, __LINE__);
  }
  if (h->indicators!=NULL) {
    sprintf(h->errstr,"Freeing indicators...");
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, NULL, h->rownum, 0, 0,
                    h->logfile, h->verbosity, 50);
    load2_free(h->indicators, __FILE__, __LINE__);
  }
  if (h->typmsizes!=NULL) {
    sprintf(h->errstr,"Freeing typmsizes...");
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, NULL, h->rownum, 0, 0,
                    h->logfile, h->verbosity, 50);
    load2_free(h->typmsizes, __FILE__, __LINE__);
  }
  if (h->fname!=NULL) {
    sprintf(h->errstr,"Freeing fname...");
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, NULL, h->rownum, 0, 0,
                    h->logfile, h->verbosity, 50);
    load2_free(h->fname, __FILE__, __LINE__);
  }
  if (h->slctstmt2!=NULL) {
    sprintf(h->errstr,"Freeing slcstmt2...");
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, NULL, h->rownum, 0, 0,
                    h->logfile, h->verbosity, 50);
    load2_free(h->slctstmt2, __FILE__, __LINE__);
  }
  if (h->udesc!=NULL) {
    sprintf(h->errstr,"Resetting sqlind...");
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, NULL, h->rownum, 0, 0,
                    h->logfile, h->verbosity, 50);
    for (col = h->udesc->sqlvar, i = 0; i < h->udesc->sqld; col++, i++) {
      col->sqlind = NULL;
    }
    sprintf(h->errstr,"Freeing udesc...");
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, NULL, h->rownum, 0, 0,
                    h->logfile, h->verbosity, 50);
#ifdef SQLDA_FREE
#ifdef DEBUG_ALLOC
printf("Do SqlFreeMem: %x\n", h->udesc);
#endif
    SqlFreeMem(h->udesc, SQLDA_FREE);
#else
    load2_free(h->udesc, __FILE__, __LINE__);
#endif
  }
}

int simple_unload (const char * slctstmt2, const char *fname)
{
  unl_handle_t *unl_handle;
  int i, k;
  char buffer[32767+1];
  int batchsize2;
  int fieldsnum;
  int all_records;
  unsigned long int bytesread;
  int sqlind;
  char delim;
  int batchsize;
  char tablename[12];

  delim='|';
  batchsize=10000;
  strcpy(tablename,"<table>");
  unl_handle=unl_create();
  if (unl_handle==NULL) {
    return 1;
  }
  k=unl_setstmt(unl_handle, slctstmt2);
  if (k!=0) {
    return 1;
  }

  k=unl_setdatahandlename(unl_handle, fname);
  if (k!=0) {
    return 1;
  }

  unl_setdelimiter(unl_handle, delim);

  if (batchsize<0) {
    batchsize2=0;
  } else {
    batchsize2=batchsize;
  }
  unl_setbatchsize(unl_handle, batchsize2);

  unl_setloghandle(unl_handle, NULL);

  unl_setverbosity(unl_handle, 0);

  unl_settablename(unl_handle, tablename);

  k=unl_prepare(unl_handle);
  if (k!=0) {
    return 1;
  }
  fieldsnum=unl_getfieldsnum(unl_handle);
  k=unl_open(unl_handle);
  if (k!=0) {
    return 1;
  }
  while (1) {
    k=unl_getrecord(unl_handle, &all_records);
    if (k!=0) {
      return 1;
    }
    if (all_records) {
      break;
    }
    if (unl_beforerow(unl_handle)) {
      return 1;
    }
    i=0; /* current field number */

    while (i<fieldsnum) {
      k=unl_getfield(unl_handle, i, buffer, sizeof(buffer), &bytesread, &sqlind);
      if (k!=0) {
        return 1;
      }
      k=unl_writefield(unl_handle, i, buffer, bytesread, sqlind,
                       unl_handle->fieldstates[i]);
      if (k!=0) {
        return 1;
      }
      i++;
    }

    if (i>=fieldsnum) {
      if (unl_afterrow(unl_handle)) {
        return 1;
      }
    }
  }
  if (unl_close(unl_handle)) {
    return 1;
  }
  if (unl_free(unl_handle)) { /* = loader_unprepare */
    return 1;
  }
  unl_handle=NULL;
  return 0; 
}

