/******************************************************************************
*
* Standard event handlers for unloading functions
*
* Author: Robert Sosnowski, Robert.Sosnowski@bzwbk.pl
*
* Version 1.70, 17-10-2005
*
* Copyright (c) 2001 - 2007. All Rights Reserved.
*
* Permission granted to use & distribute for free. Changes in source code
* require Author's acceptance.
*
* This software can be used only at your own risk. Author is not responsible 
* for any damage caused by using this software.
* 
******************************************************************************/

#include "load2cfg.h"
#include "tempfile.h"

#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_CTYPE_H
#include <ctype.h>
#endif

#include <stdio.h>
#include "unload2.h"
#include <stdlib.h>

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_ZLIB
#include <zlib.h>
#endif

EXEC SQL include sqltypes;

static char *id="@(#) $Id: unload2gz.ec 2521 2007-11-21 08:55:56Z sosnor01 $";
static const char hexdigit[] = "0123456789abcdef";

void oncleanupgzu(char *fname, void *handle, const void *logfile,
                  onlogerror_upt onlogerror, int verbose,
                  const char *datahandleoptions)
{
  char errstr[ERR_BUF_SIZE];
  gzFile fd1;
  unsigned int stat;
  
  fd1=handle;
  if (fd1!=NULL) {
    stat=gzclose((gzFile)handle);
    if (stat==EOF) {
      sprintf(errstr, "gzOS error %d",errno);
      onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, 0, 0, errno,
                 logfile, verbose, 10);
      return;
    }
  }
}


int oncrlfgzu(void *fd, char *filename, unsigned long int rownum,
              const void *logfile, onlogerror_upt onlogerror, int verbose,
              const char *datahandleoptions)
{
  char errstr[ERR_BUF_SIZE];
  gzFile *fd1;
  unsigned int stat;

  fd1=(gzFile *)fd;
  stat=gzputc(fd1,'\n');
  if (stat==EOF) {
    sprintf(errstr, "gzOS error %d",errno);
    onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, rownum, rownum,
               errno, logfile, verbose, 10);
    return 1;
  }
  return 0;
}

int oninitgzu(char *fname, const char *tablename, int batchsize, void **handle,
              char *escapech, const void *logfile, int fieldsnum, char *slctstmt2,
              onlogerror_upt onlogerror, int verbose, char *datahandleoptions)
{
  char errstr[ERR_BUF_SIZE];
  FILE *fd;
  char options[10];

  strcpy(options,"wb");
  if (datahandleoptions && datahandleoptions[0]>='0'
      && datahandleoptions[0]<='9') {
    options[2]=datahandleoptions[0];
    options[3]=0;
  }


  fd=(FILE *)logfile;

  if (fname[0]==0) {
    if (strcmp("<table>",tablename)) {
      sprintf(errstr,"file=<stdout>, tablename=%s", tablename);
    } else {
      sprintf(errstr,"file=<stdout>\n");
    }
    (*handle)=gzdopen(1,options);
    if (verbose>=10) {
      onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, 0, 0, errno, logfile,
                 verbose, 30);
    }
   } else {
    if (strcmp("<table>",tablename)) {
      sprintf(errstr, "file=%s, tablename=%s\n",fname, tablename);
    } else {
      sprintf(errstr, "file=%s\n",fname);
    }
    (*handle) = gzopen(fname,options);    
    if (verbose>=10) {
      onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, 0, 0, errno, logfile,
                 verbose, 30);
    }
  }
  if ( (*handle) == NULL ) {
    sprintf(errstr,"Cannot open file %s:%d", fname, errno);
    onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, 0, 0, errno, logfile,
               verbose, 10);
    return 1;
  }
  (*escapech)='\\';
  return 0;
}


int onwritetextblobgz(gzFile fd1, char *filename, loc_t * src,
                      const unsigned int buflen,
                      char delim, char escapech, unsigned long int rownum,
                      int fieldnum, char *fieldname, const void *logfile,
                      onlogerror_upt onlogerror, int verbose)
{ char * p;
  char buf[MAXMEMBLOB+1];  /* size enough for max size SQL datatype (non-BLOB)
                         in Informix it is CHAR(32767) */
  unsigned int j, k, len;
  FILE *fd2;
  char errstr[ERR_BUF_SIZE];
  unsigned int stat;
  char ch;
  int sqlind;
  unsigned int maxblob;
  char *envp;

  maxblob=MAXMEMBLOB;
#if defined(HAVE_GETENV)
  envp=getenv("DBBLOBBUF");
  if (envp!=NULL && isdigit(envp[0])) {
    k=atol(envp);
    if (k*1024<MAXMEMBLOB) {
      maxblob=k*1024;
    }
  }
#endif
  p=src->loc_buffer;
  sqlind=src->loc_indicator;
  if (sqlind!=-1) {
    if (p[0]=='\0' && src->loc_position==1) { /* if we have an empty string */
      buf[0]=escapech; /* empty string written as '\ ' */
      buf[1]=' ';
      buf[2]='\0';
      stat=gzputs(fd1, buf);
      if (stat==EOF) {
        sprintf(errstr, "gzOS error %d",errno);
        onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, rownum, rownum,
                   errno, logfile, verbose, 10);
        return 1;
      }
    } else {
      if (src->loc_position==1) {
        len = src->loc_size;
        j = 0;
        while ( j < len) {
          k=0;
          while (k<maxblob-2 && j<len) {
            ch=p[j];
            if (ch == '\n' || ch == escapech || ch == delim){
              buf[k]=escapech;
              k++;
            }
            buf[k]=ch;
            j++;
            k++;
          }
          stat=gzwrite(fd1, buf, k);
          if (stat==0 && k!=0) {
            sprintf(errstr, "gzOS error %d",errno);
            onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, rownum, rownum,
                       errno, logfile, verbose, 10);
            return 1;
          }
        }
      } else {
        fd2=fopen(p,"rb");
        if ( fd2 == NULL ) {
          sprintf(errstr,"Cannot open file %s:%d", p, errno);
          onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, 0, 0, errno,
                     logfile, verbose, 10);
          return 1;
        }
        while (!feof(fd2)) {
          k=0;
          while (k<maxblob-2) {
            ch=fgetc(fd2);
            if ( ferror(fd2) ) {
              sprintf(errstr,"Cannot read temp file %s:%d", p, errno);
              onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, 0, 0, errno,
                         logfile, verbose, 10);
              return 1;
            }
            if (feof(fd2)) {
              break;
            }
            if (ch == '\n' || ch == escapech || ch == delim){
              buf[k]=escapech;
              k++;
            }
            buf[k]=ch;
            k++;
          }
          stat=gzwrite(fd1, buf, k);
          if (stat==0 && k!=0) {
            sprintf(errstr, "gzOS error %d",errno);
            onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, rownum, rownum,
                       errno, logfile, verbose, 10);
            return 1;
          }
        }
        fclose(fd2);
      }
    }
  }
  if (src->loc_buffer!=NULL && sqlind!=-1) {
    if (src->loc_position==2) {
      unlink(p);
    }
    free(src->loc_buffer);
    src->loc_buffer=NULL;
  }
  return 0;
}

int onwritebyteblobgz(gzFile fd1, char *filename, loc_t * src,
                      const unsigned int buflen,
                      char delim, unsigned long int rownum,
                      int fieldnum, char *fieldname, const void *logfile,
                      onlogerror_upt onlogerror, int verbose)
{ char * p;
  char buf[MAXMEMBLOB+1];  /* size enough for max size SQL datatype (non-BLOB)
                         in Informix it is CHAR(32767) */
  unsigned int i, j, k, len;
  FILE *fd2;
  char errstr[ERR_BUF_SIZE];
  char ch;
  int j2;
  int sqlind;
  unsigned int maxblob;
  char *envp;

  maxblob=MAXMEMBLOB;
#if defined(HAVE_GETENV)
  envp=getenv("DBBLOBBUF");
  if (envp!=NULL && isdigit(envp[0])) {
    k=atol(envp);
    if (k*1024<MAXMEMBLOB) {
      maxblob=k*1024;
    }
  }
#endif
  p=src->loc_buffer;
  sqlind=src->loc_indicator;
  if (sqlind!=-1) {
    if (src->loc_position==1) {
      len = src->loc_size;
      j=0;
      while (j<len) {
        i=0;
        while (i<=maxblob && j<len) {
          ch=p[j];
          buf[i]=hexdigit[(ch>>4) & 0xF];
          i++;
          buf[i]=hexdigit[ch & 0xF];
          i++;
          j++;
        }
        k=gzwrite(fd1, buf, i);
        if (errno) {
          sprintf(errstr, "gzOS error while writing out file: %d",errno);
          onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, rownum, rownum,
                     errno, logfile, verbose, 10);
          return 1;
        }
      }
    } else {
      fd2=fopen(p,"rb");
      if ( fd2 == NULL ) {
        sprintf(errstr,"Cannot open file %s:%d", p, errno);
        onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, 0, 0, errno, logfile,
                   verbose, 10);
        return 1;
      }
      while (1) {
        i=fread(buf, 1, (maxblob-1)/2, fd2);
        if (ferror(fd2)) {
          sprintf(errstr, "OS error while reading temp file: %d:%d",errno, ferror(fd2));
          onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, rownum, rownum,
                     errno, logfile, verbose, 10);
          return 1;
        }
        j2=i-1;
        while (j2>=0) {
          buf[2*j2+1]=hexdigit[(buf[j2] & 0xF)];
          buf[2*j2]=hexdigit[((buf[j2]>>4) & 0xF)];
          j2--;
        }
        j=gzwrite(fd1, buf, 2*i); 
        if (errno) {
          sprintf(errstr, "gzOS error while writing out file:%d",errno);
          onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, rownum, rownum,
                     errno, logfile, verbose, 10);
          return 1;
        }
        if (feof(fd2)) {
          break;
        }
      }
      fclose(fd2);
    }
  }
  if (src->loc_buffer!=NULL && sqlind!=-1) {
/* additional check for sqlind!=-1 because it appeared
   that in many cases loc_buffer was not null altough memory pointed to
   was never allocated; therefore free caused memory leakage;
   fortunately all such cases occurs only for null BLOB */

    if (src->loc_position==2) {
      unlink(p);
    }
    free(src->loc_buffer);
    src->loc_buffer=NULL;
  }
  return 0;
}

int onwritesmartblobgz(gzFile fd1, char *filename, int4 sqlxid, ifx_lo_t * src,
                       char delim, unsigned long int rownum, int fieldnum,
                       char *fieldname, slob_handle_t *fieldstate,
                       const void *logfile,
                       onlogerror_upt onlogerror, int verbose)
{
  char buf[MAXMEMBLOB+1];  /* size enough for max size SQL datatype (non-BLOB)
                         in Informix it is CHAR(32767) */
  char errstr[ERR_BUF_SIZE];
  mint lofd;
  mint error;
  ifx_lo_stat_t *stats;
  char result[256];
  ifx_int8_t fsize64;
  char fsizeasc[18];
  gzFile f;
  mint n;
  unsigned int stat;

  lofd=ifx_lo_open((ifx_lo_t *)src, LO_RDONLY, &error);
  if (lofd<0) {
    sprintf(errstr, "ifx_lo_open: %d,"INT4_FORMAT",%s",
                    error,sqlca.sqlerrd[1], sqlca.sqlerrm);
    onlogerror(__FILE__, __LINE__, errstr, error, sqlca.sqlerrd[1],
               sqlca.sqlerrm, rownum, rownum, 0, logfile, verbose, 10);
    return 1;
  }
  error=ifx_lo_stat(lofd, &stats);
  if (error<0) {
    sprintf(errstr, "ifx_lo_stat: %d,"INT4_FORMAT",%s",
            error, sqlca.sqlerrd[1], sqlca.sqlerrm);
    onlogerror(__FILE__, __LINE__, errstr, error, sqlca.sqlerrd[1],
               sqlca.sqlerrm, rownum, rownum, 0, logfile, verbose, 10);
    return 1;
  }
  error=ifx_lo_stat_size(stats,&fsize64);
  if (error<0) {
    sprintf(errstr, "ifx_lo_stat_size: %d,"INT4_FORMAT",%s",
            error, sqlca.sqlerrd[1], sqlca.sqlerrm);
    onlogerror(__FILE__, __LINE__, errstr, error, sqlca.sqlerrd[1],
               sqlca.sqlerrm, rownum, rownum, 0, logfile, verbose, 10);
    return 1;
  }
  if (fieldstate->filename[0]==0) {
    if (sqlxid==XID_BLOB) {
      error=ifx_lo_filename((ifx_lo_t *)src, "blob????.???.gz", result, sizeof(result));
    } else {
      error=ifx_lo_filename((ifx_lo_t *)src, "clob????.???.gz", result, sizeof(result));
    }
    if (error<0) {
      sprintf(errstr, "ifx_lo_filename: %d,"INT4_FORMAT",%s",
              error, sqlca.sqlerrd[1], sqlca.sqlerrm);
      onlogerror(__FILE__, __LINE__, errstr, error, sqlca.sqlerrd[1], sqlca.sqlerrm,
                 rownum, rownum, 0, logfile, verbose, 10);
      return 1;
    }
    strncpy(fieldstate->filename, result, 255);
    fieldstate->offset=0;
    f=gzopen(result, "wb");
  } else {
    strncpy(result, fieldstate->filename, 255);
    f=gzopen(result, "ab");
  }
  if ( f == NULL ) {
    sprintf(errstr,"Cannot open file '%s:%d'", result, errno);
    onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, 0, 0, errno,
               logfile, verbose, 10);
    return 1;
  }
  while ((n=ifx_lo_read(lofd, buf, MAXMEMBLOB, &error))>0) {
    stat=gzwrite(f, buf, n);
    if (stat==0 && n!=0) {
      sprintf(errstr, "gzOS error %d", errno);
      onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, rownum, rownum,
                 errno, logfile, verbose, 10);
      return 1;
    }
  }
  if (n<0) {
    sprintf(errstr, "ifx_lo_read: %d, "INT4_FORMAT",%s",
            error, sqlca.sqlerrd[1], sqlca.sqlerrm);
    onlogerror(__FILE__, __LINE__, errstr, error, sqlca.sqlerrd[1],
               sqlca.sqlerrm, rownum, rownum, 0, logfile, verbose, 10);
    return 1;
  }
  error=ifx_int8toasc(&fsize64,fsizeasc, sizeof(fsizeasc)-1);
  if (error<0) {
    sprintf(errstr, "ifx_int8toasc: %d",error);
    onlogerror(__FILE__, __LINE__, errstr, error, 0, NULL, rownum, rownum,
               0, logfile, verbose, 10);
    return 1;
  }
  fsizeasc[sizeof(fsizeasc)-1]=0;
  fieldstate->size=(unsigned long)atol(fsizeasc);
  fieldstate->offset=fieldstate->offset + fieldstate->size;
  gzclose(f);
  ifx_lo_close(lofd);
  ifx_lo_stat_free(stats);

  return 0;
}


int onwritefieldgz(void *fd, int sqltype, int4 sqlxid, char *filename,
                   char *src, const unsigned int buflen,
                   char delim, char escapech, short int sqlind,
                   unsigned long int rownum,  int fieldnum,
                   char * fieldname, const void *logfile,
                   onlogerror_upt onlogerror, int verbose,
                   const char *datahandleoptions,
                   void *fieldstate)
/* returns:  0 if success
 *           1 if error;
 * src : pointer to the buffer;
 */
{
  char buf[BUF1_SIZE+1];  /* size enough for max size SQL datatype (non-BLOB)
                         in Informix it is CHAR(32767) */

  unsigned int j, k, len;
  gzFile fd1;
  char errstr[ERR_BUF_SIZE];
  unsigned int stat;
  char ch;
  slob_handle_t *slob_h;

  fd1=(gzFile )fd;
  if (sqltype==SQLTEXT || sqltype==SQLBYTES ) {
    if (sqltype==SQLTEXT) {
      stat=onwritetextblobgz(fd1, filename, (loc_t *) src, buflen, delim,
                             escapech, rownum, fieldnum, fieldname, logfile,
                             onlogerror, verbose);
    } else {
      stat=onwritebyteblobgz(fd1, filename, (loc_t *) src, buflen, delim,
                             rownum, fieldnum, fieldname, logfile, onlogerror,
                             verbose);
    }
    if (stat!=0) {
      return 1;
    }
    stat=gzputc(fd1, delim);
    if (stat==EOF) {
      sprintf(errstr, "gzOS error %d",errno);
      onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, rownum, rownum,
                 errno, logfile, verbose, 10);
      return 1;
    }
    return 0;
  }
  if (sqltype==CFIXBINTYPE && (sqlxid==XID_BLOB || sqlxid==XID_CLOB)
  && sqlind!=-1 ) {
    slob_h=(slob_handle_t *)fieldstate;
    stat=gzprintf(fd1, "%s,", off_t2hex(slob_h->offset));

    if (stat==0) {
      sprintf(errstr, "gzOS error %d",errno);
      onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, rownum, rownum,
                 errno, logfile, verbose, 10);
      return 1;
    }
    stat=onwritesmartblobgz(fd1, filename, sqlxid, (ifx_lo_t *) src,
                            escapech, rownum, fieldnum, fieldname,
                            slob_h, logfile, onlogerror, verbose);
    if (stat!=0) {
      return 1;
    }
    stat=gzprintf(fd1, "%s,%s%c", off_t2hex(slob_h->size), slob_h->filename, delim);

    if (stat==EOF) {
      sprintf(errstr, "gzOS error %d",errno);
      onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, rownum, rownum,
                 errno, logfile, verbose, 10);
      return 1;
    }
    return 0;
  }
  if (sqlind != -1 && src[0]=='\0' /* if we have an empty string */
  && (sqltype == CCHARTYPE ||
      sqltype == CVCHARTYPE || sqltype == CLVCHARTYPE)) {
    buf[0]=escapech; /* empty string written as '\ ' */
    buf[1]=' ';
    buf[2]='\0';
    stat=gzputs(fd1, buf);
    if (stat==EOF) {
      sprintf(errstr, "OS error %d",errno);
      onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, rownum, rownum,
                 errno, logfile, verbose, 10);
      return 1;
    }
  } else {
    len = stleng(src);
    j = 0;
    while ( j < len) {
      k=0;
      while (k<BUF1_SIZE-2 && j<len) {
        ch=src[j];
        if (ch == '\n' || ch == escapech || ch == delim){
          buf[k]=escapech;
          k++;
        }
        buf[k]=ch;
        j++;
        k++;
      }
      buf[k]='\0';
      stat=gzputs(fd1, buf);
      if (stat==EOF) {
        sprintf(errstr, "gzOS error %d",errno);
        onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, rownum, rownum,
                   errno, logfile, verbose, 10);
        return 1;
      }
    }
  }
  stat=gzputc(fd1, delim);
  if (stat==EOF) {
    sprintf(errstr, "gzOS error %d",errno);
    onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, rownum, rownum,
               errno, logfile, verbose, 10);
    return 1;
  }
  return 0;
}

