/******************************************************************************
*
* Standard event handlers for unloading functions
*
* Author: Robert Sosnowski, Robert.Sosnowski@bzwbk.pl
*
* Version 1.70, 17-10-2005
*
* Copyright (c) 2001 - 2007. All Rights Reserved.
*
* Permission granted to use & distribute for free. Changes in source code
* require Author's acceptance.
*
* This software can be used only at your own risk. Author is not responsible 
* for any damage caused by using this software.
* 
******************************************************************************/

#include "load2cfg.h"
#include "tempfile.h"

#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_CTYPE_H
#include <ctype.h>
#endif

#include <stdio.h>
#include "unload2.h"
#include <stdlib.h>

#ifdef HAVE_STRING_H
#include <string.h>
#endif

EXEC SQL include sqltypes;

static char *id="@(#) $Id: unload2h.ec 2521 2007-11-21 08:55:56Z sosnor01 $";
static const char hexdigit[] = "0123456789abcdef";

int onlogerror_stdu(const char * srcfile, const int pline,
                    const char *errstr, const long int sqlcode,
                    const long int isamcode, const char *ifx_str,
                    const unsigned long int rownum,
                    const unsigned long int processed,
                    const int os_errno, const void *logfile,
                    const int verbosity, const int urgency);

int oninit_stdu(char *fname, const char *tablename, int batchsize,
                void **handle, char *escapech, const void *logfile,
                int fieldsnum, char *slctstmt2,
                onlogerror_upt onlogerror, int verbose,
                const char *datahandleoptions);
int onfinish_stdu(char *fname, char *tablename, void *handle,
                  unsigned long int rownum, const void *logfile,
                  onlogerror_upt onlogerror, int verbose);
int onwritefield_stdu(void *fd, int sqltype, int4 sqlxid, char *filename,
                      char *src, const unsigned int buflen,
                      char delim, char escapech, short int sqlind,
                      unsigned long int rownum,  int fieldnum,
                      char *fieldname, const void *logfile,
                      onlogerror_upt onlogerror, int verbose,
                      const char *datahandleoptions,
                      void *fieldstate);
int onbatch_stdu(char *fname, const char *tablename, void *handle,
                 unsigned long int rownum,
                 const void *logfile,
                 onlogerror_upt onlogerror, int verbose);
void oncleanup_stdu(char *fname, void *handle, const void *logfile,
                    onlogerror_upt onlogerror, int verbose,
                    const char *datahandleoptions);
int oncrlf_stdu(void *fd, char *filename, unsigned long int rownum,
                const void *logfile,
                onlogerror_upt onlogerror, int verbose,
                const char *datahandleoptions);
int onrecordbegin_stdu(void *fd, char *filename, unsigned long int rownum,
                       const void *logfile,
                       onlogerror_upt onlogerror, int verbose,
                       const char *datahandleoptions);
int onblobfree_stdu(loc_t *simple_lo, const void *logfile,
                    onlogerror_upt onlogerror, int verbose);


int onblobfree_stdu(loc_t *simple_lo, const void *logfile,
                   onlogerror_upt onlogerror, int verbose)
/* this procedure is currently not used; blob is freed 
   inside onwritetextblob or onwritebyteblob;
   should always be used in cleanup (in future) */
{
  if (simple_lo->loc_loctype==LOCMEMORY) {
    if (simple_lo->loc_buffer!=NULL) {
      free(simple_lo->loc_buffer);
    }
    simple_lo->loc_buffer=NULL;
    simple_lo->loc_size=0;
    simple_lo->loc_bufsize=0;
  }
  if (simple_lo->loc_loctype==LOCFNAME) {
    unlink(simple_lo->loc_fname);
    free(simple_lo->loc_fname);
    simple_lo->loc_fname=NULL;
  }
  return 0;
}

mint openit(loc_t *loc, mint flag, mint bsize)
/* Parameter bsize is declared in Informix's header file
   However it is undocumented even in doc for IDS 9.40 (CSDK 2.81)!
   Is it reliable value? */
{ void *p;
  char *p1;
  FILE *ft;
  unl_handle_t *h;
  unsigned int maxblob, k;
  char *envp;

  maxblob=MAXMEMBLOB;
#if defined(HAVE_GETENV)
  envp=getenv("DBBLOBBUF");
  if (envp!=NULL && isdigit(envp[0])) {
    k=atol(envp);
    if (k*1024<MAXMEMBLOB) {
      maxblob=k*1024;
    }
  }
#endif
  h=(unl_handle_t *)loc->loc_user_env;
  loc->loc_xfercount = 0L;
  loc->loc_status=0;
  loc->loc_buffer=NULL;
  if ( (flag !=LOC_WONLY)) {
    sprintf(h->errstr, "wrong open flags for BLOB");
    h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, NULL, h->rownum,
                  h->rownum, 0, h->logfile, h->verbosity, 10);
    return -1;
  }
  loc->loc_position=2; /* it means in this case: stored in named file */

  if ((unsigned long)bsize<=maxblob) {
    loc->loc_position=1; /* it means in this case: stored in memory */
    if (bsize==0) {
      loc->loc_bufsize=0;
    } else {
      p=malloc(bsize+1);
      ((char *)p)[bsize]=0;
      if (p!=NULL) {
        loc->loc_buffer=p;
        loc->loc_bufsize=bsize+1;
      } else {
        loc->loc_position=2;
      }
    }
  }
  if (loc->loc_position==2) {
    /* in named file */
    p1=creatmpfile();
    if (p1==NULL) {
      sprintf(h->errstr, "cannot create temporary file: %d", errno);
      h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, NULL, h->rownum,
                  h->rownum, errno, h->logfile, h->verbosity, 10);
      return -1;
    }
    p=malloc(strlen(p1)+1);
    if (p==NULL) {
      sprintf(h->errstr, "cannot allocate memory for temp file name");
      h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, NULL, h->rownum,
                  h->rownum, 0, h->logfile, h->verbosity, 10);
      return -1;
    }
    ft=fopen(p1,"wb");
    if (ft==NULL) {
      free(p);
      sprintf(h->errstr, "cannot open temporary file for writing");
      h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, NULL, h->rownum,
                    h->rownum, 0, h->logfile, h->verbosity, 10);
      return -1;
    }
    strcpy((char *)p,p1);
    h->tmp_blob_file=(void *)ft;
    loc->loc_buffer=p;
    loc->loc_bufsize=strlen(p1)+1;
  }
  return 0; /* -1 = failure */
}

mint closeit(loc_t *loc)
{ FILE *ft;
  unl_handle_t *h;

  h=(unl_handle_t *)loc->loc_user_env;
  loc->loc_status=0;
  if (loc->loc_oflags & LOC_WONLY) {
    loc->loc_indicator=0;
    loc->loc_size=loc->loc_xfercount;
    if (loc->loc_position==2) {
      ft=(FILE *)h->tmp_blob_file;
      fclose(ft);
      h->tmp_blob_file=NULL;
      /* loc->loc_loctype=LOCMEMORY; */
    }
/*    if (loc->loc_position==1) {
      load_blob->loc_loctype=LOCFNAME;
      load_blob->loc_size=-1;
      load_blob->loc_fname=loc->locbuffer;
      loc->locbuffer=NULL;
    } */
  }
  return 0;
}

mint writeit(loc_t *loc, char *buffer, mint bufflen)
{ FILE *ft;
  int i;
  unl_handle_t *h;
  char *p1, *p;

  p=NULL;
  h=(unl_handle_t *)loc->loc_user_env;
  if (bufflen==0) {
    return 0;
  }
  if ((loc->loc_xfercount+bufflen>loc->loc_bufsize) && loc->loc_position==1) {
    loc->loc_position=2;
    p1=creatmpfile();
    if (p1==NULL) {
      sprintf(h->errstr, "cannot create temporary file: %d", errno);
      h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, NULL, h->rownum,
                  h->rownum, errno, h->logfile, h->verbosity, 10);
      return -1;
    }
    p=(char *)malloc(strlen(p1)+1);
    if (p==NULL) {
      sprintf(h->errstr, "cannot allocate memory for temp file name");
      h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, NULL, h->rownum,
                  h->rownum, 0, h->logfile, h->verbosity, 10);
      return -1;
    }
    ft=fopen(p1,"wb");
    if (ft==NULL) {
      free(p);
      sprintf(h->errstr, "cannot open temporary file for writing");
      h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, NULL, h->rownum,
                    h->rownum, 0, h->logfile, h->verbosity, 10);
      return -1;
    }
    strcpy((char *)p,p1);
    h->tmp_blob_file=(void *)ft;
    i=fwrite(loc->loc_buffer, 1, loc->loc_xfercount, ft);
    if (ferror(ft)) {
      free(p);
      sprintf(h->errstr, "cannot write to temporary file");
      h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, NULL, h->rownum,
                    h->rownum, 0, h->logfile, h->verbosity, 10);
      return -1;
    }
    free(loc->loc_buffer);
    loc->loc_buffer=p;
    loc->loc_bufsize=strlen(p1)+1;

  }
  i=loc->loc_xfercount;
  if (loc->loc_position==1) {
    bycopy(buffer, (void *)((char *)(loc->loc_buffer)+i), bufflen);
  } else {
    ft=(FILE *)h->tmp_blob_file;
    i=fwrite(buffer, 1, bufflen, ft);
    if (ferror(ft)) {
      free(p);
      sprintf(h->errstr, "cannot write to temporary file");
      h->onlogerror(__FILE__, __LINE__, h->errstr, 0, 0, NULL, h->rownum,
                    h->rownum, 0, h->logfile, h->verbosity, 10);
      return -1;
    }
  }
/*  loc->loc_xfercount+=bufflen; */
  return bufflen;  /* number of bytes written or -1 */
}

int onbatch_stdu(char *fname, const char *tablename, void *handle,
                 unsigned long int rownum, const void *logfile,
                 onlogerror_upt onlogerror, int verbose)
{ FILE *fd;
  char errstr[ERR_BUF_SIZE];

  fd=(FILE *) logfile;
  sprintf(errstr, "Unloaded succesfully %ld records", rownum);
  onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, rownum,
                rownum, 0, logfile, verbose, 40);
  return 0;
}

void oncleanup_stdu(char *fname, void *handle, const void *logfile,
                    onlogerror_upt onlogerror, int verbose,
                    const char *datahandleoptions)
{
  char errstr[ERR_BUF_SIZE];
  FILE * fd1;
  unsigned int stat;
  
  fd1=handle;
  if (fd1!=NULL && fd1!=(void *)stderr && fd1!=(void *)stdout) {
    stat=fclose((FILE *)handle);
    if (stat==EOF) {
      sprintf(errstr, "OS error %d:%d",errno, ferror(fd1));
      onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, 0, 0, errno,
                 logfile, verbose, 10);
      return;
    }
  }
}


int oncrlf_stdu(void *fd, char *filename, unsigned long int rownum,
                const void *logfile, onlogerror_upt onlogerror, int verbose,
                const char *datahandleoptions)
{
  char errstr[ERR_BUF_SIZE];
  FILE *fd1;
  unsigned int stat;

  fd1=(FILE *)fd;
  stat=putc('\n',fd1);
  if (stat==EOF) {
    sprintf(errstr, "OS error %d:%d",errno, ferror(fd1));
    onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, rownum, rownum,
               errno, logfile, verbose, 10);
    return 1;
  }
  return 0;
}

int onlogerror_stdu(const char *srcfile, const int pline, const char *errstr,
                    const long int sqlcode, const long int isamcode,
                    const char *ifx_str, const unsigned long int rownum,
                    const unsigned long int processed, const int os_errno,
                    const void *logfile, const int verbosity,
                    const int urgency)
{ FILE *fd;

  fd=(FILE *) logfile;
  if (urgency<=verbosity && verbosity!=0) {
    if (fd!=NULL) {
      fprintf(fd, errstr);
      fprintf(fd, "\n");
      if (urgency>=30) {
        /* notification messages */
        fprintf(stderr, errstr);
        fprintf(stderr, "\n");
      } else {
        /* error messages */
        fprintf(stderr, errstr);
        fprintf(stderr, "\n");
        if (verbosity>=50) {
          fprintf(stderr, "File: %s, line:%d\n",srcfile, pline);
        }
      }
    } else {
      fprintf(stderr, errstr);
      fprintf(stderr, "\n");
      if (verbosity>=50 && urgency<30) {
        fprintf(stderr, "File: %s, line:%d\n",srcfile, pline);
      }
    }
  }
  return 0;
}

int oninit_stdu(char *fname, const char *tablename, int batchsize,
                void **handle, char *escapech, const void *logfile,
                int fieldsnum, char *slctstmt2,
                onlogerror_upt onlogerror, int verbose,
                const char *datahandleoptions)
{
  char errstr[ERR_BUF_SIZE];
  FILE *fd;

  fd=(FILE *)logfile;


  if (fname[0]==0) {
    (*handle)=stdout;
    if (strcmp("<table>",tablename)) {
      sprintf(errstr,"file=<stdout>, tablename=%s", tablename);
    } else {
      sprintf(errstr,"file=<stdout>\n");
    }
    if (verbose>=10) {
      onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, 0, 0, errno, logfile,
                 verbose, 30);
    }
  } else {
    if (strcmp("<table>",tablename)) {
      sprintf(errstr, "file=%s, tablename=%s\n",fname, tablename);
    } else {
      sprintf(errstr, "file=%s\n",fname);
    }
#ifdef NOBINARY_FLAG
    (*handle) = fopen(fname,"w");
    if (verbose>=10) {
      onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, 0, 0, errno, logfile,
                 verbose, 30);
    }
#else
    (*handle) = fopen(fname,"wb");
    if (verbose>=10) {
      onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, 0, 0, errno, logfile,
                 verbose, 30);
    }
#endif
  }
  if ( (*handle) == NULL ) {
    sprintf(errstr,"Cannot open file %s:%d", fname, errno);
    onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, 0, 0, errno, logfile,
               verbose, 10);
    return 1;
  }
  (*escapech)='\\';
  return 0;
}


int onwritetextblob(FILE *fd1, char *filename, loc_t * src,
                    char delim, char escapech, unsigned long int rownum,
                    int fieldnum, char *fieldname, const void *logfile,
                    onlogerror_upt onlogerror, int verbose)
{ char * p;
  char buf[MAXMEMBLOB+1];  /* size enough for max size SQL datatype (non-BLOB)
                         in Informix it is CHAR(32767) */
  unsigned int j, k, len;
  FILE *fd2;
  char errstr[ERR_BUF_SIZE];
  unsigned int stat;
  char ch;
  int sqlind;
  unsigned int maxblob;
  char *envp;

  maxblob=MAXMEMBLOB;
#if defined(HAVE_GETENV)
  envp=getenv("DBBLOBBUF");
  if (envp!=NULL && isdigit(envp[0])) {
    k=atol(envp);
    if (k*1024<MAXMEMBLOB) {
      maxblob=k*1024;
    }
  }
#endif

  p=src->loc_buffer;
  sqlind=src->loc_indicator;
  if (sqlind!=-1) {
    if (p[0]=='\0' && src->loc_position==1) { /* if we have an empty string */
      buf[0]=escapech; /* empty string written as '\ ' */
      buf[1]=' ';
      buf[2]='\0';
      stat=fputs(buf, fd1);
      if (stat==EOF) {
        sprintf(errstr, "OS error %d:%d",errno, ferror(fd1));
        onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, rownum, rownum,
                   errno, logfile, verbose, 10);
        return 1;
      }
    } else {
      if (src->loc_position==1) {
        len = src->loc_size;
        j = 0;
        while ( j < len) {
          k=0;
          while (k<maxblob-2 && j<len) {
            ch=p[j];
            if (ch == '\n' || ch == escapech || ch == delim){
              buf[k]=escapech;
              k++;
            }
            buf[k]=ch;
            j++;
            k++;
          }
          stat=fwrite(buf, 1, k, fd1);
          if ( ferror(fd1) ) {
            sprintf(errstr, "OS error %d:%d",errno, ferror(fd1));
            onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, rownum, rownum,
                       errno, logfile, verbose, 10);
            return 1;
          }
        }
      } else {
        fd2=fopen(p,"rb");
        if ( fd2 == NULL ) {
          sprintf(errstr,"Cannot open file %s:%d", p, errno);
          onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, 0, 0, errno,
                     logfile, verbose, 10);
          return 1;
        }
        while (!feof(fd2)) {
          k=0;
          while (k<maxblob-2) {
            ch=fgetc(fd2);
            if ( ferror(fd2) ) {
              sprintf(errstr,"Cannot read temp file %s:%d", p, errno);
              onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, 0, 0, errno,
                         logfile, verbose, 10);
              return 1;
            }
            if (feof(fd2)) {
              break;
            }
            if (ch == '\n' || ch == escapech || ch == delim){
              buf[k]=escapech;
              k++;
            }
            buf[k]=ch;
            k++;
          }
          j=fwrite(buf, 1, k, fd1);
          if ( ferror(fd1) ) {
            sprintf(errstr, "OS error %d:%d",errno, ferror(fd1));
            onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, rownum, rownum,
                       errno, logfile, verbose, 10);
            return 1;
          }
        }
        fclose(fd2);
      }
    }
  }
  if (src->loc_buffer!=NULL && sqlind!=-1) {
    if (src->loc_position==2) {
      unlink(p);
    }
    free(src->loc_buffer);
    src->loc_buffer=NULL;
  }
  return 0;
}

int onwritebyteblob(FILE *fd1, char *filename, loc_t * src,
                    char delim, unsigned long int rownum,
                    int fieldnum, char *fieldname, const void *logfile,
                    onlogerror_upt onlogerror, int verbose)
{ char * p;
  char buf[MAXMEMBLOB+1];  /* size enough for max size SQL datatype (non-BLOB)
                         in Informix it is CHAR(32767) */
  unsigned int i, j, k, len;
  FILE *fd2;
  char errstr[ERR_BUF_SIZE];
  char ch;
  int j2;
  int sqlind;
  unsigned int maxblob;
  char *envp;

  maxblob=MAXMEMBLOB;
#if defined(HAVE_GETENV)
  envp=getenv("DBBLOBBUF");
  if (envp!=NULL && isdigit(envp[0])) {
    k=atol(envp);
    if (k*1024<MAXMEMBLOB) {
      maxblob=k*1024;
    }
  }
#endif

  p=src->loc_buffer;
  sqlind=src->loc_indicator;
  if (sqlind!=-1) {
    if (src->loc_position==1) {
      len = src->loc_size;
      j=0;
      while (j<len) {
        i=0;
        while (i<=maxblob && j<len) {
          ch=p[j];
          buf[i]=hexdigit[(ch>>4) & 0xF];
          i++;
          buf[i]=hexdigit[ch & 0xF];
          i++;
          j++;
        }
        k=fwrite(buf, 1, i, fd1);
        if (ferror(fd1)) {
          sprintf(errstr, "OS error while writing out file: %d:%d",errno, ferror(fd1));
          onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, rownum, rownum,
                     errno, logfile, verbose, 10);
          return 1;
        }
      }
    } else {
      fd2=fopen(p,"rb");
      if ( fd2 == NULL ) {
        sprintf(errstr,"Cannot open file %s:%d", p, errno);
        onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, 0, 0, errno, logfile,
                   verbose, 10);
        return 1;
      }
      while (1) {
        i=fread(buf, 1, (maxblob-1)/2, fd2);
        if (ferror(fd2)) {
          sprintf(errstr, "OS error while reading temp file: %d:%d",errno, ferror(fd2));
          onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, rownum, rownum,
                     errno, logfile, verbose, 10);
          return 1;
        }
        j2=i-1;
        while (j2>=0) {
          buf[2*j2+1]=hexdigit[(buf[j2] & 0xF)];
          buf[2*j2]=hexdigit[((buf[j2]>>4) & 0xF)];
          j2--;
        }
        j=fwrite(buf,1,2*i, fd1); 
        if (ferror(fd1)) {
          sprintf(errstr, "OS error while writing out file:%d:%d",errno, ferror(fd1));
          onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, rownum, rownum,
                     errno, logfile, verbose, 10);
          return 1;
        }
        if (feof(fd2)) {
          break;
        }
      }
      fclose(fd2);
    }
  }
  if (src->loc_buffer!=NULL && sqlind!=-1) {
/* additional check for sqlind!=-1 because it appeared
   that in many cases loc_buffer was not null altough memory pointed to
   was never allocated; therefore free caused memory leakage;
   fortunately all such cases occurs only for null BLOB */

    if (src->loc_position==2) {
      unlink(p);
    }
    free(src->loc_buffer);
    src->loc_buffer=NULL;
  }
  return 0;
}

int onwritesmartblob(FILE *fd1, char *filename, int4 sqlxid, ifx_lo_t * src,
                    char delim, unsigned long int rownum, int fieldnum,
                    char *fieldname, slob_handle_t *fieldstate,
                    const void *logfile,
                    onlogerror_upt onlogerror, int verbose)
{
  char buf[MAXMEMBLOB+1];  /* size enough for max size SQL datatype (non-BLOB)
                         in Informix it is CHAR(32767) */
  char errstr[ERR_BUF_SIZE];
  mint lofd;
  mint error;
  ifx_lo_stat_t *stats;
  char result[256];
  ifx_int8_t fsize64;
  char fsizeasc[18];
  FILE *f;
  mint n;
  
  lofd=ifx_lo_open((ifx_lo_t *)src, LO_RDONLY, &error);
  if (lofd<0) {
    sprintf(errstr, "ifx_lo_open: %d,"INT4_FORMAT",%s",
                    error, sqlca.sqlerrd[1], sqlca.sqlerrm);
    onlogerror(__FILE__, __LINE__, errstr, error, sqlca.sqlerrd[1],
               sqlca.sqlerrm, rownum, rownum, 0, logfile, verbose, 10);
    return 1;
  }
  error=ifx_lo_stat(lofd, &stats);
  if (error<0) {
    sprintf(errstr, "ifx_lo_stat: %d,"INT4_FORMAT",%s",
                    error, sqlca.sqlerrd[1], sqlca.sqlerrm);
    onlogerror(__FILE__, __LINE__, errstr, error, sqlca.sqlerrd[1],
               sqlca.sqlerrm, rownum, rownum, 0, logfile, verbose, 10);
    return 1;
  }
  error=ifx_lo_stat_size(stats,&fsize64);
  if (error<0) {
    sprintf(errstr, "ifx_lo_stat_size: %d,"INT4_FORMAT",%s",
                    error, sqlca.sqlerrd[1], sqlca.sqlerrm);
    onlogerror(__FILE__, __LINE__, errstr, error, sqlca.sqlerrd[1],
               sqlca.sqlerrm, rownum, rownum, 0, logfile, verbose, 10);
    return 1;
  }
  if (fieldstate->filename[0]==0) {
    if (sqlxid==XID_BLOB) {
      error=ifx_lo_filename((ifx_lo_t *)src, "blob????.???", result, sizeof(result));
    } else {
      error=ifx_lo_filename((ifx_lo_t *)src, "clob????.???", result, sizeof(result));
    }
    if (error<0) {
      sprintf(errstr, "ifx_lo_filename: %d, "INT4_FORMAT",%s",
                      error, sqlca.sqlerrd[1], sqlca.sqlerrm);
      onlogerror(__FILE__, __LINE__, errstr, error, sqlca.sqlerrd[1],
                 sqlca.sqlerrm, rownum, rownum, 0, logfile, verbose, 10);
      return 1;
    }
    strncpy(fieldstate->filename, result, 255);
    fieldstate->offset=0;
    f=fopen(result, "wb");
  } else {
    strncpy(result, fieldstate->filename, 255);
    f=fopen(result, "ab");
  }
  if ( f == NULL ) {
    sprintf(errstr,"Cannot open file %s:%d", result, errno);
    onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, 0, 0, errno,
               logfile, verbose, 10);
    return 1;
  }
  while ((n=ifx_lo_read(lofd, buf, MAXMEMBLOB, &error))>0) {
    fwrite(buf, 1, n, f);
    if (ferror(f)) {
      sprintf(errstr, "OS error while writing out file:%d,"INT4_FORMAT",%s:%d",
              errno, sqlca.sqlerrd[1], sqlca.sqlerrm, ferror(f));
      onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, rownum, rownum,
                 errno, logfile, verbose, 10);
      return 1;
    }
  }
  if (n<0) {
    sprintf(errstr, "ifx_lo_read: %d,"INT4_FORMAT",%s",
                    error, sqlca.sqlerrd[1], sqlca.sqlerrm);
    onlogerror(__FILE__, __LINE__, errstr, error, sqlca.sqlerrd[1],
               sqlca.sqlerrm, rownum, rownum, 0, logfile, verbose, 10);
    return 1;
  }
  fieldstate->offset=ftell(f);
  error=ifx_int8toasc(&fsize64,fsizeasc, sizeof(fsizeasc)-1);
  if (error<0) {
    sprintf(errstr, "ifx_int8toasc: %d",error);
    onlogerror(__FILE__, __LINE__, errstr, error, 0, NULL, rownum, rownum,
               0, logfile, verbose, 10);
    return 1;
  }
  fsizeasc[sizeof(fsizeasc)-1]=0;
  fieldstate->size=(unsigned long)atol(fsizeasc);
  fclose(f);
  ifx_lo_close(lofd);
  ifx_lo_stat_free(stats);

  return 0;
}


int onwritefield_stdu(void *fd, int sqltype, int4 sqlxid, char *filename,
                      char *src, const unsigned int buflen,
                      char delim, char escapech, short int sqlind,
                      unsigned long int rownum,  int fieldnum,
                      char * fieldname, const void *logfile,
                      onlogerror_upt onlogerror, int verbose,
                      const char *datahandleoptions,
                      void *fieldstate)
/* returns:  0 if success
 *           1 if error;
 * src : pointer to the buffer;
 */
{
  char buf[BUF1_SIZE+1];  /* size enough for max size SQL datatype (non-BLOB)
                         in Informix it is CHAR(32767) */

  unsigned int j, k, len;
  FILE *fd1;
  char errstr[ERR_BUF_SIZE];
  unsigned int stat;
  char ch;
  slob_handle_t *slob_h;

  fd1=(FILE *)fd;
  if (sqltype==SQLTEXT || sqltype==SQLBYTES ) {
    if (sqltype==SQLTEXT) {
      stat=onwritetextblob(fd1, filename, (loc_t *) src, delim, escapech,
                           rownum, fieldnum, fieldname,  logfile, onlogerror,
                           verbose);
    } else {
      stat=onwritebyteblob(fd, filename, (loc_t *) src, delim, rownum,
                           fieldnum, fieldname, logfile, onlogerror, verbose);
    }
    if (stat!=0) {
      return 1;
    }
    stat=putc(delim,fd1);
    if (stat==EOF) {
      sprintf(errstr, "OS error %d:%d",errno, ferror(fd1));
      onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, rownum, rownum,
                 errno, logfile, verbose, 10);
      return 1;
    }
    return 0;
  }
  if (sqltype==CFIXBINTYPE && (sqlxid==XID_BLOB || sqlxid==XID_CLOB)
  && sqlind!=-1 ) {
    slob_h=(slob_handle_t *)fieldstate;
    stat=fprintf(fd1, "%s,", off_t2hex(slob_h->offset));

    if (stat==EOF) {
      sprintf(errstr, "OS error %d:%d",errno, ferror(fd1));
      onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, rownum, rownum,
                 errno, logfile, verbose, 10);
      return 1;
    }
    stat=onwritesmartblob(fd1, filename, sqlxid, (ifx_lo_t *) src,
                          escapech, rownum, fieldnum, fieldname,
                          slob_h, logfile, onlogerror, verbose);
    if (stat!=0) {
      return 1;
    }
    stat=fprintf(fd1, "%s,%s%c", off_t2hex(slob_h->size), slob_h->filename, delim);

    if (stat==EOF) {
      sprintf(errstr, "OS error %d:%d",errno, ferror(fd1));
      onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, rownum, rownum,
                 errno, logfile, verbose, 10);
      return 1;
    }
    return 0;
  }
  if (sqlind != -1 && src[0]=='\0' /* if we have an empty string */
  && (sqltype == CCHARTYPE ||
      sqltype == CVCHARTYPE || sqltype == CLVCHARTYPE)) {
    buf[0]=escapech; /* empty string written as '\ ' */
    buf[1]=' ';
    buf[2]='\0';
    stat=fputs(buf, fd1);
    if (stat==EOF) {
      sprintf(errstr, "OS error %d:%d",errno, ferror(fd1));
      onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, rownum, rownum,
                 errno, logfile, verbose, 10);
      return 1;
    }
  } else {
    len = stleng(src);
    j = 0;
    while ( j < len) {
      k=0;
      while (k<BUF1_SIZE-2 && j<len) {
        ch=src[j];
        if (ch == '\n' || ch == escapech || ch == delim){
          buf[k]=escapech;
          k++;
        }
        buf[k]=ch;
        j++;
        k++;
      }
      buf[k]='\0';
      stat=fputs(buf,fd1);
      if (stat==EOF) {
        sprintf(errstr, "OS error %d:%d",errno, ferror(fd1));
        onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, rownum, rownum,
                   errno, logfile, verbose, 10);
        return 1;
      }
    }
  }
  stat=putc(delim,fd1);
  if (stat==EOF) {
    sprintf(errstr, "OS error %d:%d",errno, ferror(fd1));
    onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, rownum, rownum,
               errno, logfile, verbose, 10);
    return 1;
  }
  return 0;
}

int onfinish_stdu(char *fname, char *tablename, void *handle,
                  unsigned long int rownum, const void *logfile,
                  onlogerror_upt onlogerror, int verbose)
{ char errstr[ERR_BUF_SIZE];

  sprintf(errstr, "%ld records were unloaded", rownum);
  onlogerror(__FILE__, __LINE__, errstr, 0, 0, NULL, rownum, rownum,
             0, logfile, verbose, 30);
  return 0;
}

int onrecordbegin_stdu(void *fd, char *filename, unsigned long int rownum,
                       const void *logfile,
                       onlogerror_upt onlogerror, int verbose,
                       const char *datahandleoptions)
{
  return 0;
}
